Require Import String.
Require Import List.
Open Scope string_scope.

Inductive guard :=
|Gu : string -> guard.

Inductive transition_label :=
|Transition_label : string -> list (string * string) -> transition_label.

Inductive transition :=
|Local : string -> string -> transition_label -> guard -> bool -> transition
|From_sub : string -> string -> string -> transition_label -> guard -> bool -> transition
|To_sub : string -> string -> string -> transition_label -> guard -> bool -> transition.

Inductive astd :=
|Elem : string -> astd
|Automata : string -> listOfAstdStates -> list transition -> list string -> list string -> string -> astd
with listOfAstdStates :=
|EmptyAstdStates : listOfAstdStates
|ConsAstdStates : astd -> listOfAstdStates -> listOfAstdStates.

Scheme astd_mut := Induction for astd Sort Prop
                   with listOfAstdStates_mut := Induction for listOfAstdStates Sort Prop.


Inductive isLabelOfTransition : transition_label -> transition -> Prop :=
|IsLabelOfTransition_local :
   forall (n1 n2 : string)
          (label : transition_label)
          (gu : guard)
          (final : bool),
     isLabelOfTransition label (Local n1 n2 label gu final)
|IsLabelOfTransition_fromSub :
   forall (n1 n2 n1b : string)
          (label : transition_label)
          (gu : guard)
          (final : bool),
     isLabelOfTransition label (From_sub n1b n2 n1 label gu final)
|IsLabelOfTransition_toSub :
   forall (n1 n2 n2b : string)
          (label : transition_label)
          (gu : guard)
          (final : bool),
     isLabelOfTransition label (To_sub n1 n2b n2 label gu final).


Inductive isAstdNamedN : astd -> string -> Prop :=
|IsAstdNamedN_elem :
   forall (name : string),
     isAstdNamedN (Elem name) name
|IsAstdNamedN_automata :
   forall (astdName initName : string)
          (states : listOfAstdStates)
          (transitions : list transition)
          (deepFinal shallowFinal : list string),
     isAstdNamedN
       (Automata astdName
                 states
                 transitions
                 shallowFinal
                 deepFinal
                 initName)
       astdName.

Inductive isAstdNamedNFromAstdList : astd -> string -> listOfAstdStates -> Prop :=
|IsAstdNamedNFromAstdList_isElem :
   forall (searchedAstd : astd)
          (astdList_queue : listOfAstdStates)
          (name : string),
     isAstdNamedN searchedAstd name ->
     isAstdNamedNFromAstdList searchedAstd name (ConsAstdStates searchedAstd astdList_queue)
|IsAstdNamedNFromAstdList_isNotElem :
   forall (searchedAstd a : astd)
          (astdList_queue : listOfAstdStates)
          (name : string),
     isAstdNamedNFromAstdList searchedAstd name (astdList_queue) -> 
     isAstdNamedNFromAstdList searchedAstd name (ConsAstdStates a astdList_queue).

Inductive isAutomata : astd -> Prop :=
|IsAutomata : forall (name init : string)
                     (states : listOfAstdStates)
                     (transitions : list transition)
                     (deepFinal shallowFinal : list string),
                isAutomata (Automata name
                                     states
                                     transitions
                                     shallowFinal
                                     deepFinal
                                     init).

Inductive isAstdNameStateList : listOfAstdStates -> list string -> Prop :=
|IsAstdNameStateList_empty :
   isAstdNameStateList EmptyAstdStates nil
|IsAstdNameStateList_notEmpty :
   forall (a : astd)
          (s : string)
          (li : list string)
          (queue_astd : listOfAstdStates),
     isAstdNamedN a s ->
     isAstdNameStateList queue_astd li ->
     isAstdNameStateList (ConsAstdStates a queue_astd) (s::li).

Inductive stringListIsIncludedInStringList : list string -> list string -> Prop :=
|StringListIsIncludedInStringList_nil :
   forall (s : list string),
     stringListIsIncludedInStringList nil s
|StringListIsIncludedInStringList_notNil :
   forall (s1 s2 : list string)
          (s : string),
     stringListIsIncludedInStringList s1 s2 ->
     In s s2 ->
     stringListIsIncludedInStringList (s::s1) s2.

Inductive isAstdStateList : listOfAstdStates -> astd -> Prop :=
|IsAstdStateList_automata :
   forall (states : listOfAstdStates)
          (name init : string)
          (transitions : list transition)
          (dFinal sFinal : list string),
     isAstdStateList states (Automata name states transitions sFinal dFinal init).

Inductive transitionIsWellDefined : transition -> listOfAstdStates -> list string -> Prop :=
|TransitionIsWellDefined_local :
   forall (fromS toS : string)
          (label : transition_label)
          (states : listOfAstdStates)
          (stateNames : list string)
          (gu : guard)
          (final : bool),
     In fromS stateNames ->
     In toS stateNames ->
     transitionIsWellDefined (Local fromS toS label gu final) states stateNames
|TransitionIsWellDefined_fromSub :
   forall (fromS toS throughS : string)
          (label : transition_label)
          (throughAstd : astd)
          (states statesOfSub : listOfAstdStates)
          (stateNames stateNamesOfSub : list string)
          (gu : guard)
          (final : bool),
     In throughS stateNames ->
     In toS stateNames ->
     isAstdNamedNFromAstdList throughAstd throughS states ->
     isAstdStateList statesOfSub throughAstd ->
     isAstdNameStateList statesOfSub stateNamesOfSub ->
     In fromS stateNamesOfSub ->
     transitionIsWellDefined (From_sub fromS toS throughS label gu final) states stateNames
|TransitionIsWellDefined_toSub_notHistory :
   forall (fromS toS throughS : string)
          (label : transition_label)
          (throughAstd : astd)
          (states statesOfSub : listOfAstdStates)
          (stateNames stateNamesOfSub : list string)
          (gu : guard)
          (final : bool),
     In throughS stateNames ->
     In fromS stateNames ->
     isAstdNamedNFromAstdList throughAstd throughS states ->
     isAstdStateList statesOfSub throughAstd ->
     isAstdNameStateList statesOfSub stateNamesOfSub ->
     In toS stateNamesOfSub ->
     transitionIsWellDefined (To_sub fromS toS throughS label gu final) states stateNames
|TransitionIsWellDefined_toSub_history :
   forall (fromS toS throughS : string)
          (label : transition_label)
          (states : listOfAstdStates)
          (stateNames : list string)
          (gu : guard)
          (final : bool),
     In throughS stateNames ->
     In fromS stateNames ->
     toS = "H" \/ toS = "H*" ->
     transitionIsWellDefined (To_sub fromS toS throughS label gu final) states stateNames.

Inductive transitionsAreWellDefined : list transition -> listOfAstdStates -> list string -> Prop :=
|TransitionsAreWellDefined_nil :
   forall (states : listOfAstdStates)
          (stateNames : list string),
     transitionsAreWellDefined nil states stateNames
|TransitionsAreWellDefined_notNil :
   forall (tr : transition)
          (transitions : list transition)
          (stateNames : list string)
          (states : listOfAstdStates),
     transitionsAreWellDefined transitions states stateNames ->
     transitionIsWellDefined tr states stateNames ->
     transitionsAreWellDefined (tr::transitions) states stateNames.
          

Inductive noAstdIsNamedNInAstd : string -> astd -> Prop :=
|NoAstdIsNamedNInAstd_elem :
   forall (s1 s2 : string),
     s1 <> s2 ->
     noAstdIsNamedNInAstd s2 (Elem s1)
|NoAstdIsNamedNInAstd_automata :
   forall (s2 name init : string)
          (states : listOfAstdStates)
          (transitions : list transition)
          (sFinal dFinal : list string),
     noAstdIsNamedNInSubAstdList s2 states ->
     s2 <> name ->
     noAstdIsNamedNInAstd s2
                          (Automata name states transitions sFinal dFinal init)
with noAstdIsNamedNInSubAstdList : string -> listOfAstdStates -> Prop :=
|NoAstdIsNamedNInSubAstdList_empty :
   forall (s : string),
     noAstdIsNamedNInSubAstdList s EmptyAstdStates
|NoAstdIsNamedNInSubAstdList_notEmpty :
   forall (a : astd)
          (listAstdQueue : listOfAstdStates)
          (s : string),
     noAstdIsNamedNInAstd s a ->
     noAstdIsNamedNInSubAstdList s listAstdQueue ->
     noAstdIsNamedNInSubAstdList s (ConsAstdStates a listAstdQueue).

Inductive astdNameAreUniqueInAstd : astd -> Prop :=
|AstdNameAreUnique_elem :
   forall (s : string),
     astdNameAreUniqueInAstd (Elem s)
|AstdNameAreUnique_automata :
   forall (name init : string)
          (states : listOfAstdStates)
          (transitions : list transition)
          (sFinal dFinal : list string),
     noAstdIsNamedNInSubAstdList name states ->
     astdNameAreUniqueInAstdList states ->
     astdNameAreUniqueInAstd (Automata name states transitions sFinal dFinal init)
with astdNameAreUniqueInAstdList : listOfAstdStates -> Prop :=
|AstdNameAreUniqueInAstdList_empty :
   astdNameAreUniqueInAstdList EmptyAstdStates
|AstdNameAreUniqueInAstdList_notEmpty :
   forall (a : astd)
          (listAstdQueue : listOfAstdStates),
     astdNameAreUniqueInAstd a ->
     astdNameAreUniqueInAstdList listAstdQueue ->
     (forall (s : string),
        noAstdIsNamedNInAstd s a \/
        noAstdIsNamedNInSubAstdList s listAstdQueue) ->
     astdNameAreUniqueInAstdList (ConsAstdStates a listAstdQueue).

Inductive isWellDefined : astd -> Prop :=
|IsWellDefined_elem :
   forall (s : string),
     isWellDefined (Elem s)
|IsWellDefined_automata :
   forall (name init : string)
          (states : listOfAstdStates)
          (transitions : list transition)
          (dFinal sFinal stateNameList : list string),
     isAstdNameStateList states stateNameList ->
     isAstdListWellDefined states ->
     stringListIsIncludedInStringList dFinal stateNameList ->
     stringListIsIncludedInStringList sFinal stateNameList ->
     (exists (a : astd), isAstdNamedNFromAstdList a init states) ->
     astdNameAreUniqueInAstd (Automata name states transitions sFinal dFinal init) ->
     transitionsAreWellDefined transitions states stateNameList ->
     isWellDefined (Automata name states transitions sFinal dFinal init)
with isAstdListWellDefined : listOfAstdStates -> Prop :=
     |IsAstdListWellDefined_empty :
        isAstdListWellDefined EmptyAstdStates
     |IsAstdListWellDefined_notEmpty :
        forall (a : astd)
               (li : listOfAstdStates),
          isAstdListWellDefined li ->
          isWellDefined a ->
          isAstdListWellDefined (ConsAstdStates a li).

Inductive isASubQueueOfStates : listOfAstdStates -> listOfAstdStates -> Prop :=
|IsASubQueueOfStates_isElem :
   forall (states : listOfAstdStates),
     isASubQueueOfStates states states
|IsASubQueueOfStates_isNotElem :
   forall (a : astd)
          (states1 states2 : listOfAstdStates),
     isASubQueueOfStates states1 states2 ->
     isASubQueueOfStates states1 (ConsAstdStates a states2).

Theorem queueIsASubQueue :
  forall (a : astd)
         (states states2 : listOfAstdStates),
    isASubQueueOfStates (ConsAstdStates a states) states2 ->
    isASubQueueOfStates states states2.
Proof.
  induction states2.
  -
    intros IsASubQueue.
    inversion IsASubQueue.
  -
    intros IsASubQueue.
    inversion IsASubQueue;subst.
    +
      apply (IsASubQueueOfStates_isNotElem).
      apply (IsASubQueueOfStates_isElem).
    +
      apply (IsASubQueueOfStates_isNotElem).
      apply (IHstates2);assumption.
Qed.
