Require Import guard.
Require Import ASTD.
Require Import states.
Require Import String.
Require Import List.

Inductive isStateFinalInAstd : state -> astd -> Prop :=
|IsStateFinalInAstd_elem :
   forall (s : string),
     isStateFinalInAstd Elem_state (Elem s)
|IsStateFinalInAstd_shallowFinal :
   forall (stateName astdName initName : string)
          (transitions : list transition)
          (states : listOfAstdStates)
          (subState : state)
          (shallowFinal deepFinal : list string)
          (history : histState),
     In stateName shallowFinal ->
     isStateFinalInAstd
       (Automata_state stateName
                       history
                       subState)
       (Automata astdName
                 states
                 transitions
                 shallowFinal
                 deepFinal
                 initName)
|IsStateFinalInAstd_deepFinal :
   forall (stateName astdName initName : string)
          (subAstdOfSubState : astd)
          (transitions : list transition)
          (states : listOfAstdStates)
          (subState : state)
          (shallowFinal deepFinal : list string)
          (history : histState),
     In stateName deepFinal ->
     isStateFinalInAstd subState subAstdOfSubState ->
     isAstdNamedNFromAstdList subAstdOfSubState stateName states ->
     isStateFinalInAstd
       (Automata_state stateName
                       history
                       subState)
       (Automata astdName
                 states
                 transitions
                 shallowFinal
                 deepFinal
                 initName).  

Inductive isInitialStateOfAstd : state -> astd -> Prop :=
|IsInitialStateOfAstd_elem : forall (name : string),
                         isInitialStateOfAstd Elem_state (Elem name)
|IsInitialStateOfAstd_automata :
   forall (name init : string)
          (subAstd : astd)
          (states : listOfAstdStates)
          (transitions : list transition)
          (hInit : histState)
          (deepFinals shallowFinals : list string)
          (subState : state),
     isInitialStateOfAstd subState subAstd ->
     isAstdNamedNFromAstdList subAstd init states ->
     isHInit hInit states ->
     isInitialStateOfAstd (Automata_state init
                                          hInit
                                          subState)
                          (Automata name
                                    states
                                    transitions
                                    deepFinals
                                    shallowFinals
                                    init)
with isHInit : histState -> listOfAstdStates -> Prop :=
     |IsHInit_nil :
        isHInit EmptyHistState EmptyAstdStates
     |IsHInit_elem : forall (name : string)
                            (hInit : histState)
                            (states_queue : listOfAstdStates),
                       isHInit hInit states_queue ->
                       isHInit hInit (ConsAstdStates (Elem name) states_queue)
     |IsHInit_automata :
        forall (name : string)
               (s : state)
               (a : astd)
               (hInit_rest : histState)
               (states_queue : listOfAstdStates),
          isAstdNamedN a name ->
          isInitialStateOfAstd s a ->
          isAutomata a ->
          isHInit hInit_rest states_queue ->
          isHInit (ConsHistState name s hInit_rest) (ConsAstdStates a states_queue).


Inductive isHistoricStateOfNFromHistory : state -> string -> histState -> Prop :=
|IsHistoricStateOfNFromHistory_isElem :
   forall (searchedState : state)
          (searchedName : string)
          (h_queue : histState),
     isHistoricStateOfNFromHistory searchedState searchedName (ConsHistState searchedName searchedState h_queue)
|IsHistoricStateOfNFromHistory_isNotElem :
   forall (searchedState s : state)
          (searchedName n : string)
          (h_queue : histState),
     isHistoricStateOfNFromHistory searchedState searchedName h_queue ->
     isHistoricStateOfNFromHistory searchedState searchedName (ConsHistState n  s h_queue).

Inductive astd_step : state -> astd -> transition_label -> state -> Prop :=
|Automata_rule1 :
   forall (n1 n2 init nameOfAutomata gu : string)
          (label : transition_label)
          (h' h : histState)
          (s initOfN2 : state)
          (a2 a1 : astd)
          (states : listOfAstdStates)
          (transitions : list transition)
          (shallowFinal deepFinal : list string)
          (final : bool),
     In (Local n1 n2 label (Gu gu) final) transitions ->
     (final = true -> isStateFinalInAstd s a1) ->
     evaluateGuard gu ->
     histIsHistOverLoadedByNameAndState h' h n1 s ->
     isAstdNamedNFromAstdList a1 n1 states ->
     isAstdNamedNFromAstdList a2 n2 states ->
     isInitialStateOfAstd initOfN2 a2 ->
     astd_step (Automata_state n1 h s)
               (Automata nameOfAutomata states transitions shallowFinal deepFinal init)
               label
               (Automata_state n2 h' initOfN2)
|Automata_rule2 :
   forall (n1 n2 n2b init initOfN2 nameOfAutomata gu : string)
          (label : transition_label)
          (h' h hInit : histState)
          (s initOfN2b : state)
          (a2b a1 : astd)
          (states statesOfN2 : listOfAstdStates)
          (transitions transitionsOfN2 : list transition)
          (shallowFinal shallowFinalOfN2 deepFinal deepFinalOfN2 : list string)
          (final : bool),
     In (To_sub n1 n2b n2 label (Gu gu) final) transitions ->
     isAstdNamedNFromAstdList (Automata n2
                                        statesOfN2
                                        transitionsOfN2
                                        shallowFinalOfN2
                                        deepFinalOfN2
                                        initOfN2) n2 states ->
     isHInit hInit statesOfN2 ->
     isAstdNamedNFromAstdList a2b n2b statesOfN2 ->
     isInitialStateOfAstd initOfN2b a2b ->
     isAstdNamedNFromAstdList a1 n1 states ->
     (final = true -> isStateFinalInAstd s a1) ->
     evaluateGuard gu ->
     histIsHistOverLoadedByNameAndState h' h n1 s ->
     astd_step (Automata_state n1 h s)
               (Automata nameOfAutomata
                         states
                         transitions
                         shallowFinal
                         deepFinal
                         init)
               label
               (Automata_state n2 h' (Automata_state n2b hInit initOfN2b))
|Automata_rule3 :
   forall (n1 n2 n2b init initOfN2 nameOfAutomata : string)
          (label : transition_label)
          (h' h hInit : histState)
          (s initOfN2b s2b : state)
          (states statesOfN2 : listOfAstdStates)
          (transitions transitionsOfN2: list transition)
          (shallowFinal shallowFinalOfN2 deepFinal deepFinalOfN2 : list string)
          (gu : string)
          (final : bool)
          (a2b a1 : astd),
     In (To_sub n1 "H" n2 label (Gu gu) final) transitions ->
     isHistoricStateOfNFromHistory s2b n2 h ->
     isStateNamedN s2b n2b ->
     isHInit hInit statesOfN2 ->
     isInitialStateOfAstd initOfN2b a2b ->
     isAstdNamedNFromAstdList a1 n1 states ->
     isAstdNamedNFromAstdList a2b n2b statesOfN2 ->
     isAstdNamedNFromAstdList (Automata n2
                                        statesOfN2
                                        transitionsOfN2
                                        shallowFinalOfN2
                                        deepFinalOfN2
                                        initOfN2) n2 states ->
     (final = true -> isStateFinalInAstd s a1) ->
     evaluateGuard gu ->
     histIsHistOverLoadedByNameAndState h' h n1 s ->
     astd_step (Automata_state n1 h s)
               (Automata nameOfAutomata
                         states
                         transitions
                         shallowFinal
                         deepFinal
                         init)
               label
               (Automata_state n2 h' (Automata_state n2b hInit initOfN2b))
|Automata_rule4 :
   forall (n1 n2 nameOfAutomata init : string)
          (label : transition_label)
          (h' h : histState)
          (transitions : list transition)
          (states : listOfAstdStates)
          (shallowFinal deepFinal : list string)
          (gu : string)
          (final : bool)
          (a1 : astd)
          (s2b s : state),
     In (To_sub n1 "H*" n2 label (Gu gu) final) transitions ->
     (final = true -> isStateFinalInAstd s a1) ->
     evaluateGuard gu ->
     histIsHistOverLoadedByNameAndState h' h n1 s ->
     isAstdNamedNFromAstdList a1 n1 states ->
     isHistoricStateOfNFromHistory s2b n2 h ->
     astd_step (Automata_state n1 h s)
               (Automata nameOfAutomata
                         states
                         transitions
                         shallowFinal
                         deepFinal
                         init)
               label
               (Automata_state n2 h' s2b)
|Automata_rule5 :
   forall (n1 n2 n1b init nameOfAutomata initOfN1 : string)
          (label : transition_label)
          (states statesOfN1 : listOfAstdStates)
          (a2 a1 : astd)
          (transitions transitionsOfN1 : list transition)
          (h h' : histState)
          (deepFinal deepFinalOfN1 shallowFinal shallowFinalOfN1 : list string)
          (gu : string)
          (final : bool)
          (s initOfn2 : state),
     In (From_sub n1b n2 n1 label (Gu gu) final) transitions ->
     (final = true -> isStateFinalInAstd s a1) ->
     evaluateGuard gu ->
     histIsHistOverLoadedByNameAndState h' h n1 s ->
     isAstdNamedNFromAstdList a1 n1b statesOfN1 ->
     isAstdNamedNFromAstdList (Automata n1
                                        statesOfN1
                                        transitionsOfN1
                                        shallowFinalOfN1
                                        deepFinalOfN1
                                        initOfN1) n1 states -> 
     isStateNamedN s n1b ->
     isInitialStateOfAstd initOfn2 a2 ->
     isAstdNamedNFromAstdList a2 n2 states ->
     astd_step (Automata_state n1 h s)
               (Automata nameOfAutomata
                         states
                         transitions
                         shallowFinal
                         deepFinal
                         init)
               label
               (Automata_state n2 h' initOfn2)
|Automata_rule6 :
   forall (n init nameOfAutomata : string)
          (label : transition_label)
          (a : astd)
          (s1 s2 : state)
          (h : histState)
          (states : listOfAstdStates)
          (transitions : list transition)
          (deepFinal shallowFinal : list string),
     astd_step s1 a label s2 ->
     isAstdNamedNFromAstdList a n states ->
     astd_step (Automata_state n h s1)
               (Automata nameOfAutomata states transitions shallowFinal deepFinal init)
               label
               (Automata_state n h s2).

Lemma initStateIsAState :
  forall (s : state)
         (a : astd),
    isInitialStateOfAstd s a ->
    isAStateOfTheAstd s a.
Proof.
  apply (state_mut (fun (s : state) =>
                     forall (a : astd),
                       isInitialStateOfAstd s a ->
                       isAStateOfTheAstd s a)
                  (fun (h : histState) =>
                     forall (states : listOfAstdStates),
                       isHInit h states ->
                       isAHistoryStateOfAstdList h states)).
  -
    intros a IsInit.
    inversion IsInit; subst. apply (IsAStateOfTheAstd_elem).
  -
    intros name h IHh s IHs a IsInitial.
    inversion IsInitial;subst.
    apply (IsAStateOfTheAstd_automata) with subAstd.
    assumption.
    apply (IHs subAstd H2).
    apply (IHh states H5).
  -
    intros states HInit.
    induction states;subst.
    apply (IsAHistoryStateOfAstdList_nil).
    inversion HInit;subst.
    apply (IsAHistoryStateOfAstdList_notNil_elem).
    apply (IHstates H1).
  -
    intros name s IHs h IHh states IsHInit.
    induction states;inversion IsHInit;subst.
    +
      apply (IsAHistoryStateOfAstdList_notNil_elem).
      apply (IHstates H1).
    +
      inversion H6;subst.
      inversion H4;subst.
      apply (IsAHistoryStateOfAstdList_notNil_automata);[apply (IHh states H7)|apply (IHs (Automata name states0 transitions shallowFinal deepFinal init) H5)].
Qed.      

Corollary HInitIsAnHistoricState :
  forall (states : listOfAstdStates)
         (h : histState),
    isHInit h states ->
    isAHistoryStateOfAstdList h states.
Proof.
  induction states.
  -
    intros h IsHInit. inversion IsHInit;subst.
    apply (IsAHistoryStateOfAstdList_nil).
  -
    intros h IsHInit. inversion IsHInit;subst.
    +
      apply (IsAHistoryStateOfAstdList_notNil_elem). apply (IHstates h H1).
    +
      inversion H4;subst. inversion H1;subst.
      apply (IsAHistoryStateOfAstdList_notNil_automata).
      apply (IHstates hInit_rest H5).
      apply (initStateIsAState s (Automata name states0 transitions shallowFinal deepFinal init) H2).
Qed.

Lemma existsAStateInHistoryIfExistsAnAstd :
  forall (states : listOfAstdStates)
         (name : string)
         (h : histState)
         (a : astd),
    isAstdNamedNFromAstdList a name states ->
    isAutomata a ->
    isAHistoryStateOfAstdList h states ->
    exists (s : state),
      isHistoricStateOfNFromHistory s name h.
Proof.
  induction states.
  -
    intros name h a IsAstdNamed. inversion IsAstdNamed.
  -
    intros name h a0 IsAstdNamed IsAutomata IsHist.
    inversion IsHist;subst;inversion IsAstdNamed;subst.
    +
      inversion IsAutomata.
    +
      apply (IHstates name h a0);assumption.
    +
      inversion H4;subst. exists hState. apply (IsHistoricStateOfNFromHistory_isElem).
    +
      destruct (IHstates name hRest a0 H4 IsAutomata H2).
      exists x. apply (IsHistoricStateOfNFromHistory_isNotElem). assumption.
Qed.

Lemma absurd_noAstdNamed_isHistoricState :
  forall (states : listOfAstdStates)
         (h : histState)
         (name : string)
         (s : state)
         (P : Prop),
    isHistoricStateOfNFromHistory s name h ->
    isAHistoryStateOfAstdList h states ->
    noAstdIsNamedNInSubAstdList name states ->
    P.
Proof.
  induction states.
  -
    intros h name s P IsHistState IsAHistory.
    inversion IsAHistory;subst. inversion IsHistState.
  -
    intros h name s P IsHistState IsAHistory NoAstdNamed.
    inversion IsAHistory;subst; inversion NoAstdNamed;subst.
    +
      apply (IHstates h name s P IsHistState H1 H4).
    +
      inversion IsHistState;subst.
      *
        inversion H4;subst. destruct H11. reflexivity.
      *
        apply (IHstates hRest name s P H6 H2 H5).
Qed.      
    
