Require Import B_state.
Require Import B_semantics.
Require Import B_langage.
Require Import String.
Require Import ASTD.
Require Import states.
Require Import ASTDstep.
Require Import stateEquivalence.
Require Import translation.
Require Import wellConstructed.
Require Import theoremsOnUniqueBState.
Require Import theoremsOnUniqueAstdList.
Require Import theoremsOnNames.
Open Scope list_scope.

Theorem listWhereReplacedTwice :
  forall (li3 li2 li1 : bStateList)
         (v1 v2 : bState)
         (name : string),
    variableNamesAreUniqueInBStateList li2 ->
    v1 <> Elem_B_state ->
    v2 <> Elem_B_state ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState li1 li2 name v1 ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState li2 li3 name v2 ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState li1 li3 name v1.
Proof.
  induction li3;subst.
  -
    intros li2 li1 v1 v2 name VarUnique NotElem1 NoElem2 ListReplaced1 ListReplaced2.
    inversion ListReplaced2;subst.
    inversion ListReplaced1;subst.
    apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_empty).
  -
    intros li2 li1 v1 v2 name VarUnique NotElem1 NoElem2 ListReplaced1 ListReplaced2.
    inversion ListReplaced2;subst;inversion ListReplaced1;subst.
    +
      apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem).
    +
      inversion VarUnique;subst.
      apply (absurd_noVariableNamed_isListWhereReplaced2 name li3 bStateReplaced v1).
      assumption.
      assumption.
      apply (H2 name val2). apply (IsValOfVarInState_isElem).
    +
      inversion VarUnique;subst.
      apply (absurd_noVariableNamed_isListWhereReplaced name li3 bStateReplaced v2).
      assumption.
      assumption.
      apply (H2 name val). apply (IsValOfVarInState_isElem).
    +
      apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isNotElem).
      inversion VarUnique;subst.
      apply (IHli3 bStateReplaced bStateReplaced0 v1 v2 name);assumption.
Qed.      


Lemma predicateIsTrueInASubState1 :
  forall (pre : predicate)
         (v : bState)
         (var val : string)
         (vLi : bStateList),
    predicateIsTrueInState pre (Automata_B_state var val (ConsBStateList v vLi)) ->
    (forall (name :string),
       variableIsUsedInPredicate name pre ->
       noVariableIsNamedNInBStateList name vLi /\ (name <> var)) ->
    predicateIsTrueInState pre v.
Proof.
  induction pre.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    destruct (NoVarNamed s).
    apply (VariableIsUsedInPredicate_equality).
    inversion H2;subst.
    +
      destruct H0.
      reflexivity.
    +    
      inversion H5;subst.
      *
        apply (PredicateIsTrueInState_equality).
        assumption.
      *
        apply (absurd_isValOfVarInList_noVariableNamedInList vLi s s0);assumption.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_and).
    +
      apply (IHpre1 v var val vLi H1).
      intros name VarUsed.
      apply (NoVarNamed name).
      apply (VariableIsUsedInPredicate_andLeft).
      assumption.
    +
      apply (IHpre2 v var val vLi H3).
      intros name VarUsed.
      apply (NoVarNamed name).
      apply (VariableIsUsedInPredicate_andRight).
      assumption.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    +
      apply (PredicateIsTrueInState_orLeft).
      apply (IHpre1 v var val vLi H2).
      intros name VarUsed.
      apply (NoVarNamed name).
      apply (VariableIsUsedInPredicate_orLeft).
      assumption.
    +
      apply (PredicateIsTrueInState_orRight).
      apply (IHpre2 v var val vLi H2).
      intros name VarUsed.
      apply (NoVarNamed name).
      apply (VariableIsUsedInPredicate_orRight).
      assumption.
  -
    intros v var val vLi PredTrue NoVarNamed.
    apply (PredicateIsTrueInState_truePredicate).
  -
    intros v var val vLi PredTrue.
    inversion PredTrue.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    destruct (NoVarNamed s).
    apply (VariableIsUsedInPredicate_isIn).
    inversion H1;subst.
    +
      destruct H0.
      reflexivity.
    +    
      inversion H6;subst.
      *
        apply (PredicateIsTrueInState_isIn) with val0;assumption.
      *
        apply (absurd_isValOfVarInList_noVariableNamedInList vLi s val0);assumption.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_guard).
    assumption.
Qed.

Lemma predicateIsTrueInASubState2 :
  forall (pre : predicate)
         (v : bState)
         (var val : string)
         (vLi : bStateList),
    predicateIsTrueInState pre (Automata_B_state var val (ConsBStateList v vLi)) ->
    (forall (name :string),
       variableIsUsedInPredicate name pre ->
       noVariableIsNamedNInBState name v) ->
    predicateIsTrueInState pre (Automata_B_state var val vLi).
Proof.
  induction pre.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_equality).
    inversion H2;subst.
    +
      apply (IsValOfVarInState_isElem).
    +
      inversion H3;subst.
      *
        apply (absurd_isValOfVar_noVariableNamed v s s0).
        assumption.
        apply (NoVarNamed s).
        apply (VariableIsUsedInPredicate_equality).
      *
        apply (IsValOfVarInState_isNotElem).
        assumption.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_and).
    +
      apply (IHpre1 v var val vLi H1).
      intros name VarUsed.
      apply (NoVarNamed name).
      apply (VariableIsUsedInPredicate_andLeft).
      assumption.
    +
      apply (IHpre2 v var val vLi H3).
      intros name VarUsed.
      apply (NoVarNamed name).
      apply (VariableIsUsedInPredicate_andRight).
      assumption.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    +
      apply (PredicateIsTrueInState_orLeft).
      apply (IHpre1 v var val vLi H2).
      intros name VarUsed.
      apply (NoVarNamed name).
      apply (VariableIsUsedInPredicate_orLeft).
      assumption.
    +
      apply (PredicateIsTrueInState_orRight).
      apply (IHpre2 v var val vLi H2).
      intros name VarUsed.
      apply (NoVarNamed name).
      apply (VariableIsUsedInPredicate_orRight).
      assumption.
  -
    intros v var val vLi PredTrue NoVarNamed.
    apply (PredicateIsTrueInState_truePredicate).
  -
    intros v var val vLi PredTrue.
    inversion PredTrue.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_isIn) with val0;[|assumption].
    inversion H1;subst.
    +
      apply (IsValOfVarInState_isElem).
    +
      inversion H4;subst.
      *
        apply (absurd_isValOfVar_noVariableNamed v s val0).
        assumption.
        apply (NoVarNamed s).
        apply (VariableIsUsedInPredicate_isIn).
      *
        apply (IsValOfVarInState_isNotElem).
        assumption.
  -
    intros v var val vLi PredTrue NoVarNamed.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_guard).
    assumption.
Qed.

Lemma isHistOfASubQueue :
  forall (states : listOfAstdStates)
         (h : histState)
         (name : string)
         (s : state),
    isAHistoryStateOfAstdList (ConsHistState name s h) states ->
    exists (a : astd)
           (states' : listOfAstdStates),
      isAutomata a /\
      isAHistoryStateOfAstdList (ConsHistState name s h) (ConsAstdStates a states') /\
      isASubQueueOfStates (ConsAstdStates a states') states.
Proof.
  induction states.
  -
    intros states name s IsHist.
    inversion IsHist;subst.
  -
    intros h name s IsHist.
    inversion IsHist;subst.
    +
      destruct (IHstates h name s H1).
      destruct H.
      destruct H.
      destruct H0.
      exists x.
      exists x0.
      split;[assumption|split;[assumption|]].
      apply (IsASubQueueOfStates_isNotElem).
      assumption.
    +
      exists (Automata name states0 transitions sFinal dFinal init).
      exists states.
      split;[|split].
      *
        apply (IsAutomata).
      *
        assumption.
      *
        apply (IsASubQueueOfStates_isElem).
Qed.

Lemma noAstdNamed_implies_noVarNamed :
  forall (s : state)
         (a : astd)
         (v : bState)
         (name : string),
    noAstdIsNamedNInAstd name a ->
    isAStateOfTheAstd s a ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s (get_name a) ->
    noVariableIsNamedNInBState name v.
Proof.
  apply (state_mut (fun (s : state) =>
                      forall (a : astd)
                             (v : bState)
                             (name : string),
                        noAstdIsNamedNInAstd name a ->
                        isAStateOfTheAstd s a ->
                        bStateAndAstdStateAreEquivalentInAstdNamed v s (get_name a) ->
                        noVariableIsNamedNInBState name v)
                   (fun (h : histState) =>
                      forall (states : listOfAstdStates)
                             (name : string)
                             (vLi : bStateList),
                        listOfBStatesIsEquivalentToHistory vLi h ->
                        isAHistoryStateOfAstdList h states ->        
                        noAstdIsNamedNInSubAstdList name states ->
                        noVariableIsNamedNInBStateList name vLi)).
  -
    intros a v name NoAstdNamed IsAState StateEquiv.
    inversion IsAState;subst.
    inversion StateEquiv;subst.
    apply (NoVariableIsNamedNInBState_elem).
  -
    intros nameOfState h IHh s IHs a v name NoAstdNamed IsAState StateEquiv. 
    inversion IsAState;subst.
    inversion StateEquiv;subst.
    simpl.
    inversion NoAstdNamed;subst.
    apply (NoVariableIsNamedNInBState_automata).
    +
      destruct subAstd.
      *
        inversion H4;subst.
        inversion H8;subst.
        rewrite (replacingByElemBState listBSubStateRes listBSubState nameOfState H9).
        apply (IHh states name listBSubState H6 H5 H3).
      *
        apply (noVariableNamedNPreservedByReplaced listBSubState listBSubStateRes name nameOfState bSubState).
        apply (IHs (Automata s0 l l0 l1 l2 s1) bSubState name).
        apply (noAstdNamedInASubState (Automata s0 l l0 l1 l2 s1) name nameOfState states H3 H2).
        assumption.
        rewrite (astdNameUnique (Automata s0 l l0 l1 l2 s1) s0 nameOfState).
        simpl.
        assumption.
        apply (IsAstdNamedN_automata).
        apply (isAstdNamedInList_implies_isAstdNamed) with states.
        assumption.
        assumption.
        apply (IHh states name listBSubState);assumption.
    +
      unfold not. intros. subst. destruct H13. reflexivity.
  -
    intros states name vLi ListEquiv IsHist NoAstdNamed.
    inversion ListEquiv;subst.
    apply (NoVariableIsNamedNInBStateList_nil).
  -
    intros nameOfState s IHs h IHh states name vLi ListEquiv IsHist NoAstdNamed.
    inversion ListEquiv;subst.
    destruct (isHistOfASubQueue states h nameOfState s IsHist).
    destruct H.
    destruct H.
    destruct H0.
    inversion H;subst.
    inversion H0;subst.
    apply (NoVariableIsNamedNInBStateList_notNil).
    +
      apply (IHh x0 name bState_queue H2 H6).
      apply (noAstdNamedNInSubQueue states x0 name NoAstdNamed).
      apply queueIsASubQueue with (Automata name0 states0 transitions shallowFinal deepFinal init);assumption.
    +
      apply (IHs (Automata name0 states0 transitions shallowFinal deepFinal init) bst name).
      assert (noAstdIsNamedNInSubAstdList name (ConsAstdStates (Automata name0 states0 transitions shallowFinal deepFinal init) x0)).
      apply (noAstdNamedNInSubQueue states (ConsAstdStates (Automata name0 states0 transitions shallowFinal deepFinal init) x0) name NoAstdNamed H1).
      inversion H3;subst.
      assumption.
      assumption.
      simpl.
      assumption. 
Qed.

Lemma noAstdNamedInAstdList_implies_noVarNamedInList :
  forall (states : listOfAstdStates)
         (h : histState)
         (name : string)
         (vLi : bStateList),
    listOfBStatesIsEquivalentToHistory vLi h ->
    isAHistoryStateOfAstdList h states ->        
    noAstdIsNamedNInSubAstdList name states ->
    noVariableIsNamedNInBStateList name vLi.
Proof.
  induction states.
  -
    intros h name vLi ListEquiv IsHist NoAstdNamed.
    inversion IsHist;subst.
    inversion ListEquiv;subst.
    apply (NoVariableIsNamedNInBStateList_nil).
  -
    intros h name vLi ListEquiv IsHist NoAstdNamed.
    inversion NoAstdNamed;subst.
    inversion IsHist;subst.
    +
      apply (IHstates h name vLi ListEquiv H1 H3).
    +
      inversion ListEquiv;subst.
      apply (NoVariableIsNamedNInBStateList_notNil).
      *
        apply (IHstates hRest name bState_queue );assumption.
      *
        apply (noAstdNamed_implies_noVarNamed hState (Automata name0 states0 transitions sFinal dFinal init) bst name H2 H5).
        simpl. assumption.
Qed.

Lemma existsHistStateEquivToBStateList :
  forall (states : listOfAstdStates)
         (vLi vLi' : bStateList)
         (h : histState)
         (name : string)
         (s : state)
         (v : bState)
         (a : astd),
    isAstdNamedNFromAstdList a name states ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi name v ->
    astdNameAreUniqueInAstdList states ->
    isAStateOfTheAstd s a ->
    isAHistoryStateOfAstdList h states ->
    listOfBStatesIsEquivalentToHistory vLi h ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    exists (h' : histState),
      listOfBStatesIsEquivalentToHistory vLi' h' /\
      isAHistoryStateOfAstdList h' states.
Proof.
  induction states.
  -
    intros vLi vLi' h name s v a IsAstdNamed ListReplaced AstdNamesUnique IsAState IsAHistoryState ListEquiv StateEquiv.
    inversion IsAHistoryState;subst.
    inversion ListEquiv;subst. inversion ListReplaced;subst.
    exists EmptyHistState.
    split;[apply ListOfBStatesIsEquivalentToHistory_nil|apply (IsAHistoryStateOfAstdList_nil)].
  -
    intros vLi vLi' h name s v a0 IsAstdNamed ListReplaced AstdNamesUnique IsAState IsHist ListEquiv StateEquiv.
    destruct a.
    +
      inversion IsHist;subst.
      inversion IsAstdNamed;subst.
      *
        inversion IsAState;subst.
        inversion StateEquiv;subst.
        rewrite_values.
        exists h. split;assumption.
      *
        inversion AstdNamesUnique;subst.
        destruct (IHstates vLi vLi' h name s v a0 H3 ListReplaced H4 IsAState H1  ListEquiv  StateEquiv).
        destruct H. exists x.
        split;[assumption|apply (IsAHistoryStateOfAstdList_notNil_elem);assumption].
    +
      inversion IsHist;subst. inversion ListEquiv;subst.
      inversion ListReplaced;subst;inversion IsAstdNamed;subst.
      *
        inversion H4;subst.
        exists (ConsHistState name s hRest).
        split;[apply (ListOfBStatesIsEquivalentToHistory_notNil);assumption|apply (IsAHistoryStateOfAstdList_notNil_automata);assumption].
      *
        inversion AstdNamesUnique;subst.
        inversion H5;subst.
        destruct (H7 s0);[inversion H;subst;destruct H19;reflexivity|apply (absurd_noAstdNamedN_isAstdNamedN_inList states a0 s0);assumption].
      *
        inversion AstdNamesUnique;subst. inversion H4;subst.
        destruct (H9 name);[inversion H;subst;destruct H17;reflexivity|].
        apply (absurd_noVariableNamed_isListWhereReplaced2 name bState_queue bStateReplaced v);[assumption| |].
        inversion IsAState;subst. inversion StateEquiv;subst. unfold not. intros. discriminate.
        apply (noAstdNamedInAstdList_implies_noVarNamedInList states hRest);assumption.
      *
        inversion AstdNamesUnique;subst.
        destruct (IHstates bState_queue bStateReplaced hRest name s v a0 H4 H7 H6 IsAState H2 H3 StateEquiv).
        destruct H.
        exists (ConsHistState s0 hState x).
        split.
        apply (ListOfBStatesIsEquivalentToHistory_notNil);assumption.
        apply (IsAHistoryStateOfAstdList_notNil_automata);assumption.
Qed.


Theorem namesUnique_implies_varNamesUnique :
  forall (a : astd)
         (s : state)
         (v : bState)         
         (name : string),
    isAStateOfTheAstd s a ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    isAstdNamedN a name ->
    astdNameAreUniqueInAstd a ->
    variableNamesAreUniqueInBState v.
Proof.
  apply (astd_mut (fun (a : astd) =>
                     forall (s : state)
                            (v : bState)
                            (name : string),
                       isAStateOfTheAstd s a ->
                       bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
                       isAstdNamedN a name ->
                       astdNameAreUniqueInAstd a ->
                       variableNamesAreUniqueInBState v)
                  (fun (states : listOfAstdStates) =>
                     forall (h : histState)
                            (vLi : bStateList),
                       isAHistoryStateOfAstdList h states ->
                       listOfBStatesIsEquivalentToHistory vLi h ->
                       astdNameAreUniqueInAstdList states ->
                       variableNamesAreUniqueInBStateList vLi)).
  -
    intros nameOfA s v name IsAState StateEquiv IsAstdNamed NamesUnique.
    inversion IsAState. subst. inversion StateEquiv. subst.
    apply (VariableNamesAreUniqueInBState_elem).
  -
    intros nameOfA states IHstates trans sF dF init s v name IsAState StateEquiv IsAstdNamed NamesUnique.
    inversion IsAState;subst. inversion StateEquiv;subst.
    inversion NamesUnique;subst. inversion IsAstdNamed;subst.
    apply (VariableNamesAreUniqueInBState_automata).
    +
      apply (noVariableNamedNPreservedByReplaced listBSubState listBSubStateRes name nameOfState bSubState).
      apply (noAstdNamed_implies_noVarNamed subState subAstd).
      apply (noAstdNamedInASubState subAstd name nameOfState states H1 H3).
      assumption.
      rewrite (get_nameWorks subAstd nameOfState).
      assumption.
      apply (isAstdNamedInList_implies_isAstdNamed subAstd nameOfState states);assumption.
      assumption.
      apply (noAstdNamedInAstdList_implies_noVarNamedInList states h name listBSubState H4 H8 H1).
    +
      destruct (existsHistStateEquivToBStateList states listBSubState listBSubStateRes h nameOfState subState bSubState subAstd H3 H9 H12 H7 H8 H4 H6).
      destruct H.
      apply (IHstates x listBSubStateRes);assumption.
  -
    intros h vLi IsHist ListEquiv NamesUnique.
    inversion IsHist;subst. inversion ListEquiv. subst.
    apply (VariableNamesAreUniqueInBStateList_nil).
  -
    intros a IHa states IHstates h vLi IsHist ListEquiv NamesUnique.
    destruct a.
    +
      inversion IsHist;subst.
      inversion NamesUnique;subst.
      apply (IHstates h vLi);assumption.
    +
      inversion IsHist;subst.
      inversion ListEquiv;subst.
      inversion NamesUnique;subst.
      apply (variableNamesAreUniqueInBStateList_otherDef).
      *
        apply (IHa hState bst s H8 H5). apply (IsAstdNamedN_automata). assumption.
      *
        apply (IHstates hRest bState_queue H2 H3 H4).
      *
        intros name.
        destruct (H6 name).
        right. apply (noAstdNamed_implies_noVarNamed hState (Automata s l l0 l1 l2 s0));assumption.
        left. apply (noAstdNamedInAstdList_implies_noVarNamedInList states hRest);assumption.
Qed.        

Corollary namesUniqueInList_implies_varNamesUniqueInList :
  forall (states : listOfAstdStates)
         (h : histState)
         (vLi : bStateList),
    isAHistoryStateOfAstdList h states ->
    listOfBStatesIsEquivalentToHistory vLi h ->
    astdNameAreUniqueInAstdList states ->
    variableNamesAreUniqueInBStateList vLi.
Proof.
  induction states.
  -
    intros h vLi IsHist ListEquiv NamesUnique.
    inversion IsHist. subst. inversion ListEquiv. subst.
    apply (VariableNamesAreUniqueInBStateList_nil).
  -
    intros h vLi IsHist ListEquiv NamesUnique.
    destruct a.
    +
      inversion IsHist;subst.
      inversion NamesUnique;subst.
      apply (IHstates h vLi);assumption.
    +
      inversion IsHist;subst.
      inversion ListEquiv;subst.
      inversion NamesUnique;subst.
      apply (variableNamesAreUniqueInBStateList_otherDef).
      *
        apply (namesUnique_implies_varNamesUnique (Automata s l l0 l1 l2 s0) hState bst s);[assumption|assumption|apply IsAstdNamedN_automata|assumption].
      *
        apply (IHstates hRest bState_queue H2 H3 H4).
      *
        intros name.
        destruct (H6 name).
        right. apply (noAstdNamed_implies_noVarNamed hState (Automata s l l0 l1 l2 s0));assumption.
        left. apply (noAstdNamedInAstdList_implies_noVarNamedInList states hRest);assumption.
Qed.

Lemma aVarUsedInFinalPredIsInTheBState :
  forall (a : astd)
         (s : state)
         (v : bState)
         (var nameOfA : string),
    variableIsUsedInPredicate var (finalPredicate a) ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s nameOfA ->
    isAstdNamedN a nameOfA ->
    isAStateOfTheAstd s a ->
    astdNameAreUniqueInAstd a ->
    exists (val : string),
      isValOfVarInState val var v.
Proof.
  apply (astd_mut (fun (a :astd) =>
                     forall (s : state)
                            (v : bState)
                            (var nameOfA : string),
                       variableIsUsedInPredicate var (finalPredicate a) ->
                       bStateAndAstdStateAreEquivalentInAstdNamed v s nameOfA ->
                       isAstdNamedN a nameOfA ->
                       isAStateOfTheAstd s a ->
                       astdNameAreUniqueInAstd a ->
                       exists (val : string),
                         isValOfVarInState val var v)
                  (fun (states : listOfAstdStates) =>
                     forall (s : state)
                            (trans : list transition)
                            (sF dF : list string)
                            (v : bState)
                            (var name init : string),
                       variableIsUsedInPredicate var (makeDFinalPred dF states name) ->
                       isAStateOfTheAstd s (Automata name states trans sF dF init) ->
                       bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
                       astdNameAreUniqueInAstd (Automata name states trans sF dF init) ->
                       exists (val : string),
                         isValOfVarInState val var v)).
  -
    intros name s v var nameOfA VarUsed.
    inversion VarUsed.
  -
    intros name states IHstates trans sF dF init s v var nameOfA.
    intros VarUsed StateEquiv IsAstdNamed IsAState NamesUnique.
    inversion IsAstdNamed;subst.
    inversion IsAState;subst.
    inversion StateEquiv;subst.
    inversion NamesUnique;subst.
    simpl in VarUsed.
    destruct dF;destruct sF.
    +
      inversion VarUsed.
    +
      inversion VarUsed;subst.
      exists nameOfState.
      apply (IsValOfVarInState_isElem).
    +
      apply (IHstates (Automata_state nameOfState h subState) trans nil (s::dF) (Automata_B_state nameOfA nameOfState listBSubStateRes) var nameOfA init VarUsed IsAState StateEquiv).
      apply (AstdNameAreUnique_automata);assumption.
    +
      inversion VarUsed;subst.
      *
        inversion H2;subst.
        exists nameOfState.
        apply (IsValOfVarInState_isElem).
      *
        apply (IHstates (Automata_state nameOfState h subState) trans (s0::sF) (s::dF) (Automata_B_state nameOfA nameOfState listBSubStateRes) var nameOfA init H2 IsAState StateEquiv).
        apply (AstdNameAreUnique_automata);assumption.
  -
    intros s trans sF dF v var name init VarUsed.
    inversion VarUsed.
  -
    intros a IHa states IHstates s trans sF dF v var name init VarUsed IsAState StateEquiv NamesUnique.
    simpl in VarUsed.
    destruct (List.in_dec string_dec (get_name a) dF).
    +
      inversion VarUsed;subst.
      *
        {
          inversion H1;subst.
          -
            inversion H2;subst.
            inversion IsAState;subst.
            inversion StateEquiv;subst.
            exists nameOfState.
            apply (IsValOfVarInState_isElem).
          -
            inversion IsAState;subst.
            destruct a.
            +
              inversion H2.
            +
              inversion H10;subst.
              inversion StateEquiv;subst.
              inversion H7;subst.
              inversion H12;subst.
              *
                inversion H14;subst.
                destruct (IHa subState (Automata_B_state s val2 bStateLi2) var s H2 H11).
                apply (IsAstdNamedN_automata).
                rewrite <- (isAstdNamedN_uniqueInAstdList subAstd (Automata s l l0 l1 l2 s0) (ConsAstdStates (Automata s l l0 l1 l2 s0) states) s).
                assumption.
                assumption.
                apply (IsAstdNamedNFromAstdList_isElem).
                apply (IsAstdNamedN_automata).
                inversion NamesUnique;subst.
                assumption.
                inversion NamesUnique;subst.
                inversion H21;subst.
                assumption.
                exists x.
                apply (IsValOfVarInState_isNotElem).
                apply (IsValOfVarInListOfBStates_isElem).
                assumption.
              *
                destruct (IHa hState bst var s H2 H14).
                apply (IsAstdNamedN_automata).
                assumption.
                inversion NamesUnique;subst.
                inversion H19;subst.
                assumption.
                exists x.
                apply IsValOfVarInState_isNotElem.
                apply IsValOfVarInListOfBStates_isElem.
                assumption.
        }
      *
        inversion IsAState;subst.
        inversion NamesUnique;subst.
        inversion H4;subst.
        {
          inversion H9;subst.
          -
            inversion H8;subst.
            inversion H9;subst.
            destruct states;[inversion H1|].
            destruct a.
            +
              inversion StateEquiv;subst.
              destruct (IHstates (Automata_state s h Elem_state) trans sF dF (Automata_B_state name s listBSubState) var name init H1).
              *
                apply (IsAStateOfTheAstd_automata) with (Elem s).
                apply (IsAstdNamedNFromAstdList_isElem).
                apply (IsAstdNamedN_elem).
                apply (IsAStateOfTheAstd_elem).
                assumption.
              *
                apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name s h Elem_state Elem_B_state listBSubState listBSubState H12).
                apply (BStateAndAstdStateAreEquivalentInAstdNamed_elem).
                apply (replacingByElemBState_reverse).
              *
                apply (AstdNameAreUnique_automata).
                inversion H2;subst.
                assumption.
                inversion H10;subst.
                assumption.
              *
                inversion H14;subst.
                inversion H;subst.
                exists nameOfState.
                apply (IsValOfVarInState_isElem).
                exists x.
                rewrite (replacingByElemBState listBSubStateRes listBSubState nameOfState H15).
                apply (IsValOfVarInState_isNotElem).
                assumption.
            +
              inversion H6;subst.
              inversion StateEquiv;subst.
              inversion H15;subst.
              destruct (IHstates (Automata_state s (ConsHistState s hState hRest) hState) trans sF dF (Automata_B_state name s listBSubState) var name init H1).
              *
                apply (IsAStateOfTheAstd_automata) with (Automata s l l0 l1 l2 s0).
                apply (IsAstdNamedNFromAstdList_isElem).
                apply (IsAstdNamedN_automata).
                assumption.
                apply (IsAHistoryStateOfAstdList_notNil_automata);assumption.
              *
                inversion H13;subst.
                apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name s (ConsHistState s hState hRest) hState bst (ConsBStateList bst bState_queue)).
                assumption.
                assumption.
                inversion H17;subst.
                inversion H18;subst.
                apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem).
              *
                apply (AstdNameAreUnique_automata).
                inversion H2;subst.
                assumption.
                inversion H10;subst.
                assumption.
              *
                inversion H;subst.
                exists nameOfState.
                apply (IsValOfVarInState_isElem).
                exists x.
                apply (IsValOfVarInState_isNotElem).
                rewrite (replacingByElemBState listBSubStateRes listBSubState nameOfState H16).
                assumption.
          -
            inversion StateEquiv;subst.
            inversion H12;subst.
            inversion H5;subst.
            destruct states;[inversion H1|].
            destruct a;inversion H6;subst.
            +
              destruct (IHstates (Automata_state s hRest Elem_state) trans sF dF (Automata_B_state name s bState_queue) var name init H1).
              *
                apply (IsAStateOfTheAstd_automata) with (Elem s).
                apply (IsAstdNamedNFromAstdList_isElem).
                apply (IsAstdNamedN_elem).
                apply (IsAStateOfTheAstd_elem).
                assumption.
              *
                inversion H12;subst.
                apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name s hRest Elem_state Elem_B_state bState_queue bState_queue H17).
                apply (BStateAndAstdStateAreEquivalentInAstdNamed_elem).
                apply (replacingByElemBState_reverse).
              *
                apply (AstdNameAreUnique_automata).
                inversion H2.
                assumption.
                inversion H10.
                assumption.
              *
                inversion H;subst.
                exists nameOfState.
                apply (IsValOfVarInState_isElem).
                exists x.
                apply (IsValOfVarInState_isNotElem).
                inversion H15;subst.
                apply (IsValOfVarInListOfBStates_isNotElem).
                assumption.
                inversion H10;subst.
                destruct (H20 nameOfState);[inversion H0;destruct H29;reflexivity|].
                inversion H0;subst.
                apply (absurd_noVariableNamed_isListWhereReplaced2 nameOfState bState_queue bStateReplaced bSubState (isValOfVarInListOfBStates x var (ConsBStateList bst bStateReplaced)) H21).
                inversion H8;subst.
                inversion H14;subst.
                unfold not.
                intros.
                discriminate.
                apply (noAstdNamedInAstdList_implies_noVarNamedInList states hRest nameOfState bState_queue H11 H3 H25).
            +
              inversion H12;subst.
              destruct (IHstates (Automata_state s (ConsHistState s hState0 hRest0) hState0) trans sF dF (Automata_B_state name s bState_queue) var name init H1).
              *
                apply (IsAStateOfTheAstd_automata) with (Automata s l l0 l1 l2 s0).
                apply (IsAstdNamedNFromAstdList_isElem).
                apply (IsAstdNamedN_automata).
                assumption.
                assumption.
              *
                inversion H17;subst.
                apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name s (ConsHistState s hState0 hRest0) hState0 bst0 (ConsBStateList bst0 bState_queue0) (ConsBStateList bst0 bState_queue0) H17 H21).
                inversion H22;subst.
                inversion H21;subst.
                apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem).
              *
                apply (AstdNameAreUnique_automata).
                inversion H2;subst.
                assumption.
                inversion H10;subst.
                assumption.
              *
                inversion H;subst.
                exists nameOfState.
                apply (IsValOfVarInState_isElem).
                exists x.
                apply (IsValOfVarInState_isNotElem).
                inversion H15;subst.
                apply (IsValOfVarInListOfBStates_isNotElem).
                assumption.
                apply (absurd_noVariableNamed_isListWhereReplaced2 nameOfState bState_queue bStateReplaced bSubState).
                assumption.
                inversion H8;subst.
                inversion H14;subst.
                unfold not.
                intros.
                discriminate.
                inversion H10;subst.
                destruct (H23 nameOfState);[inversion H0;subst;destruct H32;reflexivity|].
                apply (noAstdNamedInAstdList_implies_noVarNamedInList (ConsAstdStates (Automata s l l0 l1 l2 s0) states) (ConsHistState s hState0 hRest0) nameOfState bState_queue H17 H6 H0).
        }
        {
          inversion H9;subst.
          -
            apply (IHstates (Automata_state nameOfState h subState) trans sF dF v var name init H1).
            apply (IsAStateOfTheAstd_automata) with subAstd.
            assumption. assumption. assumption. assumption.
            inversion H2;subst.
            inversion H10;subst.
            apply (AstdNameAreUnique_automata);assumption.
          -
            inversion StateEquiv;subst.
            inversion H12;subst.
            inversion H15;subst.
            +
              inversion H10;subst.
              destruct (H17 nameOfState).
              inversion H16;subst.
              inversion H;subst.
              destruct H28. reflexivity.
              apply (absurd_noAstdNamedN_isAstdNamedN_inList states subAstd nameOfState);assumption.
            +
              destruct (IHstates (Automata_state nameOfState hRest subState) trans sF dF (Automata_B_state name nameOfState bStateReplaced) var name init H1).
              *
                apply (IsAStateOfTheAstd_automata) with subAstd;assumption.
              *
                apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name nameOfState hRest subState bSubState bState_queue);assumption.
              *
                inversion H2;subst.
                inversion H10;subst.
                apply (AstdNameAreUnique_automata);assumption.
              *
                inversion H;subst.
                exists nameOfState.
                apply (IsValOfVarInState_isElem).
                exists x.
                apply (IsValOfVarInState_isNotElem).
                apply (IsValOfVarInListOfBStates_isNotElem).
                assumption.
        }
    +
      inversion IsAState;subst.
      destruct states;[inversion VarUsed|].
      inversion H3;subst.
      *
        inversion StateEquiv;subst.
        inversion H8;subst.
        {
          inversion H2;subst.
          inversion H7;subst.
          inversion H9;subst.
          assert (listBSubStateRes = listBSubState).
          apply (replacingByElemBState) with nameOfState;assumption.
          subst.
          destruct a0.
          -
            destruct (IHstates (Automata_state s h Elem_state) trans sF dF (Automata_B_state name s listBSubState) var name init VarUsed).
            +
              apply (IsAStateOfTheAstd_automata) with (Elem s).
              apply (IsAstdNamedNFromAstdList_isElem). apply (IsAstdNamedN_elem).
              apply (IsAStateOfTheAstd_elem).
              assumption.
            +
              apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name s h Elem_state Elem_B_state listBSubState listBSubState H5).
              apply (BStateAndAstdStateAreEquivalentInAstdNamed_elem).
              apply (replacingByElemBState_reverse).
            +
              inversion NamesUnique;subst.
              inversion H4;subst.
              inversion H14;subst.
              apply (AstdNameAreUnique_automata);assumption.
            +
              inversion H;subst.
              exists nameOfState.
              apply (IsValOfVarInState_isElem).
              exists x.
              apply (IsValOfVarInState_isNotElem).
              assumption.
          -
            inversion H1;subst.
            destruct (IHstates (Automata_state s (ConsHistState s hState hRest) hState) trans sF dF (Automata_B_state name s listBSubState) var name init VarUsed).
            +
              apply (IsAStateOfTheAstd_automata) with (Automata s l l0 l1 l2 s0).
              apply (IsAstdNamedNFromAstdList_isElem).
              apply (IsAstdNamedN_automata).
              assumption.
              assumption.
            +
              inversion H5;subst.
              apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name s (ConsHistState s hState hRest) hState bst (ConsBStateList bst bState_queue) (ConsBStateList bst bState_queue)).
              assumption.
              assumption.
              inversion H16;subst.
              inversion H13;subst.
              apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem).
            +
              inversion NamesUnique;subst.
              inversion H4;subst.
              inversion H15;subst.
              apply (AstdNameAreUnique_automata);assumption.
            +
              inversion H;subst.
              exists nameOfState.
              apply IsValOfVarInState_isElem.
              exists x.
              apply (IsValOfVarInState_isNotElem).
              assumption.
        }
        inversion H2;subst.
        inversion H5;subst.
        inversion H6;subst.
        inversion H13;subst.
        inversion H7;subst.
        inversion H10;subst.
        {
          destruct a0.
          -
            inversion H4;subst.
            destruct (IHstates (Automata_state s hRest Elem_state)trans sF dF (Automata_B_state name s bState_queue) var name init VarUsed).
            +
              apply (IsAStateOfTheAstd_automata) with (Elem s).
              apply (IsAstdNamedNFromAstdList_isElem).
              apply (IsAstdNamedN_elem).
              apply (IsAStateOfTheAstd_elem).
              apply IsAHistoryStateOfAstdList_notNil_elem.
              assumption.
            +
              apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name s hRest Elem_state Elem_B_state bState_queue bState_queue).
              assumption.
              apply (BStateAndAstdStateAreEquivalentInAstdNamed_elem).
              apply (replacingByElemBState_reverse).
            +
              inversion NamesUnique;subst.
              inversion H21;subst.
              inversion H28;subst.
              apply (AstdNameAreUnique_automata);assumption.
            +
              inversion H;subst.
              exists nameOfState.
              apply (IsValOfVarInState_isElem).
              exists x.
              apply (IsValOfVarInState_isNotElem).
              apply (IsValOfVarInListOfBStates_isNotElem).
              assumption.
          -
            inversion H4;subst.
            inversion H11;subst.
            destruct (IHstates (Automata_state s (ConsHistState s hState hRest0) hState) trans sF dF (Automata_B_state name s (ConsBStateList bst bState_queue0)) var name init VarUsed).
            +
              apply (IsAStateOfTheAstd_automata) with (Automata s l l0 l1 l2 s0).
              apply (IsAstdNamedNFromAstdList_isElem).
              apply (IsAstdNamedN_automata).
              assumption.
              apply (IsAHistoryStateOfAstdList_notNil_automata).
              assumption.
              assumption.
            +
              apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name s (ConsHistState s hState hRest0) hState bst (ConsBStateList bst bState_queue0) (ConsBStateList bst bState_queue0)).
              assumption.
              assumption.
              inversion H29;subst.
              inversion H26;subst.
              apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem).
            +
              inversion NamesUnique;subst.
              inversion H12;subst.
              inversion H31;subst.
              apply (AstdNameAreUnique_automata);assumption.
            +
              inversion H;subst.
              exists nameOfState.
              apply (IsValOfVarInState_isElem).
              exists x.
              apply (IsValOfVarInState_isNotElem).
              apply (IsValOfVarInListOfBStates_isNotElem).
              assumption.
        }
        apply (absurd_noVariableNamed_isListWhereReplaced2 nameOfState bState_queue bStateReplaced bSubState).
        assumption.
        inversion H9;subst.
        unfold not. discriminate.
        inversion NamesUnique;subst.
        inversion H27;subst.
        destruct (H23 nameOfState). inversion H;subst. destruct H34. reflexivity.
        apply (noAstdNamedInAstdList_implies_noVarNamedInList (ConsAstdStates a0 states) hRest);assumption.
      *
        inversion StateEquiv;subst.
        inversion H8;subst.
        {
          destruct (IHstates (Automata_state nameOfState h subState) trans sF dF (Automata_B_state name nameOfState listBSubStateRes) var name init VarUsed).
          -
            apply (IsAStateOfTheAstd_automata) with subAstd;assumption.
          -
            apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name nameOfState h subState bSubState listBSubState listBSubStateRes);assumption.
          -
            inversion NamesUnique;subst.
            inversion H4;subst.
            inversion H14;subst.
            apply (AstdNameAreUnique_automata);assumption.
          -
            inversion H;subst.
            exists nameOfState.
            apply IsValOfVarInState_isElem.
            exists x.
            apply (IsValOfVarInState_isNotElem).
            assumption.
        }
        {
          inversion H5;subst.
          inversion H6;subst.
          inversion H13;subst.
          inversion H10;subst.
          -
            inversion NamesUnique;subst.
            inversion H23;subst.
            destruct (H21 nameOfState).
            inversion H. destruct H30. reflexivity.
            apply (absurd_noAstdNamedN_isAstdNamedN_inList (ConsAstdStates a0 states) subAstd nameOfState);assumption.
          -
            destruct (IHstates (Automata_state nameOfState hRest subState) trans sF dF (Automata_B_state name nameOfState bStateReplaced) var name init VarUsed).
            +
              apply (IsAStateOfTheAstd_automata) with subAstd;assumption.
            +
              apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name nameOfState hRest subState bSubState bState_queue bStateReplaced);assumption.
            +
              inversion NamesUnique;subst.
              inversion H1;subst.
              inversion H24;subst.
              apply (AstdNameAreUnique_automata);assumption.
            +
              inversion H;subst.
              exists nameOfState.
              apply (IsValOfVarInState_isElem).
              exists x. apply (IsValOfVarInState_isNotElem). apply (IsValOfVarInListOfBStates_isNotElem). assumption.
        }
Qed.

Theorem finalPredicateIsTrueInASubState :
  forall (a : astd)
         (s : state)
         (li li' : bStateList)
         (name nameOfA : string)
         (vSub : bState),
    predicateIsTrueInState (finalPredicate a) (Automata_B_state name nameOfA li) ->
    bStateAndAstdStateAreEquivalentInAstdNamed vSub s nameOfA ->
    isAstdNamedN a nameOfA ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState li li' nameOfA vSub ->
    isAStateOfTheAstd s a ->
    variableNamesAreUniqueInBState (Automata_B_state name nameOfA li) ->
    astdNameAreUniqueInAstd a ->
    predicateIsTrueInState (finalPredicate a) vSub.
Proof.
  induction li.
  -
    intros li' name nameOfA vSub PredTrue StateEquiv IsAstdNamed ListReplaced IsAState BNamesUnique AstdNamesUnique.
    inversion ListReplaced;subst.
    inversion StateEquiv;subst.
    inversion IsAState;subst.
    simpl.
    apply (PredicateIsTrueInState_truePredicate).
  -
    intros li' name nameOfA vSub PredTrue StateEquiv IsAstdNamed ListReplaced IsAState BNamesUnique AstdNamesUnique.
    inversion ListReplaced;subst.
    +
      inversion StateEquiv;subst.
      apply (predicateIsTrueInASubState1 (finalPredicate a) (Automata_B_state nameOfA val2 bStateLi2) name nameOfA li PredTrue).
      intros name0 VarUsed.
      destruct (aVarUsedInFinalPredIsInTheBState a (Automata_state val2 h subState) (Automata_B_state nameOfA val2 bStateLi2) name0 nameOfA VarUsed StateEquiv IsAstdNamed IsAState AstdNamesUnique).
      split.
      *
        inversion BNamesUnique;subst.
        inversion H8;subst.
        apply (H9 name0 x H).
      *
        inversion BNamesUnique;subst.
        inversion H3;subst.
        unfold not. intros. subst.
        apply (absurd_isValOfVar_noVariableNamed (Automata_B_state nameOfA val2 bStateLi2) name x False H H10).
    +
      apply (predicateIsTrueInASubState2 (finalPredicate a) b name nameOfA li) in PredTrue.
      *
        apply (IHli bState name nameOfA vSub PredTrue StateEquiv IsAstdNamed H4 IsAState).
        inversion BNamesUnique;subst.
        inversion H1;subst.
        inversion H3;subst.
        apply (VariableNamesAreUniqueInBState_automata);assumption.
        assumption.
      *
        intros name0 VarUsed.
        inversion BNamesUnique;subst.
        inversion H3;subst.
        destruct (aVarUsedInFinalPredIsInTheBState a s vSub name0 nameOfA VarUsed StateEquiv IsAstdNamed IsAState AstdNamesUnique).
        apply (H6 x name0).
        apply (isValOfVarInStateListIfIsValOfVarInState ) with vSub;[assumption|].
        destruct a;[inversion VarUsed|].
        inversion IsAState;subst.
        inversion StateEquiv;subst.
        apply (isAStateOfListWhereReplaced nameOfA nameOfState bState li listBSubStateRes H4).
Qed.