Require Import String.
Require Import List.

Inductive operation_label :=
|Label : string -> list (string * string) -> operation_label.

Inductive predicate :=
|Equality : string -> string -> predicate
|And : predicate -> predicate -> predicate
|Or : predicate -> predicate -> predicate
|TruePredicate : predicate
|FalsePredicate : predicate
|IsIn : string -> list string -> predicate
|Guard : string -> predicate.

Inductive substitution :=
|Select : selectList -> substitution
|Parallel : substitution -> substitution -> substitution
|Affectation : string -> string -> substitution
|SkipSubstitution : substitution
with selectList :=
|EmptySelectList : selectList
|ConsSelectList : predicate -> substitution -> selectList -> selectList.

Scheme sub_mut := Induction for substitution Sort Prop
                  with selectList_mut := Induction for selectList Sort Prop.

Inductive operation :=
|Operation : operation_label -> predicate -> substitution -> operation.

Lemma absurd_equalityOr :
  forall (p0 p1 : predicate)
         (P : Prop),
    Or p0 p1 = p0 ->
    P.
Proof.
  induction p0;intros p1 P Eq;inversion Eq.
  apply (IHp0_1 p0_2 P H0).
Qed.


Inductive variableIsUsedInPredicate : string -> predicate -> Prop :=
|VariableIsUsedInPredicate_equality :
   forall (var val : string),
     variableIsUsedInPredicate var (Equality var val)
|VariableIsUsedInPredicate_andLeft :
   forall (var : string)
          (pred1 pred2 : predicate),
     variableIsUsedInPredicate var pred1 ->
     variableIsUsedInPredicate var (And pred1 pred2)
|VariableIsUsedInPredicate_andRight :
   forall (var : string)
          (pred1 pred2 : predicate),
     variableIsUsedInPredicate var pred2 ->
     variableIsUsedInPredicate var (And pred1 pred2)
|VariableIsUsedInPredicate_orLeft :
   forall (var : string)
          (pred1 pred2 : predicate),
     variableIsUsedInPredicate var pred1 ->
     variableIsUsedInPredicate var (Or pred1 pred2)
|VariableIsUsedInPredicate_orRight :
   forall (var : string)
          (pred1 pred2 : predicate),
     variableIsUsedInPredicate var pred2 ->
     variableIsUsedInPredicate var (Or pred1 pred2)
|VariableIsUsedInPredicate_isIn :
   forall (var : string)
          (vals : list string),
     variableIsUsedInPredicate var (IsIn var vals).

Inductive variableIsUsedAndModifiedInSubstitution : string -> substitution -> Prop :=
|VariableIsUsedAndModifiedInSubstitution_select :
   forall (var : string)
          (s : selectList),
     variableIsUsedAndModifiedInSelectList var s ->
     variableIsUsedAndModifiedInSubstitution var (Select s)
|VariableIsUsedAndModifiedInSubstitution_affectation :
   forall (var val : string),
     variableIsUsedAndModifiedInSubstitution var (Affectation var val)
|VariableIsUsedAndModifiedInSubstitution_parallel_left :
   forall (var : string)
          (sub1 sub2 : substitution),
     variableIsUsedAndModifiedInSubstitution var sub1 ->
     variableIsUsedAndModifiedInSubstitution var (Parallel sub1 sub2)
|VariableIsUsedAndModifiedInSubstitution_parallel_right :
   forall (var : string)
          (sub1 sub2 : substitution),
     variableIsUsedAndModifiedInSubstitution var sub2 ->
     variableIsUsedAndModifiedInSubstitution var (Parallel sub1 sub2)
with variableIsUsedAndModifiedInSelectList : string -> selectList -> Prop :=
|VariableIsUsedAndModifiedInSelectList_isElem :
   forall (var : string)
          (pre : predicate)
          (post : substitution)
          (s : selectList),
     variableIsUsedAndModifiedInSubstitution var post ->
     variableIsUsedInPredicate var pre ->
     variableIsUsedAndModifiedInSelectList var (ConsSelectList  pre post s)
|VariableIsUsedAndModifiedInSelectList_isNotElem :
   forall (var : string)
          (pre : predicate)
          (post : substitution)
          (s : selectList),
     variableIsUsedAndModifiedInSelectList var s ->
     variableIsUsedAndModifiedInSelectList var (ConsSelectList  pre post s).



Inductive substitutionContainsNoSelect : substitution -> Prop :=
|SubstitutionContainsNoSelect_affectation :
   forall (var val : string),
     substitutionContainsNoSelect (Affectation var val)
|SubstitutionContainsNoSelect_parallel :
   forall (sub1 sub2 : substitution),
     substitutionContainsNoSelect sub1 ->
     substitutionContainsNoSelect sub2 ->
     substitutionContainsNoSelect (Parallel sub1 sub2)
|SubstitutionContainsNoSelect_skipSubstitution :
   substitutionContainsNoSelect SkipSubstitution.
