Require Import guard.
Require Import String.
Require Import List.
Require Import B_langage.
Require Import B_state.


Inductive predicateIsTrueInState : predicate -> bState -> Prop :=
|PredicateIsTrueInState_truePredicate :
   forall (s : bState),
     predicateIsTrueInState TruePredicate s
|PredicateIsTrueInState_equality :
   forall (var val : string)
          (s : bState),
     isValOfVarInState val var s ->
     predicateIsTrueInState (Equality var val) s
|PredicateIsTrueInState_and :
   forall (p1 p2 : predicate)
          (s : bState),
     predicateIsTrueInState p1 s ->
     predicateIsTrueInState p2 s ->
     predicateIsTrueInState (And p1 p2) s
|PredicateIsTrueInState_orLeft :
   forall (p1 p2 : predicate)
          (s : bState),
     predicateIsTrueInState p1 s ->
     predicateIsTrueInState (Or p1 p2) s
|PredicateIsTrueInState_orRight :
   forall (p1 p2 : predicate)
          (s : bState),
     predicateIsTrueInState p2 s ->
     predicateIsTrueInState (Or p1 p2) s
|PredicateIsTrueInState_guard :
   forall (gu : string)
          (s : bState),
     evaluateGuard gu ->
     predicateIsTrueInState (Guard gu) s
|PredicateIsTrueInState_isIn :
   forall (var val : string)
          (setOfValues : list string)
          (bState : bState),
     isValOfVarInState val var bState ->
     In val setOfValues ->
     predicateIsTrueInState (IsIn var setOfValues) bState.

Inductive substitution_step : bState -> substitution -> bState -> Prop :=
|Substitution_step_skipSubstitution :
   forall (s1 : bState),
          substitution_step s1 SkipSubstitution s1
|Substitution_step_affectation :
   forall (s1 s2 : bState)
          (val var : string),
     isStateWhereVarChangedForVal s2 s1 var val ->
     substitution_step s1 (Affectation var val) s2
|Substitution_step_parallel :
   forall (sub1 sub2 : substitution)
          (s1 s2 s3 : bState),
     substitution_step s1 sub1 s2 ->
     substitution_step s2 sub2 s3 ->
     substitution_step s1 (Parallel sub1 sub2) s3
|Substitution_step_select_cons_isElem :
   forall (gu1 : predicate)
          (sub1 : substitution)
          (queue_select : selectList)
          (s1 s2 : bState),
     predicateIsTrueInState gu1 s1 ->
     substitution_step s1 sub1 s2 ->
     substitution_step s1 (Select (ConsSelectList gu1 sub1 queue_select)) s2
|Substitution_step_select_cons_isNotElem :
   forall (gu1 : predicate)
          (sub1 : substitution)
          (queue_select : selectList)
          (s1 s2 : bState),
     substitution_step s1 (Select queue_select) s2 ->
     substitution_step s1 (Select (ConsSelectList gu1 sub1 queue_select)) s2.

Inductive b_step : bState -> list operation -> operation_label -> bState -> Prop :=
|B_step : forall (pre : predicate)
                 (post : substitution)
                 (operations : list operation)
                 (label : operation_label)
                 (s1 s2 : bState),
            predicateIsTrueInState pre s1 ->
            substitution_step s1 post s2 ->
            In (Operation label pre post) operations ->
            b_step s1 operations label s2.

Theorem isValOfVarInState_preservedByInclusion :
  forall (st2 : bState)
         (val var : string)
         (st1 : bState),
    isValOfVarInState val var st1 ->
    bStateIsIncludedInBState st1 st2 ->
    isValOfVarInState val var st2.
Proof.
  apply (bState_mut
           (fun st2 =>
              forall (val var : string)
                     (st1 : bState),
                isValOfVarInState val var st1 ->
                bStateIsIncludedInBState st1 st2 ->
                isValOfVarInState val var st2)
           (fun bStList =>
              forall (val var : string)
                     (st1 : bState),
                isValOfVarInState val var st1 ->
                bStateIsIncludedInBStateList st1 bStList ->
                isValOfVarInListOfBStates val var bStList)).
  -
    intros.
    inversion H0;subst;inversion H.
  -
    intros.
    inversion H1;subst.
    +
      inversion H0.
    +
      assumption.
    +
      apply IsValOfVarInState_isNotElem.
      apply (H val var st1 H0 H4).
  -
    intros.
    inversion H0.
  -
    intros.
    inversion H2;subst.
    +
      apply IsValOfVarInListOfBStates_isElem.
      apply (H val var st1 H1 H5).
    +
      apply IsValOfVarInListOfBStates_isNotElem.
      apply (H0 val var st1 H1 H5).
Qed.

Theorem predicateIsTrue_PreservedByInclusion :
  forall (st1 st2 : bState)
         (p : predicate),
    predicateIsTrueInState p st1 ->
    bStateIsIncludedInBState st1 st2 ->
    predicateIsTrueInState p st2.
Proof.
  intros.
  induction p.
  -
    inversion H;subst.
    apply PredicateIsTrueInState_equality.
    apply (isValOfVarInState_preservedByInclusion st2 s0 s st1 H4 H0).
  -
    inversion H.
    apply PredicateIsTrueInState_and.
    +
      apply (IHp1 H3).
    +
      apply (IHp2 H5).
  -
    inversion H.
    +
      apply PredicateIsTrueInState_orLeft.
      apply (IHp1 H4).
    +
      apply PredicateIsTrueInState_orRight.
      apply (IHp2 H4).
  -
    apply PredicateIsTrueInState_truePredicate.
  -
    inversion H.
  -
    inversion H;subst.
    apply PredicateIsTrueInState_isIn with val.
    +
      apply (isValOfVarInState_preservedByInclusion st2 val s st1 H3 H0).
    +
      assumption.
  -
    inversion H;subst.
    apply PredicateIsTrueInState_guard.
    assumption.
Qed.

Lemma predicateIsTrueInBiggerState :
  forall (v : bState)
         (pre : predicate)
         (var val : string)
         (vLi : bStateList),
    predicateIsTrueInState pre v ->
    predicateIsTrueInState pre (Automata_B_state var val (ConsBStateList v vLi)).
Proof.
  induction pre.
  -
    intros var val vLi PreTrue.
    inversion PreTrue;subst.
    apply (PredicateIsTrueInState_equality).
    apply IsValOfVarInState_isNotElem.
    apply (IsValOfVarInListOfBStates_isElem).
    assumption.
  -
    intros var val vLi PredTrue.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_and).
    +
      apply (IHpre1 var val vLi H1).
    +
      apply (IHpre2 var val vLi H3).
  -
    intros var val vLi PredTrue.
    inversion PredTrue;subst.
    +
      apply (PredicateIsTrueInState_orLeft).
      apply (IHpre1 var val vLi H2).
    +
      apply (PredicateIsTrueInState_orRight).
      apply (IHpre2 var val vLi H2).
  -
    intros var val vLi PredTrue.
    apply (PredicateIsTrueInState_truePredicate).
  -
    intros var val vLi PredTrue.
    inversion PredTrue.
  -
    intros var val vLi PredTrue.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_isIn) with val0.
    +
      apply (IsValOfVarInState_isNotElem).
      apply (IsValOfVarInListOfBStates_isElem).
      assumption.
    +
      assumption.
  -
    intros var val vLi PredTrue.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_guard).
    assumption.
Qed.                              

Lemma substitutionStepInABiggerState :
  forall (sub : substitution)
         (v v' : bState)
         (var val : string)
         (vLi : bStateList),
    substitution_step v sub v' ->
    substitution_step (Automata_B_state var val (ConsBStateList v vLi))
                      sub
                      (Automata_B_state var val (ConsBStateList v' vLi)).
Proof.
  apply (sub_mut (fun (sub : substitution) =>
                    forall (v v' : bState)
                           (var val : string)
                           (vLi : bStateList),
                      substitution_step v sub v' ->
                      substitution_step (Automata_B_state var val (ConsBStateList v vLi))
                                        sub
                                        (Automata_B_state var val (ConsBStateList v' vLi)))
                 (fun (s : selectList) =>
                    forall (v v' : bState)
                           (var val : string)
                           (vLi : bStateList),
                      substitution_step v (Select s) v' ->
                      substitution_step (Automata_B_state var val (ConsBStateList v vLi))
                                        (Select s)
                                        (Automata_B_state var val (ConsBStateList v' vLi)))).
  -
    intros s IHs v v' var val vLi Step.
    apply (IHs v v' var val vLi Step).
  -
    intros sub1 IHsub1 sub2 IHsub2 v v' var val vLi Step.
    inversion Step;subst.
    apply Substitution_step_parallel with (Automata_B_state var val (ConsBStateList s2 vLi)).
    +
      apply (IHsub1 v s2 var val vLi H2).
    +
      apply (IHsub2 s2 v' var val vLi H4).
  -
    intros var1 var2 v v' var val vLi Step.
    inversion Step;subst.
    apply (Substitution_step_affectation).
    apply (IsStateWhereVarChangedForVal_isNotElem).
    apply (IsStateListWhereVarChangedForVal_isElem).
    assumption.
  -
    intros v v' var val vLi Step.
    inversion Step;subst.
    apply (Substitution_step_skipSubstitution).
  -
    intros v v' var val vLi Step.
    inversion Step.
  -
    intros pre sub IHsub sLi IHsLi v v' var val vLi Step.
    inversion Step;subst.
    +
      apply (Substitution_step_select_cons_isElem).
      *
        apply predicateIsTrueInBiggerState.
        assumption.
      *
        apply (IHsub v v' var val vLi H5).
    +
      apply Substitution_step_select_cons_isNotElem.
      apply (IHsLi v v' var val vLi H4).
Qed.

Lemma substitutionDoesNotChangeStateStructure :
  forall (sub : substitution)
         (var val : string)
         (vLi : bStateList)
         (v' : bState),
    substitution_step (Automata_B_state var val vLi) sub v' ->
    exists (val2 : string)
           (vLi' : bStateList),
      v' = (Automata_B_state var val2 vLi').
Proof.
  apply (sub_mut (fun (sub : substitution) =>
                    forall (var val : string)
                           (vLi : bStateList)
                           (v' : bState),
                      substitution_step (Automata_B_state var val vLi) sub v' ->
                      exists (val2 : string)
                             (vLi' : bStateList),
                        v' = (Automata_B_state var val2 vLi'))
                 (fun (s : selectList) =>
                    forall (var val : string)
                           (vLi : bStateList)
                           (v' : bState),
                      substitution_step (Automata_B_state var val vLi) (Select s) v' ->
                      exists (val2 : string)
                             (vLi' : bStateList),
                          (v' = (Automata_B_state var val2 vLi')))).
  -
    intros s IHs var val vLi v' Step.
    destruct (IHs var val vLi v' Step).
    destruct H.
    exists x.
    exists x0.
    assumption.
  -
    intros sub1 IHsub1 sub2 IHsub2 var val vLi v' Step.
    inversion Step;subst.
    destruct (IHsub1 var val vLi s2 H2).
    destruct H.
    subst.
    destruct (IHsub2 var x x0 v' H4).
    destruct H.
    exists x1.
    exists x2.
    assumption.
  -
    intros s s0 var val vLi v' Step.
    inversion Step;subst.
    inversion H3;subst.
    +
      exists s0.
      exists vLi.
      reflexivity.
    +
      exists val.
      exists subStatesChanged.
      reflexivity.
  -
    intros var val vLi v' Step.
    inversion Step;subst.
    exists val.
    exists vLi.
    reflexivity.
  -
    intros var val vLi v' Step.
    inversion Step.
  -
    intros p s IHs sLi IHsLi var val vLi v' Step.
    inversion Step;subst.
    +
      apply (IHs var val vLi v' H5).
    +
      apply (IHsLi var val vLi v' H4).
Qed.

Lemma substitutionInElemBState :
  forall (sub : substitution)
         (v : bState),
    substitution_step Elem_B_state sub v ->
    v = Elem_B_state.
Proof.
  apply (sub_mut (fun (sub : substitution) =>
                    forall v : bState,
                      substitution_step Elem_B_state sub v ->
                      v = Elem_B_state)
                 (fun (s : selectList) =>
                    forall (v : bState),
                      substitution_step Elem_B_state (Select s) v ->
                      v = Elem_B_state)).
  -
    intros s IHs v Step.
    apply (IHs v Step).
  -
    intros sub1 IHsub1 sub2 IHsub2 v Step.
    inversion Step;subst.
    rewrite (IHsub1 s2 H2) in H4.
    apply (IHsub2 v H4).
  -
    intros var val v Step.
    inversion Step;subst.
    inversion H3.
  -
    intros v Step.
    inversion Step.
    reflexivity.
  -
    intros v Step.
    inversion Step.
  -
    intros pre post IHpost s IHs v Step.
    inversion Step;subst;[apply (IHpost v H5)|apply (IHs v H4)].
Qed.

Lemma predicateIsTrueInBiggerState2 :
  forall (pred : predicate)
         (v : bState)
         (var val : string)
         (vLi : bStateList),
    predicateIsTrueInState pred (Automata_B_state var val vLi) ->
    predicateIsTrueInState pred (Automata_B_state var val (ConsBStateList v vLi)).
Proof.
  induction pred.
  -
    intros v var val vLi PredTrue. apply PredicateIsTrueInState_equality.
    inversion PredTrue;subst. inversion H2;subst.
    +
      apply (IsValOfVarInState_isElem).
    +
      apply (IsValOfVarInState_isNotElem). apply IsValOfVarInListOfBStates_isNotElem. assumption.
  -
    intros v var val vLi PredTrue.
    inversion PredTrue;subst.
    apply (PredicateIsTrueInState_and).
    +
      apply (IHpred1 v var val vLi H1).
    +
      apply (IHpred2 v var val vLi H3).
  -
    intros v var val vLi PredTrue.
    inversion PredTrue;subst.
    +
      apply PredicateIsTrueInState_orLeft.
      apply (IHpred1 v var val vLi H2).
    +
      apply PredicateIsTrueInState_orRight.
      apply (IHpred2 v var val vLi H2).
  -
    intros v var val vLi PredTrue.
    apply PredicateIsTrueInState_truePredicate.
  -
    intros v var val vLi PredTrue.
    inversion PredTrue.
  -
    intros v var val vLi PredTrue.
    inversion PredTrue;subst.
    apply PredicateIsTrueInState_isIn with val0.
    inversion H1;subst.
    +
      apply (IsValOfVarInState_isElem).
    +
      apply IsValOfVarInState_isNotElem.
      apply IsValOfVarInListOfBStates_isNotElem.
      assumption.
    +
      assumption.      
  -
    intros v var val vLi PredTrue.
    inversion PredTrue.
    apply PredicateIsTrueInState_guard.
    assumption.
Qed.

Lemma substitutionStepInABiggerState2 :
  forall (sub : substitution)
         (v : bState)
         (var val val2 : string)
         (vLi vLi' : bStateList),
    substitution_step (Automata_B_state var val vLi) sub (Automata_B_state var val2 vLi')->
    substitution_step (Automata_B_state var val (ConsBStateList v vLi))
                      sub
                      (Automata_B_state var val2 (ConsBStateList v vLi')).
Proof.
  apply (sub_mut (fun (sub : substitution) =>
                    forall (v : bState)
                           (var val val2 : string)
                           (vLi vLi' : bStateList),
                      substitution_step (Automata_B_state var val vLi) sub
                                        (Automata_B_state var val2 vLi') ->
                      substitution_step (Automata_B_state var val (ConsBStateList v vLi))
                                        sub
                                        (Automata_B_state var val2 (ConsBStateList v vLi')))
                 (fun (s : selectList) =>
                    forall (v : bState)
                           (var val val2 : string)
                           (vLi vLi' : bStateList),
                      substitution_step (Automata_B_state var val vLi) (Select s)
                                        (Automata_B_state var val2 vLi') ->
                      substitution_step (Automata_B_state var val (ConsBStateList v vLi))
                                        (Select s)
                                        (Automata_B_state var val2 (ConsBStateList v vLi')))).
  -
    intros s IHs v var val1 val2 vLi vLi' Step.
    apply (IHs v var val1 val2 vLi vLi' Step).
  -
    intros sub1 IHsub1 sub2 IHsub2 v var val1 val2 vLi vLi' Step.
    inversion Step;subst.
    destruct (substitutionDoesNotChangeStateStructure sub1 var val1 vLi s2 H2).
    destruct H.
    subst.
    apply Substitution_step_parallel with (Automata_B_state var x (ConsBStateList v x0)).
    +
      apply (IHsub1 v var val1 x vLi x0 H2).
    +
      apply (IHsub2 v var x val2 x0 vLi' H4).
  -
    intros s s0 v var val1 val2 vLi vLi' Step.
    inversion Step;subst.
    apply (Substitution_step_affectation).
    inversion H3;subst.
    +
      apply (IsStateWhereVarChangedForVal_isElem).
    +
      apply (IsStateWhereVarChangedForVal_isNotElem).
      apply (IsStateListWhereVarChangedForVal_isNotElem).
      assumption.
  -
    intros v var val1 val2 vLi vLi' Step.
    inversion Step;subst.
    apply Substitution_step_skipSubstitution.
  -
    intros v var val1 val2 vLi vLi' Step.
    inversion Step;subst.
  -
    intros pre post IHpost sLi IHsLi v var val1 val2 vLi vLi' Step.
    inversion Step;subst.
    +
      apply Substitution_step_select_cons_isElem.
      apply predicateIsTrueInBiggerState2.
      assumption.
      apply (IHpost v var val1 val2 vLi vLi' H5).
    +
      apply (Substitution_step_select_cons_isNotElem).
      apply (IHsLi v var val1 val2 vLi vLi' H4).
Qed.

Lemma existsAValPreservedByStateChanged :
  forall (v v' : bState)
         (var val var2 val2 : string),
    isValOfVarInState val var v' ->
    isStateWhereVarChangedForVal v' v var2 val2 ->
    exists (val' : string), isValOfVarInState val' var v.
Proof.
  apply (bState_mut (fun (v : bState) =>
                       forall (v' : bState)
                              (var val var2 val2 : string),
                         isValOfVarInState val var v' ->
                         isStateWhereVarChangedForVal v' v var2 val2 ->
                         exists (val' : string), isValOfVarInState val' var v)
                    (fun (vLi : bStateList) =>
                       forall (vLi' : bStateList)
                              (var var2 val val2: string),
                         isValOfVarInListOfBStates val var vLi' ->
                         isStateListWhereVarChangedForVal vLi' vLi var2 val2 ->
                         exists (val' : string),
                           isValOfVarInListOfBStates val' var vLi)).
  -
    intros v' var val var2 val2 IsVal IsStateChanged.
    inversion IsStateChanged;subst.
  -
    intros name value vLi IHvLi v' var val var2 val2 IsVal IsStateChanged.
    inversion IsVal;subst;inversion IsStateChanged;subst.
    +
      exists value.
      apply (IsValOfVarInState_isElem).
    +
      exists value.
      apply (IsValOfVarInState_isElem).
    +
      exists val.
      apply (IsValOfVarInState_isNotElem).
      assumption.
    +
      destruct (IHvLi bStateList var var2 val val2 H H8).
      exists x.
      apply (IsValOfVarInState_isNotElem).
      assumption.
  -
    intros vLi' var var2 val val2 IsVal IsStateChanged.
    inversion IsStateChanged.
  -
    intros v IHv vLi IHvLi vLi' var var2 val val2 IsVal IsStateList.
    inversion IsVal;subst;inversion IsStateList;subst.
    +
      destruct (IHv st var val var2 val2 H H6).                
      exists x.
      apply (IsValOfVarInListOfBStates_isElem).
      assumption.
    +
      exists val.
      apply (IsValOfVarInListOfBStates_isElem).
      assumption.
    +
      exists val.
      apply (IsValOfVarInListOfBStates_isNotElem).
      assumption.
    +
      destruct (IHvLi queueOfSubStates var var2 val val2 H H6).
      exists x.
      apply (IsValOfVarInListOfBStates_isNotElem).
      assumption.
Qed.

Lemma existsAValInListPreservedByStateChanged :
  forall (vLi vLi' : bStateList)
         (var val var2 val2 : string),
    isValOfVarInListOfBStates val var vLi' ->
    isStateListWhereVarChangedForVal vLi' vLi var2 val2 ->
    exists val' : string, isValOfVarInListOfBStates val' var vLi.
Proof.
  induction vLi.
  -
    intros vLi' var val var2 val2 IsValOfVar IsStateChanged.
    inversion IsStateChanged.
  -
    intros vLi' var val var2 val2 IsValOfVar StateChanged.
    inversion IsValOfVar;subst;inversion StateChanged;subst.
    +
      destruct (existsAValPreservedByStateChanged b st var val var2 val2 H H6).
      exists x.
      apply (IsValOfVarInListOfBStates_isElem).
      assumption.
    +
      exists val.
      apply (IsValOfVarInListOfBStates_isElem).
      assumption.
    +
      exists val.
      apply (IsValOfVarInListOfBStates_isNotElem).
      assumption.
    +
      destruct (IHvLi queueOfSubStates var val var2 val2 H H6).
      exists x.
      apply (IsValOfVarInListOfBStates_isNotElem).
      assumption.
Qed.
      
Lemma existsAValPreservedBySubstitution :
  forall (sub : substitution)
         (var val : string)
         (v v' : bState),
    isValOfVarInState val var v' ->
    substitution_step v sub v' ->
    exists (val' : string),
      isValOfVarInState val' var v.
Proof.
  apply (sub_mut (fun (sub : substitution) =>
                    forall (var val : string)
                           (v v' : bState),
                      isValOfVarInState val var v' ->
                      substitution_step v sub v' ->
                      exists (val' : string),
                        isValOfVarInState val' var v)
                 (fun (s : selectList) =>
                    forall (var val : string)
                           (v v' : bState),
                      isValOfVarInState val var v' ->
                      substitution_step v (Select s) v' ->
                      exists (val' : string),
                        isValOfVarInState val' var v)).
  -
    intros s IHs var val v v' IsVal Step.
    apply (IHs var val v v' IsVal Step).
  -
    intros sub1 IHsub1 sub2 IHsub2 var val  v v' IsVal Step.
    inversion Step;subst.
    destruct (IHsub2 var val s2 v' IsVal H4).
    apply (IHsub1 var x v s2 H H2).
  -
    intros s s0 var val v v' IsVal Step.
    inversion Step;subst.
    apply (existsAValPreservedByStateChanged v v' var val s s0 IsVal H3).
  -
    intros var val v v' IsVal Step.
    inversion Step;subst.
    exists val.
    assumption.
  -
    intros var val v v' IsVal Step.
    inversion Step.
  -
    intros pre post IHpost s IHs var val v v' IsVal Step.
    inversion Step;subst.
    +
      apply (IHpost var val v v' IsVal H5).
    +
      apply (IHs var val v v' IsVal H4).
Qed.

Lemma aChangedVariableIsInState :
  forall (v v' : bState)
         (var val : string),
    isStateWhereVarChangedForVal v' v var val ->
    exists (val' : string),
      isValOfVarInState val' var v.
Proof.
  apply (bState_mut (fun (v : bState) =>
                       forall (v' : bState)
                              (var val : string),
                         isStateWhereVarChangedForVal v' v var val ->
                         exists (val' : string),
                           isValOfVarInState val' var v)
                    (fun (vLi : bStateList) =>
                       forall (vLi' : bStateList)
                              (var val : string),
                         isStateListWhereVarChangedForVal vLi' vLi var val ->
                         exists (val' : string),
                           isValOfVarInListOfBStates val' var vLi)).
  -
    intros v' var val StateChanged.
    inversion StateChanged.
  -
    intros s s0 vLi IHvLi v' var val StateChanged.
    inversion StateChanged;subst.
    +
      exists s0.
      apply (IsValOfVarInState_isElem).
    +
      destruct (IHvLi subStatesChanged var val H5).
      exists x.
      apply (IsValOfVarInState_isNotElem).
      assumption.
  -
    intros vLi' var val StateChanged.
    inversion StateChanged.
  -
    intros v IHv vLi IHvLi vLi' var val StateChanged.
    inversion StateChanged;subst.
    +
      destruct (IHv stChanged var val H4).
      exists x.
      apply (IsValOfVarInListOfBStates_isElem).
      assumption.
    +
      destruct (IHvLi subStatesQueueChanged var val H4).
      exists x.
      apply IsValOfVarInListOfBStates_isNotElem.
      assumption.
Qed.

Lemma aModifiedVariableIsInState :
  forall (sub : substitution)
         (var : string)
         (v v' : bState),
    variableIsUsedAndModifiedInSubstitution var sub ->
    substitution_step v sub v' ->
    substitutionContainsNoSelect sub ->
    exists (val : string),
      isValOfVarInState val var v.
Proof.
  induction sub.
  -
    intros var v v' VarMod Step NoSelect.
    inversion NoSelect.
  -
    intros var v v' VarMod Step NoSelect.
    inversion NoSelect;subst.
    inversion Step;subst.
    inversion VarMod;subst.
    +
      apply (IHsub1 var v s2 H3 H4 H1).
    +
      destruct (IHsub2 var s2 v' H3 H6 H2).
      destruct (existsAValPreservedBySubstitution sub1 var x v s2 H H4).
      exists x0.
      assumption.
  -
    intros var v v' VarMod Step NoSelect.
    inversion VarMod;subst.
    inversion Step;subst.
    destruct (aChangedVariableIsInState v v' s s0 H3).
    exists x.
    assumption.
  -
    intros var v v' VarMod.
    inversion VarMod.
Qed.

Lemma isValOfVarInStateListIfIsValOfVarInState :
  forall (v : bState)
         (vLi : bStateList)
         (var val : string),
    isValOfVarInState val var v ->
    isAStateOfBStateList v vLi ->
    isValOfVarInListOfBStates val var vLi.
Proof.
  induction vLi.
  -
    intros var val IsValOfVar IsAState.
    inversion IsAState.
  -
    intros var val IsValOfVar IsAState.
    inversion IsAState;subst.
    +
      apply (IsValOfVarInListOfBStates_isElem).
      assumption.
    +
      apply (IsValOfVarInListOfBStates_isNotElem).
      apply (IHvLi var val IsValOfVar H1).
Qed.

