Require Import B_langage.
Require Import String.

Inductive bState :=
|Elem_B_state : bState
|Automata_B_state : string -> string -> bStateList -> bState
with bStateList :=
|EmptyBStateList : bStateList
|ConsBStateList : bState -> bStateList -> bStateList.

Scheme bState_mut := Induction for bState Sort Prop
                     with bStateList_mut := Induction for bStateList Sort Prop.

Inductive isValOfVarInState : string -> string -> bState -> Prop :=
|IsValOfVarInState_isElem :
   forall (var val : string)
          (queue_states : bStateList),
     isValOfVarInState val var (Automata_B_state var val queue_states)
|IsValOfVarInState_isNotElem :
   forall (var1 var2 val1  val2  : string)
          (bStateList : bStateList),
     isValOfVarInListOfBStates val1 var1 bStateList ->
     isValOfVarInState val1 var1 (Automata_B_state var2 val2 bStateList)
with isValOfVarInListOfBStates : string ->  string -> bStateList -> Prop :=
|IsValOfVarInListOfBStates_isElem :
   forall (st : bState)
          (queueOfSubStates : bStateList)
          (var val : string),
     isValOfVarInState val var st ->
     isValOfVarInListOfBStates val var (ConsBStateList st queueOfSubStates)
|IsValOfVarInListOfBStates_isNotElem :
   forall (st : bState)
          (queueOfSubStates : bStateList)
          (var val : string),
     isValOfVarInListOfBStates val var queueOfSubStates ->
     isValOfVarInListOfBStates val var (ConsBStateList st queueOfSubStates).

(*
IsStateWhereVarChangedForVal can only change a variable that is already in the bState.
It means that the value HAS to be found. In a first version, there was this case :

|IsStateWhereVarChangedForVal_nil :
   forall (var1 val1 : string),
     isStateWhereVarChangedForVal Elem_B_state Elem_B_state var1 val1

But is has been removed for the reason explained below.
*)

Inductive isStateWhereVarChangedForVal : bState -> bState -> string -> string -> Prop :=
|IsStateWhereVarChangedForVal_isElem :
   forall (var val1 val2 : string)
          (subStates : bStateList),
     isStateWhereVarChangedForVal
       (Automata_B_state var val1 subStates)
       (Automata_B_state var val2 subStates)
       var val1
|IsStateWhereVarChangedForVal_isNotElem :
   forall (var1 var2 val1 val2 : string)
          (subStates subStatesChanged : bStateList),
     isStateListWhereVarChangedForVal subStatesChanged subStates var1 val1 ->
     isStateWhereVarChangedForVal
       (Automata_B_state var2 val2 subStatesChanged)
       (Automata_B_state var2 val2 subStates) var1 val1
with isStateListWhereVarChangedForVal : bStateList -> bStateList -> string -> string -> Prop :=
|IsStateListWhereVarChangedForVal_isElem :
   forall (var val : string)
          (stChanged st : bState)
          (subStatesQueue : bStateList),
     isStateWhereVarChangedForVal stChanged st var val ->
     isStateListWhereVarChangedForVal
       (ConsBStateList stChanged subStatesQueue) (ConsBStateList st subStatesQueue) var val
|IsStateListWhereVarChangedForVal_isNotElem :
   forall (var val : string)
          (st : bState)
          (subStatesQueue subStatesQueueChanged : bStateList),
     isStateListWhereVarChangedForVal subStatesQueueChanged subStatesQueue var val ->
     isStateListWhereVarChangedForVal
       (ConsBStateList st subStatesQueueChanged) (ConsBStateList st subStatesQueue) var val.

Inductive bStateIsIncludedInBState : bState -> bState -> Prop :=
|BStateIsIncludedInBState_elem :
   forall (st : bState),
     bStateIsIncludedInBState Elem_B_state st
|BStateIsIncludedInBState_automata_isElem :
   forall (st : bState),
     bStateIsIncludedInBState st st
|BStateIsIncludedInBState_automata_isNotElem :
   forall (var val : string)
          (listOfSubStates : bStateList)
          (st : bState),
     bStateIsIncludedInBStateList st listOfSubStates ->
     bStateIsIncludedInBState st (Automata_B_state var val listOfSubStates)
with bStateIsIncludedInBStateList : bState -> bStateList -> Prop :=
|BStateIsIncludedInBStateList_isElem :
   forall (st searchedSt : bState)
          (bStateList : bStateList),
     bStateIsIncludedInBState searchedSt st ->
     bStateIsIncludedInBStateList searchedSt (ConsBStateList st bStateList)
|BStateIsIncludedInBStateList_isNotElem :
   forall (st searchedSt : bState)
          (bStateList : bStateList),
     bStateIsIncludedInBStateList searchedSt bStateList ->
     bStateIsIncludedInBStateList searchedSt (ConsBStateList st bStateList).

Inductive isAStateOfBStateList : bState -> bStateList -> Prop :=
|IsAStateOfBStateList_isElem :
   forall (v : bState)
          (li : bStateList),
     isAStateOfBStateList v (ConsBStateList v li)
|IsAStateOfBStateList_isNotElem :
   forall (v v' : bState)
          (li : bStateList),
     isAStateOfBStateList v li ->
     isAStateOfBStateList v (ConsBStateList v' li).


Inductive isBStateNamedN : bState -> string -> Prop :=
|IsBStateNamedN_automata_B_state :
   forall (var val : string)
          (vLi : bStateList),
     isBStateNamedN (Automata_B_state var val vLi) var.



(* Inductive variableIsNotUsedInPredicate : string -> predicate -> Prop := *)
(* |VariableIsNotUsedInPredicate_equality_variable : *)
(*    forall (var1 var2 val : string), *)
(*      var1 <> var2 -> *)
(*      variableIsNotUsedInPredicate var1 (Equality_variable var2 val) *)
(* |VariableIsNotUsedInPredicate_isIn : *)
(*    forall (var1 var2 : string) *)
(*           (setOfValues : list string), *)
(*      var1 <> var2 -> *)
(*      variableIsNotUsedInPredicate var1 (IsIn var2 setOfValues) *)
(* |VariableIsUsedInPrediate_and : *)
(*    forall (var : string) *)
(*           (pred1 pred2 : predicate), *)
(*      variableIsNotUsedInPredicate var pred1 -> *)
(*      variableIsNotUsedInPredicate var pred2 -> *)
(*      variableIsNotUsedInPredicate var (And pred1 pred2) *)
(* |VariableIsUsedInPrediate_or : *)
(*    forall (var : string) *)
(*           (pred1 pred2 : predicate), *)
(*      variableIsNotUsedInPredicate var pred1 -> *)
(*      variableIsNotUsedInPredicate var pred2 -> *)
(*      variableIsNotUsedInPredicate var (Or pred1 pred2) *)
(* |VariableIsNotUsedInPrediate_guard : *)
(*    forall (gu var : string), *)
(*      variableIsNotUsedInPredicate var (Guard gu) *)
(* |VariableIsNotUsedInPredicate_truePredicate : *)
(*    forall (var : string), *)
(*      variableIsNotUsedInPredicate var TruePredicate *)
(* |VariableIsNotUsedInPredicate_falsePredicate : *)
(*    forall (var : string), *)
(*      variableIsNotUsedInPredicate var FalsePredicate *)
(* |VariableIsNotUsedInPredicate_equality_constant : *)
(*    forall (var cst1 cst2 : string), *)
(*      variableIsNotUsedInPredicate var (Equality_constant cst1 cst2). *)

Inductive noVariableIsNamedNInBState : string -> bState -> Prop :=
|NoVariableIsNamedNInBState_elem :
   forall (s : string),
     noVariableIsNamedNInBState s Elem_B_state
|NoVariableIsNamedNInBState_automata :
   forall (var val s : string)
          (bSubStateList : bStateList),
     noVariableIsNamedNInBStateList s bSubStateList ->
     var <> s ->
     noVariableIsNamedNInBState s (Automata_B_state var val bSubStateList)
with noVariableIsNamedNInBStateList : string -> bStateList -> Prop :=
|NoVariableIsNamedNInBStateList_nil :
   forall (s : string),
     noVariableIsNamedNInBStateList s EmptyBStateList
|NoVariableIsNamedNInBStateList_notNil :
   forall (s : string)
          (st : bState)
          (bStateListQueue : bStateList),
     noVariableIsNamedNInBStateList s bStateListQueue ->
     noVariableIsNamedNInBState s st ->
     noVariableIsNamedNInBStateList s (ConsBStateList st bStateListQueue).

Inductive variableNamesAreUniqueInBState : bState -> Prop :=
|VariableNamesAreUniqueInBState_elem :
   variableNamesAreUniqueInBState Elem_B_state
|VariableNamesAreUniqueInBState_automata :
   forall (var val : string)
          (bSubStates : bStateList),
     noVariableIsNamedNInBStateList var bSubStates ->
     variableNamesAreUniqueInBStateList bSubStates ->
     variableNamesAreUniqueInBState (Automata_B_state var val bSubStates)
with variableNamesAreUniqueInBStateList : bStateList -> Prop :=
|VariableNamesAreUniqueInBStateList_nil :
   variableNamesAreUniqueInBStateList EmptyBStateList
|VariableNamesAreUniqueInBStateList_notNil :
   forall (st : bState)
          (bStateQueue : bStateList),
     variableNamesAreUniqueInBStateList bStateQueue ->
     (forall (var val : string),
        isValOfVarInState  val var st ->
        noVariableIsNamedNInBStateList var bStateQueue) ->
     (forall (val var : string),
        isValOfVarInListOfBStates val var bStateQueue ->
        noVariableIsNamedNInBState var st) ->
     variableNamesAreUniqueInBState st ->
     variableNamesAreUniqueInBStateList (ConsBStateList st bStateQueue).

Lemma absurd_isStateWhereValChanged_noVariableNamed :
  forall (v v' : bState)
         (var val : string)
         (P : Prop),
    isStateWhereVarChangedForVal v' v var val ->
    noVariableIsNamedNInBState var v ->
    P.
Proof.
  apply (bState_mut (fun (v : bState) =>
                       forall (v' : bState)
                              (var val : string)
                              (P : Prop),
                         isStateWhereVarChangedForVal v' v var val ->
                         noVariableIsNamedNInBState var v ->
                         P)
                    (fun (vLi : bStateList) =>
                       forall (vLi' : bStateList)
                              (var val : string)
                              (P : Prop),
                         noVariableIsNamedNInBStateList var vLi ->
                         isStateListWhereVarChangedForVal vLi' vLi var val ->
                         P)).
  -  
    intros v' var val P VarChanged NoVarNamed.
    inversion VarChanged.
  -
    intros var val vLi IHvLi.
    intros v' var0 val0 P VarChanged NoVarNamed.
    inversion VarChanged;subst.
    +
      inversion NoVarNamed;subst.
      destruct H4.
      reflexivity.
    +
      inversion NoVarNamed;subst.
      apply (IHvLi subStatesChanged var0 val0 P H2 H5).
  -
    intros vLi' var val P NoVarNamed VarChanged.
    inversion VarChanged.
  -
    intros v IHv vLi IHvLi vLi' var val P NoVarNamed VarChanged.
    inversion NoVarNamed.
    inversion VarChanged;subst.
    +
      apply (IHv stChanged var val P H9 H3).
    +
      apply (IHvLi subStatesQueueChanged var val P H2 H9).
Qed.

Lemma absurd_isStateListWhereValChanged_noVariableNamed :
  forall (vLi vLi' : bStateList)
         (var val : string)
         (P : Prop),
    noVariableIsNamedNInBStateList var vLi ->
    isStateListWhereVarChangedForVal vLi' vLi var val ->
    P.
Proof.
  induction vLi.
  -
    intros vLi' var val P NoVarNamed IsStateList.
    inversion IsStateList.
  -
    intros vLi' var val P NoVarNamed IsStateList.
    inversion NoVarNamed;subst.
    inversion IsStateList;subst.
    +
      apply (absurd_isStateWhereValChanged_noVariableNamed b stChanged var val);assumption.
    +
      apply (IHvLi subStatesQueueChanged var val);assumption.
Qed.
