Require Import translation.
Require Import B_langage.
Require Import ASTD.
Require Import B_semantics.
Require Import B_state.
Require Import String.
Require Import Theorems.
Require Import mergeWorks.
Require Import wellConstructed.

Lemma labelsAreEqual_implies_isLabelOfTransition :
  forall (t : transition)
         (nameOfOp : string)
         (args : list (string * string)),
    labelsAreEqual (Label nameOfOp args) (getTransitionLabel t) = true ->
    isLabelOfTransition (Transition_label nameOfOp args) t.
Proof.
  intros.
  destruct t;destruct t;simpl in H; unfold labelsAreEqual in H.
  destruct (string_dec nameOfOp s1);subst;[ |discriminate].
  rewrite (paramAreEqualsWorks args l H).
  apply (IsLabelOfTransition_local).
  destruct (string_dec nameOfOp s2);subst;[ |discriminate].
  rewrite (paramAreEqualsWorks args l H).
  apply (IsLabelOfTransition_fromSub).
  destruct (string_dec nameOfOp s2);subst;[ |discriminate].
  rewrite (paramAreEqualsWorks args l H).
  apply (IsLabelOfTransition_toSub).
Qed.

Theorem addTransitionWorks :
  forall (t : transition)
         (l opes : list operation)
         (v v' : bState)
         (args : list (string * string))
         (name nameOfOp: string)
         (states : listOfAstdStates),
    addTransition t l name states = Some opes ->
    operationsAreWellConstructed l ->
    b_step v opes (Label nameOfOp args) v'->
    exists (pre : predicate)
           (post : substitution),
      (makePrecondition t name states = Some pre /\
       makePostcondition t name states = Some post /\
       ((predicateIsTrueInState pre v /\
         substitution_step v post v' /\
         isLabelOfTransition (Transition_label nameOfOp args) t)\/
        b_step v l (Label nameOfOp args) v')).
Proof.
  induction l.
  -
    intros opes v v' args name nameOfOp states.
    intros addTrans WellCons step.
    simpl in addTrans.
    case_eq (makePrecondition t name states);intros;rewrite H in addTrans;[|discriminate].
    case_eq (makePostcondition t name states);intros;rewrite H0 in addTrans;[|discriminate].
    inversion addTrans.
    exists p.
    exists s.
    split;[|split];[reflexivity|reflexivity|].
    rewrite <- H2 in step.
    inversion step;subst.
    simpl in H4.
    destruct H4;inversion H2;subst.
    left.
    split;[|split].
    +
      assumption.
    +
      inversion H3;subst;[|inversion H10].
      assumption.
    +
      destruct t;simpl in H5;destruct t;simpl in H5;inversion H5;subst.
      *
        apply IsLabelOfTransition_local.
      *
        apply IsLabelOfTransition_fromSub.
      *
        apply IsLabelOfTransition_toSub.
  -
    intros opes v v' args name nameOfOp states.
    intros addTrans WellCons step.
    simpl in addTrans.
    case_eq (makePrecondition t name states);intros;rewrite H in addTrans;[|discriminate].
    case_eq (makePostcondition t name states);intros;rewrite H0 in addTrans;[|discriminate].
    exists p.
    exists s.
    split;[|split];[reflexivity|reflexivity|].
    destruct a.
    case_eq (labelsAreEqual o (getTransitionLabel t));intros;rewrite H1 in addTrans.
    +
      destruct s0;[|discriminate|discriminate |discriminate ].
      inversion addTrans.
      rewrite <- H3 in step.
      inversion step;subst.
      simpl in H5.
      destruct H5.
      *
        inversion H3;subst.
        inversion H4;subst.
        left.
        split;[assumption|split;[assumption|]].
        apply (labelsAreEqual_implies_isLabelOfTransition t nameOfOp args H1).
        right.
        apply B_step with (pre := p0) (post := (Select s0)).
        inversion WellCons;subst.
        apply (wellConstructedGuardIsTrue s0 p0 v v' H11 H10).
        assumption.
        simpl.
        left.
        reflexivity.
      *
        right.
        apply (B_step pre post).
        assumption.
        assumption.
        simpl.
        right.
        assumption.
    +
      case_eq (addTransition t l name states);intros;rewrite H2 in addTrans;[|discriminate].
      inversion addTrans;subst.
      inversion step;subst.
      simpl in H5.
      inversion H5;subst.
      inversion H6;subst.
      *
        right.
        apply (B_step pre post).
        assumption.
        assumption.
        simpl.
        left.
        reflexivity.
      *
        inversion WellCons;subst.
        destruct (IHl l0 v v' args name nameOfOp states H2 H9).
        apply (B_step pre post).
        assumption.
        assumption.
        assumption.
        destruct H7.
        destruct H7.
        destruct H8.
        destruct H10.
        rewrite H in H7;inversion H7;subst.
        rewrite H0 in H8;inversion H8;subst.
        destruct H10.
        left.
        split;assumption.
        inversion H10;subst.
        right.
        apply (B_step pre0 post0).
        assumption.
        assumption.
        simpl.
        right.
        assumption.
Qed.

