Require Import translation.
Require Import ASTD.
Require Import String.
Require Import B_state.
Require Import states.
Require Import stateEquivalence.
Require Import ASTDstep.
Require Import B_semantics.
Require Import B_langage.
Require Import theoremsOnUniqueBState.
Require Import theoremsOnUniqueAstdList.
Require Import theoremsOnSubstitutions.
Require Import Theorems.

Lemma initAllVarContainsNoSelect :
  forall (a : astd),
    substitutionContainsNoSelect (initializeAllVariable a).
Proof.
  apply (astd_mut (fun (a : astd) =>
                     substitutionContainsNoSelect (initializeAllVariable a))
                  (fun (states : listOfAstdStates) =>
                     substitutionContainsNoSelect (initializeFromAstdList states))).
  -
    intros s.
    simpl.
    apply (SubstitutionContainsNoSelect_skipSubstitution).
  -
    intros name states IHstates trans sF dF init.
    simpl.
    apply (SubstitutionContainsNoSelect_parallel).
    +
      apply (SubstitutionContainsNoSelect_affectation).
    +
      assumption.
  -
    simpl.
    apply (SubstitutionContainsNoSelect_skipSubstitution).
  -
    intros a IHa states IHstate.
    simpl.
    apply (SubstitutionContainsNoSelect_parallel);assumption.
Qed.

Corollary initAllVarListContainsNoSelect :
  forall (states : listOfAstdStates),
    substitutionContainsNoSelect (initializeFromAstdList states).
Proof.
  induction states.
  -
    simpl. apply SubstitutionContainsNoSelect_skipSubstitution.
  -
    simpl. apply (SubstitutionContainsNoSelect_parallel);[apply initAllVarContainsNoSelect|assumption].
Qed.

Lemma initAllVarDoesNotChangeValueIfNoAstdNamed :
  forall (a : astd)
         (var val1 val2 : string)
         (vLi vLi' : bStateList),
    noAstdIsNamedNInAstd var a ->
    substitution_step (Automata_B_state var val1 vLi)
                      (initializeAllVariable a)
                      (Automata_B_state var val2 vLi') ->
    val1 = val2.
Proof.
  apply (astd_mut (fun (a : astd) =>
                     forall (var val1 val2 : string)
                            (vLi vLi' : bStateList),
                       noAstdIsNamedNInAstd var a ->
                       substitution_step (Automata_B_state var val1 vLi)
                                         (initializeAllVariable a)
                                         (Automata_B_state var val2 vLi') ->
                       val1 = val2)
                  (fun (states : listOfAstdStates) =>
                     forall (var val1 val2 : string)
                            (vLi vLi' : bStateList),
                       noAstdIsNamedNInSubAstdList var states ->
                       substitution_step (Automata_B_state var val1 vLi)
                                         (initializeFromAstdList states)
                                         (Automata_B_state var val2 vLi') ->
                       val1 = val2)).
  -
    intros name var val1 val2 vLi vLi' NoAstdNamed Step.
    simpl in Step.
    inversion Step.
    reflexivity.
  -
    intros name states IHstates trans sF dF init var val1 val2 vLi vLi'.
    intros NoAstdNamed Step.
    simpl in Step.
    inversion Step;subst.
    inversion NoAstdNamed;subst.
    inversion H2;subst.
    inversion H6;subst.
    absurd (name = name);[assumption | reflexivity].
    apply (IHstates var val1 val2 subStatesChanged vLi' H3 H4).
  -
    intros val val1 val2 vLi vLi' NoAstdNamed Step.
    inversion Step;subst.
    reflexivity.
  -
    intros a IHa states IHstates var val1 val2 vLi vLi' NoAstdNamed Step.
    simpl in Step.
    inversion Step;subst.
    destruct (substitutionDoesNotChangeStateStructure (initializeAllVariable a) var val1 vLi s2 H2).
    destruct H.
    subst.
    inversion NoAstdNamed;subst.
    rewrite <- (IHa var val1 x vLi x0 H3 H2) in H4.
    apply (IHstates var val1 val2 x0 vLi' H5 H4).
Qed.

Lemma substitutionDoesNotChangeVariable :
  forall (sub : substitution)
         (var val val' : string)
         (vLi vLi' : bStateList),
    substitution_step (Automata_B_state var val vLi) sub (Automata_B_state var val' vLi') ->
    substitutionContainsNoSelect sub ->
    (forall (name : string),
       variableIsUsedAndModifiedInSubstitution name sub ->
       name <> var) ->
    val = val'.
Proof.
  induction sub.
  -
    intros var val val' vLi vLi' Step NoSelect.
    inversion NoSelect.
  -
    intros var val val' vLi vLi' Step NoSelect NotVar.
    inversion Step;subst.
    destruct (substitutionDoesNotChangeStateStructure sub1 var val vLi s2 H2).
    destruct H.
    subst.
    inversion NoSelect;subst.
    rewrite (IHsub1 var val x vLi x0 H2 H1).
    apply (IHsub2 var x val' x0 vLi' H4 H3).
    intros name VarMod.
    apply (NotVar name).
    apply (VariableIsUsedAndModifiedInSubstitution_parallel_right). assumption.
    intros name VarMod.
    apply (NotVar name).
    apply (VariableIsUsedAndModifiedInSubstitution_parallel_left). assumption.
  -
    intros var val val' vLi vLi' Step NoSelect  NotVar.
    inversion Step;subst.
    inversion H3;subst.
    +
      destruct (NotVar s).
      apply (VariableIsUsedAndModifiedInSubstitution_affectation).
      reflexivity.
    +
      reflexivity.
  -
    intros var val val' vLi vLi' Step NoSelect NotVar.
    inversion Step;subst.
    reflexivity.
Qed.

Lemma absurd_noAstdNamedAndVarModified :
  forall (a : astd)
         (var : string)
         (P : Prop),
    noAstdIsNamedNInAstd var a ->
    variableIsUsedAndModifiedInSubstitution var (initializeAllVariable a) ->
    P.
Proof.
  apply (astd_mut (fun (a : astd) =>
                     forall (var : string)
                            (P : Prop),
                       noAstdIsNamedNInAstd var a ->
                       variableIsUsedAndModifiedInSubstitution var (initializeAllVariable a) ->
                       P)
                  (fun (states : listOfAstdStates) =>
                     forall (var : string)
                            (P : Prop),
                       noAstdIsNamedNInSubAstdList var states ->
                       variableIsUsedAndModifiedInSubstitution var (initializeFromAstdList states) ->
                       P)).
  -
    intros name var P NoAstdNamed VarMod.
    inversion VarMod.
  -
    intros name states IHstates trans sF dF init var P NoAstdNamed VarUsed.
    inversion NoAstdNamed;subst.
    inversion VarUsed;subst.
    +
      inversion H1;subst. destruct H7. reflexivity.
    +
      apply (IHstates var P H2 H1).
  -
    intros var P NoAstdNamed VarUsed.
    inversion VarUsed.
  -
    intros a IHa states IHstates var P NoAstdNamed VarUsed.
    inversion NoAstdNamed;subst.
    inversion VarUsed;subst.
    +
      apply (IHa var P H2 H1).
    +
      apply (IHstates var P H3 H1).
Qed.

Corollary absurd_noAstdNamedAndVarModified_inList :
  forall (states : listOfAstdStates)
         (var : string)
         (P : Prop),
    noAstdIsNamedNInSubAstdList var states ->
    variableIsUsedAndModifiedInSubstitution var (initializeFromAstdList states) ->
    P.
Proof.
  intros states var P NoAstdNamed VarUsed.
  induction states;inversion VarUsed;inversion NoAstdNamed;subst;[apply (absurd_noAstdNamedAndVarModified a var);assumption|apply (IHstates);assumption].
Qed.

Theorem initAllVarWorks :
  forall (a : astd)
         (v v' : bState)
         (s : state)
         (name : string),
    isAStateOfTheAstd s a ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    isAstdNamedN a name ->
    isWellDefined a ->
    variableNamesAreUniqueInBState v ->
    substitution_step v (initializeAllVariable a) v' ->
    exists (s' : state),
      bStateAndAstdStateAreEquivalentInAstdNamed v' s' name /\
      isInitialStateOfAstd s' a.
Proof.
  apply (astd_mut (fun (a : astd) =>
                     forall (v v' : bState)
                            (s : state)
                            (name : string),
                       isAStateOfTheAstd s a ->
                       bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
                       isAstdNamedN a name ->
                       isWellDefined a ->
                       variableNamesAreUniqueInBState v ->
                       substitution_step v (initializeAllVariable a) v' ->
                       exists (s' : state),
                         bStateAndAstdStateAreEquivalentInAstdNamed v' s' name /\
                         isInitialStateOfAstd s' a)
                  (fun (states : listOfAstdStates) =>
                     forall (vLi vLi' : bStateList)
                            (var val : string)
                            (h : histState),
                       isAHistoryStateOfAstdList h states ->
                       listOfBStatesIsEquivalentToHistory vLi h ->
                       astdNameAreUniqueInAstdList states ->
                       noAstdIsNamedNInSubAstdList var states ->
                       isAstdListWellDefined states ->
                       variableNamesAreUniqueInBState (Automata_B_state var val vLi) ->
                       substitution_step (Automata_B_state var val vLi)
                                         (initializeFromAstdList states)
                                         (Automata_B_state var val vLi') ->
                       exists (h' : histState),
                         listOfBStatesIsEquivalentToHistory vLi' h' /\
                         isHInit h' states)).
  -
    intros name v v' s name' IsAState StateEquiv IsAstdNamed WellDef VarNameUnique Step.
    inversion Step;subst.
    inversion IsAState;subst.
    exists Elem_state.
    split;[assumption|apply (IsInitialStateOfAstd_elem)].
  -
    intros name states IHstates trans sF dF init v v' s name' IsAState StateEquiv IsAstdNamed WellDef BNamesUnique Step.
    inversion IsAstdNamed;subst.
    inversion IsAState;subst. inversion StateEquiv;subst. inversion WellDef;subst.
    inversion H16;subst.
    simpl in Step. inversion Step;subst.
    assert (s2 = (Automata_B_state name' init listBSubStateRes)). apply (affectationInUniqueState name' nameOfState init listBSubStateRes s2 H5 BNamesUnique). subst.
    destruct (substitutionDoesNotChangeStateStructure (initializeFromAstdList states) name' init listBSubStateRes v' H18). destruct H. subst.
    assert (init = x).
    apply (substitutionDoesNotChangeVariable (initializeFromAstdList states) name' init x listBSubStateRes x0 H18).
    apply initAllVarListContainsNoSelect.
    intros name VarUsed.
    destruct (string_dec name name');subst;[apply (absurd_noAstdNamedAndVarModified_inList states name');assumption|assumption].
    subst.
    destruct (existsHistStateEquivToBStateList states listBSubState listBSubStateRes h nameOfState subState bSubState subAstd H3 H9 H19 H7 H8 H4 H6).
    destruct H.
    destruct (IHstates listBSubStateRes x0 name' x x1 H0 H H19 H1 H12).
    inversion BNamesUnique;subst.
    apply (VariableNamesAreUniqueInBState_automata);assumption.
    assumption.
    destruct H2. destruct H15.
    destruct x3.
    +
      exists (Automata_state x x2 Elem_state).
      split.
      *
        apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem) with (bSubState := Elem_B_state) (listBSubState := x0).
        assumption.
        apply (BStateAndAstdStateAreEquivalentInAstdNamed_elem).
        apply (replacingByElemBState_reverse).
      *
        apply (IsInitialStateOfAstd_automata) with (Elem s).
        apply (IsInitialStateOfAstd_elem).
        assumption.
        assumption.
    +
      destruct (existsAStateInHistoryIfExistsAnAstd states x x2 (Automata s l l0 l1 l2 s0) H15). apply (IsAutomata). apply (HInitIsAnHistoricState states x2 H10).
      exists (Automata_state x x2 x3). split.
      *
        destruct (existsABStateInBStateListEquivToAStateInHistory x2 x x3 x0 H2 H20).
        destruct H21.
        apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem) with (bSubState := x4) (listBSubState := x0).
        assumption.
        assumption.
        apply (historicStateInHInitIsInitial states x2 x (Automata s l l0 l1 l2 s0)) in H20.
        inversion H20;subst. inversion H21;subst.
        apply (replacingByAStateInList).
        assumption.
        assumption.
        assumption.
        apply (IsAutomata).
        assumption.
      *
        apply (IsInitialStateOfAstd_automata) with (Automata s l l0 l1 l2 s0).
        apply (historicStateInHInitIsInitial states x2 x (Automata s l l0 l1 l2 s0) x3 H20 H15 H10). apply (IsAutomata).
        assumption.
        assumption.
        assumption.
  -
    intros vLi vLi' var val h IsHist StateEquiv AstdNameUnique NoAstdNamed WellDef BNamesUnique Step.
    simpl in Step.
    inversion Step;subst. inversion IsHist;subst. inversion StateEquiv;subst.
    exists EmptyHistState.
    split;[apply (ListOfBStatesIsEquivalentToHistory_nil)|apply (IsHInit_nil)].
  -
    intros a IHa states IHstates vLi vLi' var val h IsHist StateEquiv AstdNameUnique NoAstdNamed WellDef BNamesUnique Step.
    simpl in Step.
    inversion Step;subst.
    destruct a.
    +
      simpl in H2.
      inversion H2;subst.
      inversion IsHist;subst.
      inversion NoAstdNamed;subst.
      inversion AstdNameUnique;subst.
      inversion H3;subst.
      inversion WellDef;subst.
      destruct (IHstates vLi vLi' var val h H1 StateEquiv H7 H6 H9 BNamesUnique H4).
      exists x. destruct H.
      split.
      assumption.
      apply (IsHInit_elem).
      assumption.
    +
      inversion IsHist;subst. inversion StateEquiv;subst.
      destruct (substitutionDoesNotChangeStateStructure (initializeAllVariable (Automata s l l0 l1 l2 s0)) var val (ConsBStateList bst bState_queue) s2 H2).
      destruct H. subst.
      destruct (substitutionPreserveSubStateStructure1 (initializeAllVariable (Automata s l l0 l1 l2 s0)) var val x bst bState_queue x0 H2).
      *
        apply (initAllVarContainsNoSelect).
      *
        intros name VarMod.
        inversion AstdNameUnique;subst.
        split.
        destruct (H8 name).
        apply (absurd_noAstdNamedAndVarModified (Automata s l l0 l1 l2 s0) name);assumption.
        apply (noAstdNamedInAstdList_implies_noVarNamedInList states hRest);assumption.
        inversion BNamesUnique;subst.
        inversion H12;subst.
        unfold not. intros. subst.
        inversion NoAstdNamed;subst. apply (absurd_noAstdNamedAndVarModified (Automata s l l0 l1 l2 s0) name);assumption.
      *
        destruct H. destruct H0. subst.
        {
          inversion WellDef;subst.
          destruct (IHa bst x1 hState s H10 H7). apply (IsAstdNamedN_automata).
          assumption. inversion BNamesUnique;subst. inversion H12;subst.
          assumption. assumption.
          destruct H.
          destruct (substitutionPreserveSubStateStructure2 (initializeFromAstdList states) var x x x1 bState_queue vLi' H4).
          -
            apply initAllVarListContainsNoSelect.
          -
            intros name VarMod.
            inversion AstdNameUnique;subst.
            split.
            +
              destruct (H14 name).
              apply (noAstdNamed_implies_noVarNamed x0 (Automata s l l0 l1 l2 s0) x1 name H9).
              apply (initStateIsAState);assumption.
              simpl.
              assumption.
              apply (absurd_noAstdNamedAndVarModified_inList states name);assumption.
            +
              unfold not. intros. subst.
              inversion NoAstdNamed;subst. apply (absurd_noAstdNamedAndVarModified_inList states name);assumption.
          -
            inversion AstdNameUnique;subst.
            destruct H9. subst.
            inversion NoAstdNamed;subst.
            destruct (IHstates bState_queue x2 var x hRest H3 H5 H14 H18 H6).
            inversion BNamesUnique;subst.
            inversion H16;subst.
            inversion H20;subst.
            apply VariableNamesAreUniqueInBState_automata;assumption.
            apply (substitutionInASmallerState) with x1.
            assumption. apply initAllVarListContainsNoSelect.
            intros name VarMod.
            split.
            +
              destruct (H15 name).
              apply (noAstdNamed_implies_noVarNamed x0 (Automata s l l0 l1 l2 s0) x1 name H9).
              apply (initStateIsAState);assumption.
              simpl. assumption.
              apply (absurd_noAstdNamedAndVarModified_inList states name);assumption.
            +
              unfold not. intros. subst.
              apply (absurd_noAstdNamedAndVarModified_inList states name);assumption.
            +
              destruct H9.
              exists (ConsHistState s x0 x3).
              split.
              apply (ListOfBStatesIsEquivalentToHistory_notNil);assumption.
              apply (IsHInit_automata). apply (IsAstdNamedN_automata).
              assumption. apply (IsAutomata). assumption.
        }
Qed.        
