Require Import ASTD.
Require Import states.
Require Import B_langage.
Require Import B_state.
Require Import B_semantics.
Require Import String.
Require Import ASTDstep.
Require Import stateEquivalence.
Require Import translation.
Require Import List.
Require Import Theorems.
Require Import theoremsOnNames.
Require Import theoremsOnUniqueAstdList.
Require Import theoremsOnUniqueBState.

Lemma existsAnAstdIfMakeDFPredTrue :
  forall (states : listOfAstdStates)
         (vLi : bStateList)
         (dF : list string)
         (name nameOfState : string),
    predicateIsTrueInState (makeDFinalPred dF states name) (Automata_B_state name nameOfState vLi) ->
    astdNameAreUniqueInAstdList states ->
    variableNamesAreUniqueInBState (Automata_B_state name nameOfState vLi) ->
    exists (a : astd),
      isAstdNamedNFromAstdList a nameOfState states.
Proof.
  induction states.
  -
    intros vLi dF name nameOfState.
    intros PredTrue.
    inversion PredTrue.
  -
    intros vLi dF name nameOfState.
    intros PredTrue AstdNamesUnique BNamesUnique.
    simpl in PredTrue.
    destruct (in_dec string_dec (get_name a) dF).
    +
      inversion PredTrue;subst.
      *
        inversion H2;subst.
        inversion H1;subst.
        exists a.
        apply (IsAstdNamedNFromAstdList_isElem).
        rewrite_values.
        apply (get_nameWorks2).
      *
        inversion AstdNamesUnique;subst.
        destruct (IHstates vLi dF name nameOfState H2 H3 BNamesUnique).
        exists x.
        apply (IsAstdNamedNFromAstdList_isNotElem).
        assumption.
    +
      inversion AstdNamesUnique;subst.
      destruct (IHstates vLi dF name nameOfState PredTrue H2 BNamesUnique).
      exists x.
      apply (IsAstdNamedNFromAstdList_isNotElem).
      assumption.
Qed.              

Theorem isFinalWorks :
  forall (a : astd)
         (s : state)
         (name : string)
         (v : bState),
    predicateIsTrueInState (finalPredicate a) v ->
    astdNameAreUniqueInAstd a ->
    isAStateOfTheAstd  s a ->
    variableNamesAreUniqueInBState v ->
    isAstdNamedN a name ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    isStateFinalInAstd s a.
Proof.
  apply (astd_mut
           (fun (a : astd) =>
              forall (s : state) (nameOfa : string) (v : bState),
                predicateIsTrueInState (finalPredicate a) v ->
                astdNameAreUniqueInAstd a ->
                isAStateOfTheAstd s a ->
                variableNamesAreUniqueInBState v ->
                isAstdNamedN a nameOfa ->
                bStateAndAstdStateAreEquivalentInAstdNamed v s nameOfa ->
                isStateFinalInAstd s a)
           (fun (li : listOfAstdStates) =>
              forall (bSubState : bState)
                     (listOfBSubStates vLi : bStateList)
                     (subAstd : astd)
                     (name nameOfState : string)
                     (subState : state)
                     (dFinal : list string),
                predicateIsTrueInState (makeDFinalPred dFinal li name)
                                       (Automata_B_state name
                                                         nameOfState
                                                         listOfBSubStates) ->
                astdNameAreUniqueInAstdList li ->
                isAStateOfTheAstd subState subAstd ->
                variableNamesAreUniqueInBState (Automata_B_state
                                                  name nameOfState
                                                  listOfBSubStates) ->
                listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState listOfBSubStates vLi nameOfState bSubState ->
                isAstdNamedNFromAstdList subAstd nameOfState li ->
                bStateAndAstdStateAreEquivalentInAstdNamed bSubState subState nameOfState ->
                In nameOfState dFinal /\ isStateFinalInAstd subState subAstd)).
  -
    intros name s nameOfA v.
    intros PredTrue NameUnqiue IsAState.
    inversion IsAState;subst.
    intros.
    apply IsStateFinalInAstd_elem.
  -
    intros name states IHstates trans sF dF init s nameOfA v.
    intros PredTrue AstdNameUnique IsAState VarUnique IsAstdNamed StateEquiv.
    simpl in PredTrue.
    destruct sF,dF.
    +
      inversion PredTrue.
    +
      inversion IsAstdNamed;subst.
      inversion IsAState;subst.
      inversion StateEquiv;subst.
      inversion AstdNameUnique;subst.
      destruct (IHstates bSubState listBSubStateRes listBSubState subAstd nameOfA nameOfState subState (s0::dF) PredTrue H12 H7 VarUnique H9 H3 H6).
      apply IsStateFinalInAstd_deepFinal with subAstd;assumption.
    +
      inversion IsAstdNamed;subst.
      inversion IsAState;subst.
      inversion StateEquiv;subst.
      inversion PredTrue;subst.
      apply IsStateFinalInAstd_shallowFinal.
      rewrite_values.
      assumption.
    +
      inversion PredTrue;subst.
      *
        inversion IsAState;subst.
        inversion StateEquiv;subst.
        inversion H2;subst.
        inversion IsAstdNamed;subst.
        apply IsStateFinalInAstd_shallowFinal.
        rewrite_values.
        assumption.
      *
        inversion IsAState;subst.
        inversion StateEquiv;subst.
        inversion IsAstdNamed;subst.
        inversion AstdNameUnique;subst.
        destruct (IHstates bSubState listBSubStateRes listBSubState subAstd nameOfA nameOfState subState (s1::dF) H2 H13 H8 VarUnique H10 H4 H7).
        apply IsStateFinalInAstd_deepFinal with subAstd;assumption.
  -
    intros bSubState vLi vLi' subAstd name nameOfStae subState dF.
    intros PredTrue.
    simpl in PredTrue.
    inversion PredTrue.
  -
    intros a IHa states IHstates bSubSt vLi vLi' subAstd name nameOfState subState dF.
    intros PredTrue AstdNameUnique IsAState BNameUnique ListReplaced IsAstdNamed StateEquiv.
    simpl in PredTrue.
    destruct (in_dec string_dec (get_name a) dF).
    +
      inversion PredTrue;subst.
      *
        inversion H2;subst.
        inversion H1;subst.
        rewrite_values.
        split;[assumption|].
        assert (a = subAstd).
        apply (isAstdNamedN_uniqueInAstdList a subAstd (ConsAstdStates a states) (get_name a)).
        apply (IsAstdNamedNFromAstdList_isElem).
        apply (get_nameWorks2).
        assumption.
        assumption.
        subst.
        apply (IHa subState (get_name subAstd) bSubSt).
        apply (finalPredicateIsTrueInASubState subAstd subState vLi vLi' name (get_name subAstd) bSubSt H4 StateEquiv).
        apply (isAstdNamedInList_implies_isAstdNamed subAstd (get_name subAstd) (ConsAstdStates subAstd states) IsAstdNamed).
        assumption.
        assumption.
        assumption.
        inversion AstdNameUnique.
        assumption.
        inversion AstdNameUnique.
        assumption.
        assumption.
        inversion BNameUnique.
        apply (variableNamesUnique_preservedByInclusion (get_name subAstd) bSubSt vLi vLi' H7 ListReplaced).
        apply (isAstdNamedInList_implies_isAstdNamed subAstd (get_name subAstd) (ConsAstdStates subAstd states) IsAstdNamed).
        assumption.
      *
        inversion AstdNameUnique;subst.
        apply (IHstates bSubSt vLi vLi' subAstd name nameOfState subState dF H2 H3 IsAState BNameUnique ListReplaced).
        destruct (existsAnAstdIfMakeDFPredTrue states vLi dF name nameOfState H2 H3 BNameUnique).
        rewrite (isAstdNamedN_uniqueInAstdList subAstd x (ConsAstdStates a states) nameOfState IsAstdNamed).
        assumption.
        apply (IsAstdNamedNFromAstdList_isNotElem).
        assumption.
        assumption.
        assumption.
    +
      inversion AstdNameUnique;subst.
      destruct (existsAnAstdIfMakeDFPredTrue states vLi dF name nameOfState PredTrue H2 BNameUnique).
      apply (IHstates bSubSt vLi vLi' subAstd name nameOfState subState dF PredTrue H2 IsAState BNameUnique ListReplaced).
        rewrite (isAstdNamedN_uniqueInAstdList subAstd x (ConsAstdStates a states) nameOfState IsAstdNamed).
        assumption.
        apply (IsAstdNamedNFromAstdList_isNotElem).
        assumption.
        assumption.
        assumption.
Qed.
