Require Import translation.
Require Import B_state.
Require Import ASTD.
Require Import states.
Require Import String.
Require Import B_semantics.
Require Import B_langage.
Require Import B_state.
Require Import stateEquivalence.
Require Import ASTDstep.
Require Import initAllVarWorks.
Require Import theoremsOnSubstitutions.
Require Import theoremsOnNames.
Require Import Theorems.
Require Import isFinalWorks.
Open Scope list_scope.
Require Import theoremsOnUniqueAstdList.
Require Import theoremsOnUniqueBState.

Lemma existsAStateOfAnAstdInStateEquivToHistory :
  forall (states : listOfAstdStates)
         (a : astd)
         (name : string)
         (vLi : bStateList)
         (h : histState),
    isAstdNamedNFromAstdList a name states ->
    isAutomata a ->
    isAHistoryStateOfAstdList h states ->
    listOfBStatesIsEquivalentToHistory vLi h ->
    astdNameAreUniqueInAstdList states ->
    exists (s : state)
           (v : bState),
      isAStateOfTheAstd s a /\
      bStateAndAstdStateAreEquivalentInAstdNamed v s name /\
      isAStateOfBStateList v vLi.
Proof.
  induction states.
  -
    intros a name vLi h IsAstdNamed IsAuto IsHist ListEquiv NamesUnique.
    inversion IsAstdNamed.
  -
    intros a0 name vLi h IsAstdNamed IsAuto IsHist ListEquiv NamesUnique.
    destruct a0;inversion IsAuto;subst.
    inversion IsAstdNamed;subst.
    +
      inversion H2;subst.
      inversion IsHist;subst.
      inversion ListEquiv;subst.
      exists hState.
      exists bst.
      split;[assumption|split].
      assumption.
      apply (IsAStateOfBStateList_isElem).
    +
      inversion NamesUnique;subst.
      inversion IsHist;subst.
      *
        apply (IHstates (Automata s l l0 l1 l2 s0) name vLi h H2 IsAuto H5 ListEquiv H3).
      *
        inversion ListEquiv;subst.
        destruct (IHstates (Automata s l l0 l1 l2 s0) name bState_queue hRest H2 IsAuto H6 H8 H3).
        destruct H.
        destruct H.
        destruct H0.
        exists x.
        exists x0.
        split;[assumption |split].
        assumption.
        apply (IsAStateOfBStateList_isNotElem).
        assumption.
Qed.

Lemma existsAStateOfAnAstdInHistory :
  forall (states : listOfAstdStates)
         (a a' : astd)
         (name name' : string)
         (vLi vLi' : bStateList)
         (v' : bState)
         (s' : state)
         (h : histState),
    isAstdNamedNFromAstdList a name states ->
    isAutomata a ->
    isAHistoryStateOfAstdList h states ->
    listOfBStatesIsEquivalentToHistory vLi h ->
    isAstdNamedNFromAstdList a' name' states ->
    isAStateOfTheAstd s' a' ->
    bStateAndAstdStateAreEquivalentInAstdNamed v' s' name' ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi name' v' ->
    astdNameAreUniqueInAstdList states ->
    exists (s : state)
           (v : bState),
      isAStateOfTheAstd s a /\
      bStateAndAstdStateAreEquivalentInAstdNamed v s name /\
      isAStateOfBStateList v vLi'.
Proof.
  intros states a a' name name' vLi vLi' v' s' h.
  intros IsAstdNamed1 IsAuto IsHist ListEquiv IsAstdNamed2 IsAState StateEquiv ListReplaced NameUnique.
  destruct (string_dec name name').
  -
    subst.
    assert (a = a').
    apply (isAstdNamedN_uniqueInAstdList a a' states name' IsAstdNamed1 IsAstdNamed2 NameUnique).
    subst.
    exists s'.
    exists v'.
    split;[assumption|split].
    +
      assumption.
    +
      inversion IsAuto;subst.
      inversion IsAState;subst.
      inversion StateEquiv;subst.
      apply (isAStateOfListWhereReplaced) with vLi.
      assumption.
  -
    destruct (existsAStateOfAnAstdInStateEquivToHistory states a name vLi h IsAstdNamed1 IsAuto IsHist ListEquiv NameUnique).
    destruct H.
    destruct H.
    destruct H0.
    exists x.
    exists x0.
    split;[assumption | split].
    assumption.
    destruct v'.
    +
      rewrite (replacingByElemBState vLi' vLi name' ListReplaced).
      assumption.
    +
      inversion IsAuto;subst.
      inversion H;subst.
      inversion H0;subst.
      apply (isAStateOfListWhereReplacedNotNamed name' name s0 nameOfState vLi vLi' b listBSubStateRes).
      unfold not.
      intros.
      unfold not in n.
      apply n.
      rewrite H2.
      reflexivity.
      assumption.
      rewrite <- (nameOfReplacingState s s0 name' vLi vLi' b) in ListReplaced.
      assumption.
      assumption.
Qed.

Lemma overLoadedByElemState :
  forall (h : histState)
         (name : string),
    histIsHistOverLoadedByNameAndState h h name Elem_state.
Proof.
  induction h.
  -
    intro.
    apply (Hist2IsHist1OverLoadedByNameAndState_nil).
  -
    intro.
    apply (Hist2IsHist1OverLoadedByNameAndState_isNotElem).
    apply (IHh name).
Qed.


Lemma histStateEquivToBStateListExists :
  forall (vLi vLi' : bStateList)
         (h : histState)
         (s : state)
         (name : string)
         (v : bState),
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi name v ->
    listOfBStatesIsEquivalentToHistory vLi h ->
    exists (h' : histState),
      listOfBStatesIsEquivalentToHistory vLi' h' /\
      histIsHistOverLoadedByNameAndState h' h name s.
Proof.
  induction vLi.
  -
    intros vLi' h s name v StateEquiv StateReplaced EquivToHist.
    inversion StateReplaced;subst.
    inversion EquivToHist;subst.
    exists EmptyHistState.
    split.
    +
      apply ListOfBStatesIsEquivalentToHistory_nil.
    +
      inversion StateEquiv;subst.
      apply (Hist2IsHist1OverLoadedByNameAndState_nil).
  -
    intros vLi' h s name v StateEquiv StateReplaced EquivToHist.
    inversion StateReplaced;subst;inversion StateEquiv;subst;inversion EquivToHist;subst.
    +
      inversion H7;subst.
      exists (ConsHistState name0 (Automata_state val2 h0 subState) h_queue).
      split.
      *
        apply (ListOfBStatesIsEquivalentToHistory_notNil );assumption.
      *
        apply (Hist2IsHist1OverLoadedByNameAndState_isElem).
    +
      exists (ConsHistState name0 st h_queue).
      split.
      *
        rewrite (replacingByElemBState bStateReplaced vLi name H4).
        assumption.
      *
        apply overLoadedByElemState.
    +
      destruct (IHvLi bStateReplaced h_queue (Automata_state n1 h0 subState) name (Automata_B_state name n1 listBSubStateRes) StateEquiv H4 H5).
      destruct H2.
      exists (ConsHistState name0 st x).
      split.
      *
        apply (ListOfBStatesIsEquivalentToHistory_notNil);assumption.
      *
        apply (Hist2IsHist1OverLoadedByNameAndState_isNotElem);assumption.
Qed.


Lemma existAStateAfterInitAllVar :
  forall (a : astd)
         (s : state)
         (name : string)
         (v : bState),
    isAStateOfTheAstd s a ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    isAstdNamedN a name ->
    variableNamesAreUniqueInBState v ->
    astdNameAreUniqueInAstd a ->
    exists (v' : bState),
      substitution_step v (initializeAllVariable a) v'.
Proof.
  apply (astd_mut (fun (a : astd) =>
                     forall (s : state)
                            (name : string)
                            (v : bState),
                       isAStateOfTheAstd s a ->
                       bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
                       isAstdNamedN a name ->
                       variableNamesAreUniqueInBState v ->
                       astdNameAreUniqueInAstd a ->
                       exists (v' : bState),
                         substitution_step v (initializeAllVariable a) v')
                  (fun (states : listOfAstdStates) =>
                     forall (h : histState)
                            (var val : string)
                            (vLi : bStateList),
                       isAHistoryStateOfAstdList h states ->
                       listOfBStatesIsEquivalentToHistory vLi h ->
                       variableNamesAreUniqueInBStateList vLi ->
                       astdNameAreUniqueInAstdList states ->
                       exists (vLi' : bStateList),
                         substitution_step (Automata_B_state var val vLi) (initializeFromAstdList states) (Automata_B_state var val vLi'))).
  -
    intros nameOfA s name v IsAState StateEquiv IsAstdNamed.
    simpl.
    exists v.
    apply (Substitution_step_skipSubstitution).
  -
    intros nameOfA states IHstates trans dF sF init s name v.
    intros IsAState StateEquiv  IsAstdNamed BNamesUnique AstdNamesUnique.
    inversion IsAState;subst.
    inversion StateEquiv;subst.
    inversion BNamesUnique;subst.
    inversion AstdNamesUnique;subst.
    simpl.
    destruct (existsAHistStateEquivToList states h subAstd nameOfState subState bSubState listBSubState listBSubStateRes H8 H3 H7 H4 H9 H6 H5 H14).
    destruct H.
    destruct (IHstates x name init listBSubStateRes H0 H H5 H14).
    exists (Automata_B_state name init x0).
    apply (Substitution_step_parallel) with (Automata_B_state name init listBSubStateRes).
    +
      apply (Substitution_step_affectation).
      inversion IsAstdNamed;subst.
      apply (IsStateWhereVarChangedForVal_isElem).
    +
      assumption.
  -
    intros h var val vLi IsHist ListEquiv BNamesUnique.
    simpl.
    exists vLi.
    apply (Substitution_step_skipSubstitution).
  -
    intros a IHa states IHstates h var val vLi IsHist ListEquiv BNamesUnique AstdNamesUnique.
    destruct a.
    +
      inversion IsHist;subst.
      inversion AstdNamesUnique;subst.
      destruct (IHstates h var val vLi H1 ListEquiv BNamesUnique H3).
      exists x.
      simpl.
      apply (Substitution_step_parallel) with (Automata_B_state var val vLi).
      *
        apply (Substitution_step_skipSubstitution).
      *
        assumption.
    +
      simpl.
      inversion IsHist;subst.
      inversion ListEquiv;subst.
      inversion BNamesUnique;subst.
      inversion AstdNamesUnique;subst.
      destruct (IHa hState s bst H8 H5).
      apply (IsAstdNamedN_automata).
      assumption.
      assumption.
      destruct (IHstates hRest var val bState_queue H2 H3).
      assumption.
      assumption.
      exists (ConsBStateList x x0).
      apply (Substitution_step_parallel) with (Automata_B_state var val (ConsBStateList x bState_queue)).
      *
        apply substitutionStepInABiggerState.
        simpl in H.
        assumption.
      *
        apply substitutionStepInABiggerState2.
        assumption.
Qed.

Lemma substitutionInASubState :
  forall (var val val' var1 val1 val2 : string)
         (vLi vLi' vLiSub vLiSub' : bStateList)
         (sub : substitution),
    variableNamesAreUniqueInBState (Automata_B_state var val vLi) ->
    substitution_step (Automata_B_state var val vLi) sub (Automata_B_state var val' vLi') ->
    substitution_step (Automata_B_state var1 val1 vLiSub) sub (Automata_B_state var1 val2 vLiSub') ->
    substitutionContainsNoSelect sub ->
    isAStateOfBStateList (Automata_B_state var1 val1 vLiSub) vLi ->
    (listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi var1 (Automata_B_state var1 val2 vLiSub') /\ val = val').
Proof.
  induction vLi.
  -
    intros vLi' vLiSub vLiSub' sub NamesUnique Step1 Step2 NoSelect IsAState.
    inversion IsAState.
  -
    intros vLi' vLiSub vLiSub' sub NamesUnique Step1 Step2 NoSelect IsAState.
    inversion IsAState;subst.
    +
      destruct (substitutionPreserveSubStateStructure1 sub var val val' (Automata_B_state var1 val1 vLiSub) vLi vLi' Step1 NoSelect).
      inversion NamesUnique;subst.
      intros name VarMod.
      split.
      *
        inversion H3;subst.
        destruct (aModifiedVariableIsInState sub name (Automata_B_state var1 val1 vLiSub) (Automata_B_state var1 val2 vLiSub') VarMod Step2 NoSelect).
        apply (H4 name x H).
      *
        inversion H1;subst.
        {
          destruct (string_dec var name).
          -
            subst.
            destruct (aModifiedVariableIsInState sub name (Automata_B_state var1 val1 vLiSub) (Automata_B_state var1 val2 vLiSub') VarMod Step2 NoSelect).
            apply (absurd_isValOfVar_noVariableNamed (Automata_B_state var1 val1 vLiSub) name x);assumption.
          -
            assumption.
        }
      *
        destruct H.
        destruct H0.
        subst.
        split;[|reflexivity].
        assert (x = (Automata_B_state var1 val2 vLiSub')).
        inversion NamesUnique;subst.
        inversion H4;subst.
        apply (substitutionWithoutSelectIsDeterministic sub (Automata_B_state var1 val1 vLiSub) x (Automata_B_state var1 val2 vLiSub') H0 Step2 NoSelect H7).
        subst.
        apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem).
    +
      destruct (substitutionPreserveSubStateStructure2 sub var val val' b vLi vLi' Step1 NoSelect).
      inversion NamesUnique;subst.
      intros name VarMod.
      inversion H4;subst.
      destruct (aModifiedVariableIsInState sub name (Automata_B_state var1 val1 vLiSub) (Automata_B_state var1 val2 vLiSub') VarMod Step2 NoSelect).
      split.
      *
        apply (H6 x name).
        apply (isValOfVarInStateListIfIsValOfVarInState (Automata_B_state var1 val1 vLiSub) vLi name x H H1).
      *
        {
          destruct (string_dec var name).
          -
            subst.
            inversion H2;subst.
            apply (absurd_isValOfVarInList_noVariableNamedInList vLi name x).
            apply (isValOfVarInStateListIfIsValOfVarInState (Automata_B_state var1 val1 vLiSub) vLi name x H H1).
            assumption.
          -
            assumption.
        }
      *
        destruct H.
        subst.
        split;[|reflexivity].
        apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isNotElem).
        {
          destruct (IHvLi x vLiSub vLiSub' sub).
          -
            inversion NamesUnique;subst.
            inversion H4;subst.
            inversion H2;subst.
            apply (VariableNamesAreUniqueInBState_automata).
            +
              assumption.
            +
              assumption.
          -
            apply (substitutionInASmallerState) with b.
            assumption.
            assumption.
            inversion NamesUnique.
            intros name VarMod;subst.
            inversion H4;subst.
            destruct (aModifiedVariableIsInState sub name (Automata_B_state var1 val1 vLiSub) (Automata_B_state var1 val2 vLiSub') VarMod Step2 NoSelect).
            split.
            +
              apply (H6 x0 name).
              apply (isValOfVarInStateListIfIsValOfVarInState (Automata_B_state var1 val1 vLiSub) vLi name x0 H H1).
            +
              destruct (string_dec var name).
              *
                subst.
                inversion H2;subst.
                apply (absurd_isValOfVarInList_noVariableNamedInList vLi name x0).
                apply (isValOfVarInStateListIfIsValOfVarInState (Automata_B_state var1 val1 vLiSub) vLi name x0 H H1).
                assumption.
              *
                assumption.
          -
            assumption.
          -
            assumption.
          -
            assumption.
          -
            assumption.
        }
Qed.

Lemma makePrePostWorks_forLocalFinal :
  forall (n1 n2 : string)
         (label : transition_label)
         (v v' : bState)
         (s : state)
         (gu : guard)
         (sFinal dFinal : list string)
         (name init : string)
         (pre : predicate)
         (trans : list transition)
         (post : substitution)
         (states : listOfAstdStates),
    makePrecondition (Local n1 n2 label gu true) name states = Some pre ->
    makePostcondition (Local n1 n2 label gu true) name states = Some post ->
    predicateIsTrueInState pre v ->
    substitution_step v post v' ->
    variableNamesAreUniqueInBState v ->
    isAStateOfTheAstd s (Automata name states (Local n1 n2 label gu true :: trans) sFinal dFinal init) ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    isAstdListWellDefined states ->
    astdNameAreUniqueInAstd (Automata name states (Local n1 n2 label gu true :: trans) sFinal dFinal init) ->
    exists s' : state,
      bStateAndAstdStateAreEquivalentInAstdNamed v' s' name /\
      astd_step s (Automata name states (Local n1 n2 label gu true :: trans) sFinal dFinal init) label s'.
Proof.
  intros n1 n2 label v v' s gu sF dF name init pre trans post states.
  intros MakePre MakePost PreTrue PostStep BNameUnique IsAState StateEquiv WellDef AstdNamesUnique.
  simpl in MakePre.
  simpl in MakePost.
  case_eq (getAstdFromStateList n1 states);intros;rewrite H in MakePre;[|discriminate].
  case_eq (getAstdFromStateList n2 states);intros;rewrite H0 in MakePost;[|discriminate].
  inversion MakePre.
  inversion MakePost.
  subst.
  inversion IsAState;subst.
  inversion StateEquiv;subst.
  inversion PostStep;subst.
  rewrite (affectationInUniqueState name nameOfState n2 listBSubStateRes s2 H4 BNameUnique) in H12.
  apply (getAstdFromStateListWorks) in H0.
  inversion PreTrue;subst.
  inversion H3;subst.
  inversion H7;subst.
  inversion H15;subst.
  rewrite_values.
  apply (getAstdFromStateListWorks) in H.
  inversion AstdNamesUnique;subst.
  rewrite (isAstdNamedN_uniqueInAstdList subAstd a states nameOfState H5 H H22) in H9.
  apply (finalPredicateIsTrueInASubState a subState  listBSubStateRes listBSubState name nameOfState bSubState) in H13;[ |assumption | |assumption |assumption|assumption |].
  apply (isFinalWorks a subState nameOfState bSubState) in H13;[ | |assumption | | | ].
  destruct a0.
  -
    simpl in H12.
    inversion H12;subst.
    destruct (histStateEquivToBStateListExists listBSubState listBSubStateRes h subState nameOfState bSubState H8 H11 H6).
    destruct H1.
    exists (Automata_state n2 x Elem_state).
    split.
    +
      apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem name n2 x Elem_state Elem_B_state listBSubStateRes listBSubStateRes H1).
      apply (BStateAndAstdStateAreEquivalentInAstdNamed_elem).
      apply (replacingByElemBState_reverse).
    +
      destruct gu.
      apply Automata_rule1 with (gu := s0) (a2 := Elem s) (a1 := a) (final := true).
      *
        simpl.
        left.
        reflexivity.
      *
        intros.
        assumption.
      *
        assumption.
      *
        assumption.
      *
        assumption.
      *
        assumption.
      *
        apply (IsInitialStateOfAstd_elem).
  -
    destruct (existsAStateOfAnAstdInHistory states (Automata s l l0 l1 l2 s0) a n2 nameOfState listBSubState listBSubStateRes bSubState subState h H0);[apply IsAutomata|assumption |assumption |assumption |assumption |assumption |assumption |assumption | ].
    destruct H1.
    destruct H1.
    destruct H14.
    destruct (existAStateAfterInitAllVar (Automata s l l0 l1 l2 s0) x n2 x0 H1 H14).
    apply (isAstdNamedInList_implies_isAstdNamed) with states.
    assumption.
    inversion BNameUnique;subst.
    apply (variableNamesAreUniqueInABSubState x0 listBSubStateRes H24 H18).
    apply astdNamesUniqueInASubAstd with (states := states) (name := n2);assumption.
    destruct (initAllVarWorks (Automata s l l0 l1 l2 s0) x0 x1 x n2 H1 H14).
    apply (isAstdNamedInList_implies_isAstdNamed) with states.
    assumption.
    apply subAstdWellDefined with (states := states) (name := n2);assumption.
    apply (variableNamesAreUniqueInABSubState) with listBSubStateRes.
    inversion BNameUnique.
    assumption.
    assumption.
    assumption.
    destruct H20.
    destruct (histStateEquivToBStateListExists listBSubState listBSubStateRes h subState nameOfState bSubState).
    assumption.
    assumption.
    assumption.
    destruct H23.
    exists (Automata_state n2 x3 x2).
    split.
    +
      destruct (substitutionDoesNotChangeStateStructure (initializeAllVariable (Automata s l l0 l1 l2 s0)) name n2 listBSubStateRes v' H12).
      destruct H25;subst.
      rewrite <- (initAllVarDoesNotChangeValueIfNoAstdNamed (Automata s l l0 l1 l2 s0) name n2 x4 listBSubStateRes x5).
      *
        apply (BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem) with (listBSubState := listBSubStateRes) (bSubState := x1).
        assumption.
        assumption.
        inversion H1;subst.
        inversion H14;subst.
        destruct (substitutionDoesNotChangeStateStructure (initializeAllVariable (Automata s l l0 l1 l2 s0)) n2 nameOfState0 listBSubStateRes0 x1 H19).
        destruct H25;subst.
        destruct (substitutionInASubState  name n2 x4 n2 nameOfState0 x listBSubStateRes x5 listBSubStateRes0 x0 (initializeAllVariable (Automata s l l0 l1 l2 s0))).
        apply (uniquePreservedBySubstitution (Affectation name n2) (Automata_B_state name nameOfState listBSubStateRes) (Automata_B_state name n2 listBSubStateRes) BNameUnique).
        rewrite <- (affectationInUniqueState name nameOfState n2 listBSubStateRes s2 H4 BNameUnique).
        assumption.
        assumption.
        assumption.
        apply (initAllVarContainsNoSelect).
        assumption.
        assumption.
      *
        apply (noAstdNamedInASubState  (Automata s l l0 l1 l2 s0)  name n2 states H17 H0).
      *
        assumption.
    +
      destruct gu.
      apply (Automata_rule1) with (gu := s1) (a1 := a) (a2 := (Automata s l l0 l1 l2 s0)) (final := true).
      simpl. left. reflexivity.
      intros.
      assumption.
      assumption.
      assumption.
      assumption.
      assumption.
      assumption.
  -
    apply (astdNamesUniqueInASubAstd states nameOfState a H H22).
  -
    inversion BNameUnique.
    apply (variableNamesAreUniqueInStateWhereReplaced bSubState nameOfState listBSubStateRes listBSubState H11 H20).
  -
    apply isAstdNamedInList_implies_isAstdNamed with states. assumption.
  -
    assumption.
  -
    apply isAstdNamedInList_implies_isAstdNamed with states. assumption.
  -
    apply (astdNamesUniqueInASubAstd states nameOfState a H H22).
Qed.

Lemma makePrePostWorks_forLocalNonFinal :
  forall (n1 n2 : string)
         (label : transition_label)
         (v v' : bState)
         (s : state)
         (gu : guard)
         (sFinal dFinal : list string)
         (name init : string)
         (pre : predicate)
         (trans : list transition)
         (post : substitution)
         (states : listOfAstdStates),
    makePrecondition (Local n1 n2 label gu false) name states = Some pre ->
    makePostcondition (Local n1 n2 label gu false) name states = Some post ->
    predicateIsTrueInState pre v ->
    substitution_step v post v' ->
    variableNamesAreUniqueInBState v ->
    isAStateOfTheAstd s (Automata name states (Local n1 n2 label gu false :: trans) sFinal dFinal init) ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    astdNameAreUniqueInAstd (Automata name states (Local n1 n2 label gu false :: trans) sFinal dFinal init) ->
    isAstdListWellDefined states ->
    exists s' : state,
      bStateAndAstdStateAreEquivalentInAstdNamed v' s' name /\
      astd_step s (Automata name states (Local n1 n2 label gu false :: trans) sFinal dFinal init) label s'.
Admitted.

Lemma makePrePostWorks_forFromSubFinal :
  forall (n1 n2 n1b : string)
         (label : transition_label)
         (v v' : bState)
         (s : state)
         (gu : guard)
         (sFinal dFinal : list string)
         (name init : string)
         (pre : predicate)
         (trans : list transition)
         (post : substitution)
         (states : listOfAstdStates),
    makePrecondition (From_sub n1b n2 n1 label gu true) name states = Some pre ->
    makePostcondition (From_sub n1b n2 n1 label gu true) name states = Some post ->
    predicateIsTrueInState pre v ->
    substitution_step v post v' ->
    variableNamesAreUniqueInBState v ->
    isAStateOfTheAstd s (Automata name states (From_sub n1b n2 n1 label gu true :: trans) sFinal dFinal init) ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    astdNameAreUniqueInAstd (Automata name states (From_sub n1b n2 n1 label gu true :: trans) sFinal dFinal init) ->
    isAstdListWellDefined states ->
    exists s' : state,
      bStateAndAstdStateAreEquivalentInAstdNamed v' s' name /\
      astd_step s (Automata name states (From_sub n1b n2 n1 label gu true :: trans) sFinal dFinal init) label s'.
Admitted.

Lemma makePrePostWorks_forFromSubNonFinal :
  forall (n1 n2 n1b : string)
         (label : transition_label)
         (v v' : bState)
         (s : state)
         (gu : guard)
         (sFinal dFinal : list string)
         (name init : string)
         (pre : predicate)
         (trans : list transition)
         (post : substitution)
         (states : listOfAstdStates),
    makePrecondition (From_sub n1b n2 n1 label gu false) name states = Some pre ->
    makePostcondition (From_sub n1b n2 n1 label gu false) name states = Some post ->
    predicateIsTrueInState pre v ->
    substitution_step v post v' ->
    variableNamesAreUniqueInBState v ->
    isAStateOfTheAstd s (Automata name states (From_sub n1b n2 n1 label gu false :: trans) sFinal dFinal init) ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    astdNameAreUniqueInAstd (Automata name states (From_sub n1b n2 n1 label gu false :: trans) sFinal dFinal init) ->
    isAstdListWellDefined states ->
    exists s' : state,
      bStateAndAstdStateAreEquivalentInAstdNamed v' s' name /\
      astd_step s (Automata name states (From_sub n1b n2 n1 label gu false :: trans) sFinal dFinal init) label s'.
Admitted.

Lemma makePrePostWorks_forToSubFinal :
  forall (n1 n2 n2b : string)
         (label : transition_label)
         (v v' : bState)
         (s : state)
         (gu : guard)
         (sFinal dFinal : list string)
         (name init : string)
         (pre : predicate)
         (trans : list transition)
         (post : substitution)
         (states : listOfAstdStates),
    makePrecondition (To_sub n1 n2b n2 label gu true) name states = Some pre ->
    makePostcondition (To_sub n1 n2b n2 label gu true) name states = Some post ->
    predicateIsTrueInState pre v ->
    substitution_step v post v' ->
    variableNamesAreUniqueInBState v ->
    isAStateOfTheAstd s (Automata name states (To_sub n1 n2b n2 label gu true :: trans) sFinal dFinal init) ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    astdNameAreUniqueInAstd (Automata name states (To_sub n1 n2b n2 label gu true :: trans) sFinal dFinal init) ->
    isAstdListWellDefined states ->
    exists s' : state,
      bStateAndAstdStateAreEquivalentInAstdNamed v' s' name /\
      astd_step s (Automata name states (To_sub n1 n2b n2 label gu true :: trans) sFinal dFinal init) label s'.
Admitted.

Lemma makePrePostWorks_forToSubNonFinal :
  forall (n1 n2 n2b : string)
         (label : transition_label)
         (v v' : bState)
         (s : state)
         (gu : guard)
         (sFinal dFinal : list string)
         (name init : string)
         (pre : predicate)
         (trans : list transition)
         (post : substitution)
         (states : listOfAstdStates),
    makePrecondition (To_sub n1 n2b n2 label gu false) name states = Some pre ->
    makePostcondition (To_sub n1 n2b n2 label gu false) name states = Some post ->
    predicateIsTrueInState pre v ->
    substitution_step v post v' ->
    variableNamesAreUniqueInBState v ->
    isAStateOfTheAstd s (Automata name states (To_sub n1 n2b n2 label gu false :: trans) sFinal dFinal init) ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    astdNameAreUniqueInAstd (Automata name states (To_sub n1 n2b n2 label gu false :: trans) sFinal dFinal init) ->
    isAstdListWellDefined states ->
    exists s' : state,
      bStateAndAstdStateAreEquivalentInAstdNamed v' s' name /\
      astd_step s (Automata name states (To_sub n1 n2b n2 label gu false :: trans) sFinal dFinal init) label s'.
Admitted.

Theorem makePrePostWorks :
  forall (v v' : bState)
         (s : state)
         (nameOfA init nameOfOp : string)
         (args : list (string * string))
         (states : listOfAstdStates)
         (sFinal dFinal : list string)
         (pre : predicate)
         (t : transition)
         (trans : list transition)
         (post : substitution),
    makePrecondition t nameOfA states = Some pre ->
    makePostcondition t nameOfA states = Some post ->
    astdNameAreUniqueInAstd (Automata nameOfA states (t::trans) sFinal dFinal init) ->
    isAstdListWellDefined states ->
    predicateIsTrueInState pre v ->
    substitution_step v post v' ->
    variableNamesAreUniqueInBState v ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s nameOfA ->
    isAStateOfTheAstd s (Automata nameOfA states (t::trans) sFinal dFinal init) ->
    isLabelOfTransition (Transition_label nameOfOp args) t ->
    exists (s' : state),
      bStateAndAstdStateAreEquivalentInAstdNamed v' s' nameOfA /\
      astd_step s (Automata nameOfA states (t :: trans) sFinal dFinal init)
                (Transition_label nameOfOp args) s'.
Proof.
  intros v v' s nameOfA init nameOfOp args states sF dF pre t trans post.
  intros MakePre MakePost AstdNameUnique ListWellDefined PreTrue PostStep BNameUnique StateEquiv IsAState IsLabel.
  destruct t;destruct b;inversion IsLabel;subst.
  -
    apply (makePrePostWorks_forLocalFinal s0 s1 (Transition_label nameOfOp args) v v' s g sF dF nameOfA init pre trans post states);assumption.
  -
    apply (makePrePostWorks_forLocalNonFinal s0 s1 (Transition_label nameOfOp args) v v' s g sF dF nameOfA init pre trans post states);assumption.
  -
    apply (makePrePostWorks_forFromSubFinal s2 s1 s0 (Transition_label nameOfOp args) v v' s g sF dF nameOfA init pre trans post states);assumption.
  -
    apply (makePrePostWorks_forFromSubNonFinal s2 s1 s0 (Transition_label nameOfOp args) v v' s g sF dF nameOfA init pre trans post states);assumption.
  -
    apply (makePrePostWorks_forToSubFinal s0 s2 s1 (Transition_label nameOfOp args) v v' s g sF dF nameOfA init pre trans post states);assumption.
  -
    apply (makePrePostWorks_forToSubNonFinal s0 s2 s1 (Transition_label nameOfOp args) v v' s g sF dF nameOfA init pre trans post states);assumption.
Qed.
