Require Import translation.
Require Import B_semantics.
Require Import String.
Require Import B_langage.
Require Import B_state.
Require Import Theorems.
Require Import wellConstructed.

Lemma paramAreEqualsWorks :
  forall (args1 args2 : list (string * string)),
    paramAreEquals args1 args2 = true ->
    args1 = args2.
Proof.
  induction args1.
  -
    intros.
    inversion H.
    destruct args2.
    +
      reflexivity.
    +
      discriminate.
  -
    intros.
    simpl in H.
    destruct args2.
    +
      destruct a.
      discriminate.
    +
      destruct a.
      destruct p.
      destruct (string_dec s s1);
        destruct (string_dec s0 s2);subst.
      *
        rewrite (IHargs1 args2 H).
        reflexivity.
      *
        discriminate.
      *
        discriminate.
      *
        discriminate.
Qed.

Theorem opLabelsAreEqualWorks :
  forall (o0 o1 : operation_label),
    opLabelsAreEqual o0 o1 = true ->
    o0 = o1.
Proof.
  intros.
  destruct o0.
  destruct o1.
  unfold opLabelsAreEqual in H.
  destruct (string_dec s s0);subst.
  rewrite (paramAreEqualsWorks l l0 H).
  reflexivity.
  discriminate.
Qed.

Lemma addOperationWorks :
  forall (l liRes : list operation)
         (a : operation)
         (v v' : bState)
         (label : operation_label),
    addOperation a l = Some liRes ->
    operationsAreWellConstructed l ->
    b_step v liRes label v' ->
    (b_step v l label v' \/
     b_step v (a :: nil) label v').
Proof.
  induction l.
  -
    intros liRes a v v' label AddOp Step.
    simpl in AddOp.
    induction a.
    inversion AddOp;subst.
    right.
    assumption.
  -
    intros liRes op v v' label AddOp WellCons Step.
    simpl in AddOp.
    destruct op.
    destruct a.
    case_eq (opLabelsAreEqual o o0);intros;rewrite H in AddOp.
    +
      destruct s0;inversion AddOp.
      subst.
      inversion Step;subst.
      inversion H2;subst.
      *
        inversion H3;subst.
        inversion H1;subst.
        right.
        apply B_step with (pre := p) (post := s).
        assumption.
        assumption.
        simpl.
        left.
        reflexivity.
        left.
        apply B_step with (pre := p0) (post := Select s0).
        inversion WellCons;subst.
        apply (wellConstructedGuardIsTrue s0 p0 v v' H10 H9).
        assumption.
        rewrite (opLabelsAreEqualWorks label o0 H).
        simpl.
        left.
        reflexivity.
      *
        left.
        apply B_step with (pre := pre) (post := post).
        assumption.
        assumption.
        simpl.
        right.
        assumption.
    +
      case_eq (addOperation (Operation o p s) l);intros;rewrite H0 in AddOp;[|discriminate].
      inversion AddOp;subst.
      inversion Step;subst.
      simpl in H3.
      destruct H3.
      *
        inversion H3;subst.
        left.
        apply B_step with (pre := pre) (post := post);[assumption|assumption|].
        simpl.
        left.
        reflexivity.
      *
        inversion WellCons;subst.
        destruct (IHl l0 (Operation o p s) v v' label H0 H6).
        apply B_step with (pre := pre) (post := post);assumption.
        inversion H4;subst.
        left.
        apply B_step with (pre := pre0) (post := post0).
        assumption.
        assumption.
        simpl.
        right.
        assumption.
        right.
        assumption.
Qed.


Theorem mergeIsOneOrTheOther :
  forall (li1 li2 liRes : list operation)
         (l : operation_label)
         (v v' : bState),
    mergeOperationList li1 li2 = Some liRes ->
    operationsAreWellConstructed li1 ->
    operationsAreWellConstructed li2 ->
    b_step v liRes l v' ->
    (b_step v li1 l v' \/
     b_step v li2 l v').
Proof.
  induction li1.
  -
    intros li2 liRes l v v' Merge Step.
    simpl in Merge.
    inversion Merge;subst.
    right.
    assumption.
  -
    intros li2 liRes l v v' Merge WellConsLi1 WellConsLi2 Step.
    simpl in Merge.
    case_eq (mergeOperationList li1 li2);intros;rewrite H in Merge;[|discriminate].
    destruct (addOperationWorks l0 liRes a v v' l Merge).
    inversion WellConsLi1;subst.
    apply (wellConstructed_preservedByMerge li1 li2 l0 WellConsLi2 H2 H).
    assumption.
    *
      inversion WellConsLi1;subst.
      destruct (IHli1 li2 l0 l v v' H H3 WellConsLi2 H0).
      inversion H1.
      subst.
      left.
      apply B_step with (pre := pre) (post := post);[assumption|assumption|].
      simpl.
      right.
      assumption.
      right.
      assumption.
    *
      left.
      inversion H0;subst.
      inversion H3;subst;[|inversion H4].
      apply B_step with (pre := pre) (post := post).
      assumption.
      assumption.
      left.
      reflexivity.
Qed.
