Require Import B_state.
Require Import B_langage.
Require Import B_semantics.
Require Import translation.
Require Import String.
Require Import Theorems.
Require Import ASTD.
Require Import stateEquivalence.
Require Import states.
Require Import mergeWorks.
Require Import wellConstructed.
Require Import theoremsOnUniqueAstdList.
Require Import theoremsOnUniqueBState.
Require Import theoremsOnNames.

Lemma bStepIsStillPossibleWithMoreOpes:
  forall (op : operation)
         (v v' : bState)
         (name : string)
         (args : list (string * string))          
         (opes : list operation),
    b_step v (opes) (Label name args) v' ->
    b_step v (op::opes) (Label name args) v'.
Proof.
  intros op v v' name args opes.
  intros step.
  inversion step;subst.
  apply (B_step pre post (op::opes) (Label name args) v v' H H0).
  simpl.
  right.
  assumption.
Qed.  

Theorem modifyPrePostWorks2 :
  forall (opes : list operation)
         (name nameOfState : string)
         (vLi vLi' listBSubState : bStateList)
         (bSubState vSub' : bState)
         (a : astd)
         (subState : state)
         (states : listOfAstdStates)
         (label : operation_label),
    b_step (Automata_B_state name nameOfState vLi)
           (modifyPrePost opes name nameOfState) label
           (Automata_B_state name nameOfState vLi') ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState
      vLi
      listBSubState nameOfState bSubState ->
    translateAstd a = Some opes ->
    isAstdNamedNFromAstdList a nameOfState states ->
    isAStateOfTheAstd subState a ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState
      vLi' vLi nameOfState vSub' ->
    bStateAndAstdStateAreEquivalentInAstdNamed bSubState subState nameOfState ->
    b_step bSubState opes label vSub'.
Proof.
  induction opes.
  -
    intros name nameOfState vLi vLi' listBSubState bSubState vSub' a subState states label Step.
    simpl in Step.
    inversion Step;subst.
    inversion H1.
  -
    intros name nameOfState vLi vLi' listBSubState bSubState vSub' a0 subState states label Step ListReplace TranslateA IsAstdNamed IsAState ListReplaced StateEquiv.
    simpl in Step.
    destruct a.
    inversion Step;subst.
    destruct H1;subst.
    +
      inversion H1;subst.
      apply (B_step) with (pre := p) (post := s).
Admitted.      


 
Theorem modifyPrePostWorks :
  forall (opes : list operation)
         (name nameOfA nameOfOp : string)
         (v v' : bState)
         (args : list (string * string)),
    b_step v (modifyPrePost opes name nameOfA)  (Label nameOfOp args) v' ->
    isValOfVarInState nameOfA name v ->
    variableNamesAreUniqueInBState v ->
    b_step v opes (Label nameOfOp args) v'.
Proof.
  induction opes.
  -
    intros name nameOfA nameOfOp v v' args.
    intros step valOfVar varUnique.
    simpl in step.
    inversion step;subst.
    inversion H1.
  -
    intros name nameOfA nameOfOp v v' args.
    intros step valOfVar varUnique.
    simpl in step.
    destruct a.
    inversion step.
    inversion H1;subst.
    +
      inversion H6;subst.
      inversion H;subst.
      apply B_step with (pre := p) (post := s).
      *
        assumption.
      *
        inversion H0;subst.
        inversion H11;subst.
        rewrite (affectationInUniqueState2 name nameOfA v s2 valOfVar H8 varUnique) in H12.
        assumption.
        inversion H10.
      *
        simpl.
        left.
        reflexivity.
    +
      apply bStepIsStillPossibleWithMoreOpes.
      apply (IHopes name nameOfA nameOfOp v v' args);[|assumption|assumption].
      apply (B_step pre post);assumption.
Qed.

Lemma isValOfVar_modifyPrePost :
  forall (opes : list operation)
         (nameOfA name : string)
         (l : operation_label)
         (v v' : bState),
    b_step v (modifyPrePost opes nameOfA name) l v' ->
    isValOfVarInState name nameOfA v.
Proof.
  induction opes.
  -
    intros nameOfA name l v v'.
    intros step.
    simpl in step.
    inversion step;subst.
    inversion H1.
  -
    intros nameOfA name l v v' step.
    simpl in step.
    destruct a.
    inversion step.
    inversion H1;subst.
    +
      inversion H6;subst.
      inversion H;subst.
      inversion H4;subst.
      assumption.
    +
      apply (IHopes nameOfA name l v v').
      apply (B_step pre post);assumption.
Qed.

Lemma isAstdNamedNInModifyPrePost :
  forall (name nameOfState : string)
         (vLi vLi' : bStateList)
         (a : astd)
         (l : operation_label)
         (opes : list operation),
    b_step (Automata_B_state name nameOfState vLi) (modifyPrePost opes name (get_name a)) l (Automata_B_state name nameOfState vLi') ->
    variableNamesAreUniqueInBState (Automata_B_state name nameOfState vLi) ->
    isAstdNamedN a nameOfState.
Proof.
  intros.
  apply (isValOfVar_modifyPrePost opes name (get_name a) l (Automata_B_state name nameOfState vLi) (Automata_B_state name nameOfState vLi')) in H.
  rewrite_values.
  apply get_nameWorks2.
Qed.


Theorem mergeWithTranslateStateList :
  forall (a subAstd : astd)
         (states : listOfAstdStates)
         (vLi vLi' : bStateList)
         (name nameOfState : string)
         (label : operation_label)
         (opes1 opes2 opes3 : list operation),
    mergeOperationList (modifyPrePost opes3 name (get_name a)) opes1 = Some opes2 ->
    operationsAreWellConstructed opes3 ->
    operationsAreWellConstructed opes1 ->
    isAstdNamedNFromAstdList subAstd nameOfState states ->
    astdNameAreUniqueInAstdList (ConsAstdStates a states) ->
    b_step (Automata_B_state name nameOfState vLi) opes2
           label (Automata_B_state name nameOfState vLi') ->
    variableNamesAreUniqueInBState (Automata_B_state name nameOfState vLi) ->
    b_step (Automata_B_state name nameOfState vLi) opes1
     label (Automata_B_state name nameOfState vLi').
Proof.
  intros a subAstd states vLi vLi' name nameOfState label opes1 opes2 opes3.
  intros Merge WellCons3 WellCons1 IsAstdNamed AstdNamesUnique Step BNameUnique.
  destruct (mergeIsOneOrTheOther (modifyPrePost opes3 name (get_name a)) opes1 opes2 label (Automata_B_state name nameOfState vLi) (Automata_B_state name nameOfState vLi') Merge).
  apply (modifyPrePostMakesWellConstructed).
  assumption.
  assumption.
  -
    inversion AstdNamesUnique.
    destruct (H4 nameOfState).
    +
      apply (absurd_noAstdNamed_isAstdNamed nameOfState a).
      apply  (isAstdNamedNInModifyPrePost name nameOfState vLi vLi' a label  opes3 H BNameUnique).
      assumption.
    +
      apply (absurd_noAstdNamedN_isAstdNamedN_inList) with (states := states) (a := subAstd) (name := nameOfState);assumption.
  -
    assumption.
Qed.