Require Import states.
Require Import B_langage.
Require Import B_state.
Require Import B_semantics.
Require Import String.
Require Import ASTD.
Require Import ASTDstep.
Require Import theoremsOnUniqueAstdList.

Inductive listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState :
  bStateList -> bStateList -> string -> bState -> Prop :=
|ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_empty :
   forall (name : string),
     listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState EmptyBStateList EmptyBStateList name Elem_B_state
|ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem :
   forall (var val val2 : string)
          (bStateLi1 bStateLi2 bState_queue : bStateList),
     listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState
       (ConsBStateList (Automata_B_state var val2 bStateLi2) bState_queue)
       (ConsBStateList (Automata_B_state var val bStateLi1) bState_queue)
       var
       (Automata_B_state var val2 bStateLi2)
|ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isNotElem :
   forall (searchedName : string)
          (st newSt: bState)
          (bStateReplaced bState : bStateList),
     listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState
       bStateReplaced bState searchedName newSt ->
     listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState
       (ConsBStateList st bStateReplaced) (ConsBStateList st bState) searchedName newSt.

Inductive bStateAndAstdStateAreEquivalentInAstdNamed : bState -> state -> string -> Prop :=
|BStateAndAstdStateAreEquivalentInAstdNamed_elem :
   forall (name : string),
     bStateAndAstdStateAreEquivalentInAstdNamed Elem_B_state Elem_state name
|BStateAndAstdStateAreEquivalentInAstdNamed_subStateElem :
   forall (nameOfA n1 : string)
          (h : histState)
          (subState : state)
          (bSubState : bState)
          (listBSubState listBSubStateRes: bStateList),
     listOfBStatesIsEquivalentToHistory listBSubState h ->
     bStateAndAstdStateAreEquivalentInAstdNamed bSubState subState n1 ->
     listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState listBSubStateRes listBSubState n1 bSubState ->
     bStateAndAstdStateAreEquivalentInAstdNamed
       (Automata_B_state nameOfA n1 listBSubStateRes)
       (Automata_state n1 h subState)
       nameOfA
with listOfBStatesIsEquivalentToHistory : bStateList -> histState -> Prop :=
|ListOfBStatesIsEquivalentToHistory_nil :
   listOfBStatesIsEquivalentToHistory EmptyBStateList EmptyHistState
|ListOfBStatesIsEquivalentToHistory_notNil :
   forall (bst : bState)
          (st : state)
          (name : string)
          (h_queue : histState)
          (bState_queue : bStateList),
     listOfBStatesIsEquivalentToHistory bState_queue h_queue ->
     bStateAndAstdStateAreEquivalentInAstdNamed bst st name ->
     listOfBStatesIsEquivalentToHistory (ConsBStateList bst bState_queue) (ConsHistState name st h_queue).

(* Inductive astdStateAndBStateAreEquivalentInAstdNamed : state -> bState -> string -> Prop := *)
(* |AstdStateAndBStateAreEquivalentInAstdNamed_elem : *)
(*    forall (s : string), *)
(*      astdStateAndBStateAreEquivalentInAstdNamed Elem_state Elem_B_state s *)
(* |AstdStateAndBStateAreEquivalentInAstdNamed_automata : *)
(*    forall (nameAstd nameState : string) *)
(*           (h : histState) *)
(*           (listOfBState : bStateList) *)
(*           (subState : state), *)
(*      historyIsEquivalentToListOfBState h listOfBState -> *)
(*      isHistoricStateOfNFromHistory subState nameState h -> *)
(*      astdStateAndBStateAreEquivalentInAstdNamed (Automata_state nameState h subState) *)
(*                                                 (Automata_B_state nameAstd nameState *)
(*                                                                   listOfBState) nameAstd *)
(* with historyIsEquivalentToListOfBState : histState -> bStateList -> Prop := *)
(* |HistoryIsEquivalentToListOfBState_nil : *)
(*    historyIsEquivalentToListOfBState EmptyHistState EmptyBStateList *)
(* |HistoryIsEquivalentToListOfBState_notNil : *)
(*    forall (st : state) *)
(*           (bSt : bState) *)
(*           (name : string) *)
(*           (bState_queue : bStateList) *)
(*           (h_queue : histState), *)
(*      historyIsEquivalentToListOfBState h_queue bState_queue -> *)
(*      astdStateAndBStateAreEquivalentInAstdNamed st bSt name -> *)
(*      historyIsEquivalentToListOfBState (ConsHistState name st h_queue) (ConsBStateList bSt bState_queue). *)

Lemma replacingByElemBState :
  forall (li2 li : bStateList)
         (name : string),
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState li2 li name Elem_B_state ->
    li2 = li.
Proof.
  induction li2.
  -
    intros li name listReplaced.
    inversion listReplaced.
    reflexivity.
  -
    intros li name listReplaced.
    inversion listReplaced;subst.
    rewrite (IHli2 bState name H4).
    reflexivity.
Qed.

Lemma replacingByElemBState_reverse :
  forall (li : bStateList)
         (name : string),
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState li li name Elem_B_state.
Proof.
  induction li.
  -
    intro.
    apply ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_empty.
  -
    intro.
    apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isNotElem).
    apply (IHli name).
Qed.

Lemma nameOfReplacingState :
  forall (var val name : string)
         (vLi vLi' vLiSubState : bStateList),
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi name (Automata_B_state var val vLiSubState) ->
    name = var.
Proof.
  induction vLi.
  -
    intros vLi' bLiSubState Replaced.
    inversion Replaced.
  -
    intros vLi' vLiSubState Replaced.
    inversion Replaced;subst.
    +
      reflexivity.
    +
      apply (IHvLi bStateReplaced vLiSubState H4).
Qed.

Lemma existsAListWhereReplaced :
  forall (li1 li2 : bStateList)
         (a : astd)
         (s s2 : state)
         (name : string)
         (bSubState bSubState2 : bState),
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState
      li2 li1 name bSubState ->
    bStateAndAstdStateAreEquivalentInAstdNamed bSubState s name ->
    isAStateOfTheAstd s a ->
    bStateAndAstdStateAreEquivalentInAstdNamed bSubState2 s2 name ->
    isAStateOfTheAstd s2 a ->
    exists (li' : bStateList),
      listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState
        li' li1 name bSubState2.
Proof.
  induction li1.
  -
    intros li2 a s s2 name bSubState bSubState2.
    intros listReplaced stateEquiv isAState stateEquiv2 isAState2.
    inversion listReplaced;subst.
    inversion stateEquiv;subst.
    inversion isAState;subst.
    inversion isAState2;subst.
    inversion stateEquiv2;subst.
    exists EmptyBStateList.
    apply ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_empty.
  -
    intros li2 a s s2 name bSubState bSubState2 listReplaced.
    intros stateEquiv isAState stateEquiv2 isAState2.
    destruct a.
    +
      inversion isAState2;subst.
      inversion stateEquiv2;subst.
      exists (ConsBStateList b li1).
      inversion isAState;subst.
      inversion stateEquiv;subst.
      rewrite (replacingByElemBState li2 (ConsBStateList b li1) name) in listReplaced.
      assumption.
      assumption.
    +
      inversion isAState;subst.
      inversion stateEquiv;subst.
      inversion isAState2;subst.
      inversion stateEquiv2;subst.
      inversion listReplaced;subst.
      *
        exists (ConsBStateList (Automata_B_state name nameOfState0 listBSubStateRes0) li1).
        apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem).
      *
        destruct (IHli1 bStateReplaced (Automata s0 l l0 l1 l2 s1) (Automata_state nameOfState h subState) (Automata_state nameOfState0 h0 subState0) name (Automata_B_state name nameOfState listBSubStateRes) (Automata_B_state name nameOfState0 listBSubStateRes0) H16 stateEquiv isAState stateEquiv2 isAState2).
        exists (ConsBStateList b x).
        apply (ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isNotElem).
        assumption.
Qed.


Theorem variableNamesUnique_preservedByInclusion :
  forall (name : string)
         (bSubState : bState)
         (vLi listBSubState : bStateList),
    variableNamesAreUniqueInBStateList vLi ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState
      vLi
      listBSubState name bSubState ->
    variableNamesAreUniqueInBState bSubState.
Proof.
  induction vLi.
  -
    intros liRes NamesUnique ListReplaced.
    inversion ListReplaced.
    apply (VariableNamesAreUniqueInBState_elem).
  -
    intros liRes NamesUnique ListReplaced.
    inversion NamesUnique;subst.
    inversion ListReplaced;subst.
    +
      assumption.
    +
      apply (IHvLi bState);assumption.
Qed.

Lemma absurd_noVariableNamed_isListWhereReplaced :
  forall (name : string)
         (vLi vLi' : bStateList)
         (v : bState)
         (P : Prop),
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi name v ->
    v <> Elem_B_state ->
    noVariableIsNamedNInBStateList name vLi' ->
    P.
Proof.
  induction vLi.
  -
    intros vLi' v P ListReplaced NotElem NoVarNamed.
    inversion ListReplaced;subst.
    absurd (Elem_B_state = Elem_B_state);[assumption|reflexivity].
  -
    intros vLi' v P ListReplaced NotElem NoVarNamed.
    inversion ListReplaced;subst.
    +
      inversion NoVarNamed;subst.
      inversion H3.
      absurd (name = name);[assumption|reflexivity].
    +
      inversion NoVarNamed;subst.
      apply (IHvLi bStateReplaced v P H4 NotElem H2).
Qed.

Lemma absurd_noVariableNamed_isListWhereReplaced2 :
  forall (name : string)
         (vLi vLi' : bStateList)
         (v : bState)
         (P : Prop),
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi name v ->
    v <> Elem_B_state ->
    noVariableIsNamedNInBStateList name vLi ->
    P.
Proof.
  induction vLi.
  -
    intros vLi' v P ListReplaced NotElem NoVarNamed.
    inversion ListReplaced;subst.
    destruct NotElem. reflexivity.
  -
    intros vLi' v P ListReplaced NotElem NoVarNamed.
    inversion ListReplaced;subst.
    +
      inversion NoVarNamed;subst.
      inversion H3;subst.
      destruct H6. reflexivity.
    +
      inversion NoVarNamed;subst.
      apply (IHvLi bStateReplaced v P H4 NotElem H2).
Qed.

Lemma existsAHistStateEquivToList :
  forall (states : listOfAstdStates)
         (h : histState)
         (a : astd)
         (name : string)
         (s : state)
         (v : bState)
         (vLi vLi' : bStateList),
    isAHistoryStateOfAstdList h states ->
    isAstdNamedNFromAstdList a name states ->
    isAStateOfTheAstd s a ->
    listOfBStatesIsEquivalentToHistory vLi h ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi name v ->
    bStateAndAstdStateAreEquivalentInAstdNamed v s name ->
    variableNamesAreUniqueInBStateList vLi' ->
    astdNameAreUniqueInAstdList states ->
    exists (h' : histState),
      listOfBStatesIsEquivalentToHistory vLi' h' /\
      isAHistoryStateOfAstdList h' states.
Proof.
  induction states.
  -
    intros h a name s v vLi vLi' IsHist IsAstdNamed.
    inversion IsAstdNamed.
  -
    intros h a0 name s v vLi vLi' IsHist IsAstdNamed IsAState EquivToHist StateReplaced StateEquiv BNamesUnique AstdNamesUnique.
    destruct a.
    +
      inversion IsHist;subst.
      inversion IsAstdNamed;subst.
      *
        inversion IsAState;subst.
        inversion StateEquiv;subst.
        rewrite (replacingByElemBState vLi' vLi name StateReplaced).
        exists h.
        split.
        assumption.
        apply (IsAHistoryStateOfAstdList_notNil_elem).
        assumption.
      *
        inversion AstdNamesUnique;subst.
        destruct (IHstates h a0 name s v vLi vLi' H1 H3 IsAState EquivToHist StateReplaced StateEquiv BNamesUnique H4).
        destruct H.
        exists x.
        split.
        assumption.
        apply (IsAHistoryStateOfAstdList_notNil_elem).
        assumption.
    +
      inversion IsHist;subst.
      inversion H8;subst.
      inversion EquivToHist;subst.
      inversion IsAstdNamed;subst.
      *
        inversion H5;subst.
        inversion H6;subst.
        {
          inversion StateReplaced;subst.
          -
            exists (ConsHistState name s hRest).
            split.
            +
              apply (ListOfBStatesIsEquivalentToHistory_notNil);assumption.
            +
              apply (IsAHistoryStateOfAstdList_notNil_automata);assumption.
          -
            inversion BNamesUnique;subst.
            apply (absurd_noVariableNamed_isListWhereReplaced name bState_queue bStateReplaced v).
            +
              assumption.
            +
              inversion IsAState;subst.
              inversion StateEquiv;subst.
              unfold not;intros.
              discriminate.
            +
              apply (H7 name nameOfState).
              apply (IsValOfVarInState_isElem).
        }
      *
        {
          inversion StateReplaced;subst.
          -
            inversion H6;subst.
            inversion AstdNamesUnique;subst.
            destruct (H11 s0).
            +
              apply (absurd_noAstdNamed_isAstdNamed s0 (Automata s0 l l0 l1 l2 s1)).
              apply (IsAstdNamedN_automata).
              assumption.
            +
              apply (absurd_noAstdNamedN_isAstdNamedN_inList states a0 s0);assumption.
          -
            inversion BNamesUnique;subst.
            inversion AstdNamesUnique;subst.
            destruct (IHstates hRest a0 name s v bState_queue bStateReplaced H2 H5 IsAState H3 H12 StateEquiv H1 H15).
            destruct H.
            exists (ConsHistState s0 (Automata_state nameOfState h subState) x).
            split.
            +
              apply ListOfBStatesIsEquivalentToHistory_notNil;assumption.
            +
             apply IsAHistoryStateOfAstdList_notNil_automata;assumption.
        }
Qed.                  


Lemma existsABStateInBStateListEquivToAStateInHistory :
  forall (h : histState)
         (name : string)
         (s : state)
         (vLi : bStateList),
    listOfBStatesIsEquivalentToHistory vLi h ->
    isHistoricStateOfNFromHistory s name h ->
    exists v : bState,
      (bStateAndAstdStateAreEquivalentInAstdNamed v s name /\
       isAStateOfBStateList v vLi).
Proof.
  induction h.
  -
    intros name s vLi ListEquiv IsHist. inversion IsHist.
  -
    intros name s1 vLi ListEquiv IsHist.
    inversion ListEquiv;subst. inversion IsHist;subst.
    +
      exists bst. split;[assumption|apply (IsAStateOfBStateList_isElem)].
    +
      destruct (IHh name s1 bState_queue H2 H3). destruct H. exists x.
      split;[assumption|apply IsAStateOfBStateList_isNotElem;assumption].
Qed.

Lemma replacingByAStateInList :
  forall (var val : string)
         (vLi vLiSubState : bStateList),
    isAStateOfBStateList (Automata_B_state var val vLiSubState) vLi ->
    listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi vLi var (Automata_B_state var val vLiSubState).
Proof.
  intros var val vLi vLiSubState IsAState.
  induction vLi;inversion IsAState;subst.
  -
    apply ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isElem.
  -
    apply ListOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState_isNotElem.
    apply (IHvLi H1).
Qed.