Require Import ASTD.
Require Import String.


Inductive state :=
|Elem_state : state
|Automata_state : string -> histState -> state -> state
with histState :=
|EmptyHistState : histState
|ConsHistState : string -> state -> histState -> histState.

Scheme state_mut := Induction for state Sort Prop
                    with histState_mut := Induction for histState Sort Prop.

Inductive isRestOfListWithoutCoupleStringState :
  histState -> histState -> string -> state -> Prop :=
|IsRestOfListWithoutCoupleStringState_nil :
   forall (name : string)
          (s : state),
     isRestOfListWithoutCoupleStringState EmptyHistState EmptyHistState name s
|IsRestOfListWithoutCoupleStringState_isElem :
   forall (name : string)
          (s : state)
          (h_rest h : histState),
     isRestOfListWithoutCoupleStringState h_rest h name s ->
     isRestOfListWithoutCoupleStringState h_rest (ConsHistState name s h) name s
|IsRestOfListWithoutCoupleStringState_isNotElem_ordered :
   forall (searchedName name : string)
          (searchedState s : state)
          (h_res_queue h_starting_rest : histState),
     isRestOfListWithoutCoupleStringState h_res_queue h_starting_rest searchedName searchedState ->
     isRestOfListWithoutCoupleStringState (ConsHistState name s h_res_queue) (ConsHistState name s h_starting_rest) searchedName searchedState
|IsRestOfListWithoutCoupleStringState_isNotElem_notOrdered :
   forall (searchedName name : string)
          (searchedState s : state)
          (h_res_queue h_starting h_starting_rest : histState),
     isRestOfListWithoutCoupleStringState h_res_queue h_starting_rest searchedName searchedState ->
     isRestOfListWithoutCoupleStringState h_starting_rest h_starting name s ->
     isRestOfListWithoutCoupleStringState (ConsHistState name s h_res_queue) h_starting searchedName searchedState.

Inductive isAutomataState : state -> Prop :=
|IsAutomataState_def :
   forall (name : string)
          (h : histState)
          (s : state),
     isAutomataState (Automata_state name h s).

Inductive histIsHistOverLoadedByNameAndState :
  histState -> histState -> string -> state -> Prop :=
|Hist2IsHist1OverLoadedByNameAndState_nil :
   forall (name : string),
     histIsHistOverLoadedByNameAndState EmptyHistState EmptyHistState name Elem_state
|Hist2IsHist1OverLoadedByNameAndState_isElem :
   forall (name nameSubState : string)
          (ancientState newState stateSubState : state)
          (h_queue hSubState : histState),
     histIsHistOverLoadedByNameAndState
       (ConsHistState name (Automata_state nameSubState hSubState stateSubState)  h_queue)
       (ConsHistState name ancientState h_queue)
       name
       newState
|Hist2IsHist1OverLoadedByNameAndState_isNotElem :
   forall (searchedName name : string)
          (newState s : state)
          (h h' : histState),
     histIsHistOverLoadedByNameAndState h' h searchedName newState ->
     histIsHistOverLoadedByNameAndState (ConsHistState name s h') (ConsHistState name s h) searchedName newState.

Inductive isHistoryFromHistoryWithoutNameAndState : histState -> histState -> string -> state -> Prop :=
|IsHistoryFromHistoryWithoutNameAndState_isElem :
   forall (searchedName : string)
          (searchedState : state)
          (h_queue : histState),
     isHistoryFromHistoryWithoutNameAndState h_queue
                                             (ConsHistState searchedName searchedState h_queue)
                                             searchedName
                                             searchedState
|IsHistoryFromHistoryWithoutNameAndState_isNotElem :
   forall (searchedName name : string)
          (searchedState s : state)
          (h_rest h : histState),
     isHistoryFromHistoryWithoutNameAndState h_rest h searchedName searchedState ->
     isHistoryFromHistoryWithoutNameAndState (ConsHistState name s h_rest)
                                             h
                                             searchedName
                                             searchedState.

Inductive isStateNamedN : state -> string -> Prop :=
|IsStateNamedN_automata : forall (name : string)
                                 (h : histState)
                                 (s : state),
                            isStateNamedN (Automata_state name h s) name.

Inductive isAStateOfTheAstd : state -> astd -> Prop :=
|IsAStateOfTheAstd_elem :
   forall (name : string),
     isAStateOfTheAstd Elem_state (Elem name)
|IsAStateOfTheAstd_automata :
   forall (name init nameOfState : string)
          (h : histState)
          (subAstd : astd)
          (states : listOfAstdStates)
          (transitions : list transition)
          (subState : state)
          (sFinal dFinal : list string),
     isAstdNamedNFromAstdList subAstd nameOfState states ->
     isAStateOfTheAstd subState subAstd ->
     isAHistoryStateOfAstdList h states ->
     isAStateOfTheAstd (Automata_state nameOfState h subState)
                       (Automata name
                                 states
                                 transitions
                                 sFinal
                                 dFinal
                                 init)
with isAHistoryStateOfAstdList : histState -> listOfAstdStates -> Prop :=
|IsAHistoryStateOfAstdList_nil :
   isAHistoryStateOfAstdList EmptyHistState EmptyAstdStates
|IsAHistoryStateOfAstdList_notNil_elem :
   forall (name : string)
          (states_queue : listOfAstdStates)
          (h : histState),
     isAHistoryStateOfAstdList h states_queue ->
     isAHistoryStateOfAstdList h (ConsAstdStates (Elem name) states_queue)
|IsAHistoryStateOfAstdList_notNil_automata :
   forall (name init : string)
          (states_queue states : listOfAstdStates)
          (hRest : histState)
          (transitions : list transition)
          (sFinal dFinal : list string)
          (hState : state),
     isAHistoryStateOfAstdList hRest states_queue ->
     isAStateOfTheAstd hState (Automata name
                                        states
                                        transitions
                                        sFinal
                                        dFinal
                                        init) ->
     isAHistoryStateOfAstdList (ConsHistState name hState hRest)
                               (ConsAstdStates (Automata name
                                                         states
                                                         transitions
                                                         sFinal
                                                         dFinal
                                                         init)
                                               states_queue).

Lemma isAStateWithLessTrans :
  forall (name init : string)
         (states : listOfAstdStates)
         (trans : list transition)
         (sF dF : list string)
         (t : transition)
         (s : state),
    isAStateOfTheAstd s (Automata name states (t::trans) sF dF init) ->
    isAStateOfTheAstd s (Automata name states trans sF dF init).
Proof.
  intros name init states trans sF dF t s IsAState.
  inversion IsAState;subst.
  apply (IsAStateOfTheAstd_automata) with subAstd;assumption.
Qed.         
