Require Import B_semantics.
Require Import B_state.
Require Import String.
Require Import B_langage.
Require Import theoremsOnUniqueBState.

Theorem affectationInUniqueState :
  forall (var val1 val2 : string)
         (li : bStateList)
         (v : bState),
    substitution_step (Automata_B_state var val1 li) (Affectation var val2) v ->
    variableNamesAreUniqueInBState (Automata_B_state var val1 li) ->
    v = (Automata_B_state var val2 li).
Proof.
  intros var val1 val2 li v Step NameUnique.
  inversion Step;subst.
  inversion H3;subst.
  reflexivity.
  inversion NameUnique.
  apply (absurd_isStateListWhereValChanged_noVariableNamed li subStatesChanged var val2);assumption.
Qed.  


Lemma noVariableNamedPreservedByStateChanged :
  forall (v v' : bState)
         (var val name : string),
    noVariableIsNamedNInBState name v ->
    isStateWhereVarChangedForVal v' v var val ->
    noVariableIsNamedNInBState name v'.
Proof.
  apply (bState_mut (fun (v : bState) =>
                       forall (v' : bState)
                              (var val name : string),
                         noVariableIsNamedNInBState name v ->
                         isStateWhereVarChangedForVal v' v var val ->
                         noVariableIsNamedNInBState name v')
                    (fun (vLi : bStateList) =>
                       forall (vLi' : bStateList)
                              (var val name : string),
                         noVariableIsNamedNInBStateList name vLi ->
                         isStateListWhereVarChangedForVal vLi' vLi var val ->
                         noVariableIsNamedNInBStateList name vLi')).
  -
    intros v' var val name NoVarNamed StateChanged.
    inversion StateChanged.
  -
    intros var0 val0 vLi IHvLi v' var val name NoVarNamed StateChanged.
    inversion NoVarNamed;inversion StateChanged;subst.
    +
      apply (NoVariableIsNamedNInBState_automata);assumption.
    +
      apply (NoVariableIsNamedNInBState_automata);[|assumption].
      apply (IHvLi subStatesChanged var val name H2 H11).
  -
    intros vLi' var val name NoVarNamed IsStateList.
    inversion IsStateList.
  -
    intros v IHv vLi IHvLi vLi' var val name NoVarNamed ListChanged.
    inversion NoVarNamed;subst.
    inversion ListChanged;subst.
    +
      apply (NoVariableIsNamedNInBStateList_notNil);[assumption|apply (IHv stChanged var val name H3 H6)].
    +
      apply NoVariableIsNamedNInBStateList_notNil;[apply (IHvLi subStatesQueueChanged var val name H2 H6)|assumption].
Qed.

Lemma noVariableNamedPreservedBySubstitution :
  forall (sub : substitution)
         (v v' : bState)
         (var : string),
    noVariableIsNamedNInBState var v ->
    substitution_step v sub v' ->
    noVariableIsNamedNInBState var v'.
Proof.
  apply (sub_mut (fun (sub : substitution) =>
                    forall (v v' : bState)
                           (var : string),
                      noVariableIsNamedNInBState var v ->
                      substitution_step v sub v' ->
                      noVariableIsNamedNInBState var v')
                 (fun (s : selectList) =>
                    forall (v v' : bState)
                           (var : string),
                      noVariableIsNamedNInBState var v ->
                      substitution_step v (Select s) v' ->
                      noVariableIsNamedNInBState var v')).
  -
    intros s IHs v v' var NoVarNamed Step.
    apply (IHs v v' var NoVarNamed Step).
  -
    intros sub1 IHsub1 sub2 IHsub2 v v' var NoVarNamed Step.
    inversion Step;subst.
    apply (IHsub2 s2 v' var).
    +
      apply (IHsub1 v s2 var NoVarNamed H2).
    +
      assumption.
  -
    intros s s0 v v' var NoVarNamed Step.
    inversion Step;subst.
    apply (noVariableNamedPreservedByStateChanged v v' s s0);assumption.
  -
    intros v v' var NoVarNamed Step.
    inversion Step. subst. assumption.
  -
    intros v v' var NoVarNamed Step.
    inversion Step.
  -
    intros pre post IHpost s IHs v v' var NoVarNamed Step.
    inversion Step;subst.
    +
      apply (IHpost v v' var NoVarNamed H5).
    +
      apply (IHs v v' var NoVarNamed H4).
Qed.

Lemma uniquePreservedByIsStateWhereVarChangedForVal :
  forall (v v' : bState)
         (var val : string),
    variableNamesAreUniqueInBState v ->
    isStateWhereVarChangedForVal v' v var val ->
    variableNamesAreUniqueInBState v'.
Proof.
  apply (bState_mut (fun (v : bState) =>
                       forall (v' : bState)
                              (var val : string),
                         variableNamesAreUniqueInBState v ->
                         isStateWhereVarChangedForVal v' v var val ->
                         variableNamesAreUniqueInBState v')
                    (fun (vLi : bStateList) =>
                       forall (vLi' : bStateList)
                              (var val name : string),
                         variableNamesAreUniqueInBStateList vLi ->
                         noVariableIsNamedNInBStateList name vLi ->
                         isStateListWhereVarChangedForVal vLi' vLi var val ->
                         variableNamesAreUniqueInBStateList vLi' /\
                         noVariableIsNamedNInBStateList name vLi')).
  -
    intros v' var val NamesUnique StateChanged.
    inversion StateChanged;subst.
  -
    intros var val vLi IHvLi v' newVar newVal NamesUnique StateChanged.
    inversion NamesUnique;subst.
    inversion StateChanged;subst.
    +
      apply (VariableNamesAreUniqueInBState_automata  newVar newVal vLi H1 H3).
    +
      destruct (IHvLi subStatesChanged newVar newVal var H3 H1 H7).
      apply (VariableNamesAreUniqueInBState_automata  var val subStatesChanged H0 H).
  -
    intros vLi' var val name NamesUnique NoVarNamed IsListChanged.
    inversion IsListChanged.
  -
    intros v IHv vLi IHvLi vLi' var val name NamesUnique NoVarNamed IsListChanged.
    inversion NamesUnique;subst.
    inversion IsListChanged;subst.
    +
      inversion NoVarNamed;subst.
      split.
      *
        {
          apply (VariableNamesAreUniqueInBStateList_notNil).
          -
            assumption.
          -
            intros var1 val1 IsValOfVar.
            destruct (existsAValPreservedByStateChanged v stChanged var1 val1 var val IsValOfVar H8).
            apply (H2 var1 x H).
          -
            intros var1 val1 IsValOfVar.
            apply (noVariableNamedPreservedByStateChanged v stChanged var val).
            apply (H3 var1 val1 IsValOfVar).
            assumption.
          -
            apply (IHv stChanged var val H4 H8).
        }
      *
        apply (NoVariableIsNamedNInBStateList_notNil);[assumption|apply (noVariableNamedPreservedByStateChanged v stChanged var val name H7 H8)].
    +
      inversion NoVarNamed;subst.
      split.
      *
        destruct (IHvLi subStatesQueueChanged var val name H1 H6 H8).
        apply (VariableNamesAreUniqueInBStateList_notNil);[assumption| | |assumption].
        intros var1 val1 IsValOfVar.
        apply (IHvLi subStatesQueueChanged var val var1 H1).
        apply (H2 var1 val1  IsValOfVar).
        assumption.
        intros val1 var1 IsValOfVar.
        destruct (existsAValInListPreservedByStateChanged vLi subStatesQueueChanged var1 val1 var val IsValOfVar H8).
        apply (H3 x var1 H5).
      *
        destruct (IHvLi subStatesQueueChanged var val name H1 H6 H8).
        apply NoVariableIsNamedNInBStateList_notNil;assumption.
Qed.        


Theorem uniquePreservedBySubstitution :
  forall (sub : substitution)
         (v v' : bState),
    variableNamesAreUniqueInBState v ->
    substitution_step v sub v' ->
    variableNamesAreUniqueInBState v'.
Proof.
  apply (sub_mut (fun (sub : substitution) =>
                    forall (v v' : bState),
                      variableNamesAreUniqueInBState v ->
                      substitution_step v sub v' ->
                      variableNamesAreUniqueInBState v')
                 (fun (s : selectList) =>
                    forall (v v' : bState),
                      variableNamesAreUniqueInBState v ->
                      substitution_step v (Select s) v' ->
                      variableNamesAreUniqueInBState v')).
  -
    intros s IHs v v' NamesUnique Step.
    apply (IHs v v' NamesUnique Step).
  -
    intros sub1 IHsub1 sub2 IHsub2 v v' NamesUnique Step.
    inversion Step;subst.
    apply (IHsub2 s2 v');[|assumption].
    apply (IHsub1 v s2 NamesUnique H2).
  -
    intros var val v v' NamesUnique Step.
    inversion Step;subst.
    apply (uniquePreservedByIsStateWhereVarChangedForVal v v' var val NamesUnique H3).
  -
    intros v v' NamesUnique Step.
    inversion Step;subst.
    assumption.
  -
    intros v v' NamesUnique Step.
    inversion Step.
  -
    intros pre post IHpost s IHs v v' NamesUnique Step.
    inversion Step;subst.
    +
      apply (IHpost v v' NamesUnique H5).
    +
      apply (IHs v v' NamesUnique H4).
Qed.

Lemma affectationInUniqueStateIsDeterministic :
  forall (v v' v'' : bState)
         (var val : string),
    isStateWhereVarChangedForVal v' v var val ->
    isStateWhereVarChangedForVal v'' v var val ->
    variableNamesAreUniqueInBState v ->
    v' = v''.
Proof.
  apply (bState_mut (fun (v : bState) =>
                       forall (v' v'' : bState)
                              (var val : string),
                         isStateWhereVarChangedForVal v' v var val ->
                         isStateWhereVarChangedForVal v'' v var val ->
                         variableNamesAreUniqueInBState v ->
                         v' = v'')
                    (fun (vLi : bStateList) =>
                       forall (vLi' vLi'' : bStateList)
                              (var val : string),
                         isStateListWhereVarChangedForVal vLi' vLi var val ->
                         isStateListWhereVarChangedForVal vLi'' vLi var val ->
                         variableNamesAreUniqueInBStateList vLi ->
                         vLi' = vLi'')).
  -
    intros v' v'' var val IsState.
    inversion IsState.
  -
    intros s s0 vLi IHvLi v' v'' var val IsState1 IsState2 NamesUnique.
    inversion NamesUnique;subst.
    inversion IsState1;inversion IsState2;subst.
    +
      reflexivity.
    +
      apply (absurd_isStateListWhereValChanged_noVariableNamed vLi subStatesChanged var val);assumption.
    +
      apply (absurd_isStateListWhereValChanged_noVariableNamed vLi subStatesChanged var val);assumption.
    +
      rewrite (IHvLi subStatesChanged subStatesChanged0 var val H7 H14 H3).
      reflexivity.
  -
    intros vLi' vLi'' var val IsStateList.
    inversion IsStateList.
  -
    intros v IHv vLi IHvLi vLi' vLi'' var val IsState1 IsState2 NamesUnique.
    inversion NamesUnique;subst.
    inversion IsState1;inversion IsState2;subst.
    +
      rewrite (IHv stChanged stChanged0 var val H8 H14 H4).
      reflexivity.
    +
      apply (absurd_isStateListWhereValChanged_noVariableNamed vLi subStatesQueueChanged var val).
      destruct (aChangedVariableIsInState v stChanged var val H8).
      apply (H2 var x H).
      assumption.
    +
      apply (absurd_isStateListWhereValChanged_noVariableNamed vLi subStatesQueueChanged var val).
      destruct (aChangedVariableIsInState v stChanged var val H14).
      apply (H2 var x H).
      assumption.
    +
      rewrite (IHvLi subStatesQueueChanged subStatesQueueChanged0 var val H8 H14 H1).
      reflexivity.
Qed.

Lemma substitutionWithoutSelectIsDeterministic :
  forall (sub : substitution)
         (v v' v'' : bState),
    substitution_step v sub v' ->
    substitution_step v sub v'' ->
    substitutionContainsNoSelect sub ->
    variableNamesAreUniqueInBState v ->
    v' = v''.
Proof.
  induction sub.
  -
    intros v v' v'' Step1 Step2 NoSelect.
    inversion NoSelect.
  -
    intros v v' v'' Step1 Step2 NoSelect NamesUnique.
    inversion NoSelect;subst.
    inversion Step1;subst.
    inversion Step2;subst.
    assert (s2 = s0).
    +
      apply (IHsub1 v s2 s0 H4 H5 H1 NamesUnique).
    +
      subst.
      apply (IHsub2 s0 v' v'' H6 H8 H2).
      apply (uniquePreservedBySubstitution sub1 v s0 NamesUnique H5).
  -
    intros v v' v'' Step1 Step2 NoSelect NamesUnique.
    inversion Step1;subst.
    inversion Step2;subst.
    apply (affectationInUniqueStateIsDeterministic v v' v'' s s0 H3 H4 NamesUnique).
  -
    intros v v' v'' Step1 Step2 NoSelect NamesUnique.
    inversion Step1;subst.
    inversion Step2;subst.
    reflexivity.
Qed.

Lemma substitutionPreserveSubStateStructure1 :
  forall (sub : substitution)
         (var val1 val2 : string)
         (v : bState)
         (vLi vLi' : bStateList),
    substitution_step (Automata_B_state var val1 (ConsBStateList v vLi)) sub (Automata_B_state var val2 vLi') ->
    substitutionContainsNoSelect sub ->
    (forall (name : string),
       variableIsUsedAndModifiedInSubstitution name sub ->
       (noVariableIsNamedNInBStateList name vLi /\ var <> name)) ->
    exists (v'' : bState),
      vLi' = ConsBStateList v'' vLi /\
      substitution_step v sub v'' /\
      val1 = val2.
Proof.
  induction sub.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect.
    inversion NoSelect.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    inversion NoSelect;subst.
    destruct (substitutionDoesNotChangeStateStructure sub1 var val1 (ConsBStateList v vLi) s2 H2).
    destruct H.
    subst.
    destruct (IHsub1 var val1 x v vLi x0 H2 H1).
    intros name VarMod.
    apply (NoVarNamed name).
    apply (VariableIsUsedAndModifiedInSubstitution_parallel_left).
    assumption.
    destruct H.
    destruct H0.
    subst.
    destruct (IHsub2 var x val2 x1 vLi vLi' H4 H3).
    intros name VarMod.
    apply (NoVarNamed name).
    apply (VariableIsUsedAndModifiedInSubstitution_parallel_right).
    assumption.
    destruct H.
    destruct H5.
    subst.
    exists x0.
    split;[reflexivity|split;[|reflexivity]].
    apply (Substitution_step_parallel) with x1;assumption.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    destruct (NoVarNamed s).
    apply (VariableIsUsedAndModifiedInSubstitution_affectation).
    inversion H3;subst.
    +
      destruct H0.
      reflexivity.
    +
      inversion H9;subst.
      *
        exists stChanged.
        split;[reflexivity|split;[|reflexivity]].
        apply (Substitution_step_affectation).
        assumption.
      *
        apply (absurd_isStateListWhereValChanged_noVariableNamed vLi subStatesQueueChanged s s0);assumption.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    exists v.
    split;[reflexivity|split;[apply Substitution_step_skipSubstitution|reflexivity]].
Qed.                  

Lemma substitutionPreserveSubStateStructure2 :
  forall (sub : substitution)
         (var val1 val2 : string)
         (v : bState)
         (vLi vLi' : bStateList),
    substitution_step (Automata_B_state var val1 (ConsBStateList v vLi)) sub (Automata_B_state var val2 vLi') ->
    substitutionContainsNoSelect sub ->
    (forall (name : string),
       variableIsUsedAndModifiedInSubstitution name sub ->
       (noVariableIsNamedNInBState name v /\ var <> name)) ->
    exists (vLi'' : bStateList),
      vLi' = ConsBStateList v vLi'' /\
      val1 = val2.
Proof.
  induction sub.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion NoSelect.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    inversion NoSelect;subst.
    destruct (substitutionDoesNotChangeStateStructure sub1 var val1 (ConsBStateList v vLi) s2 H2).
    destruct H.
    subst.
    destruct (IHsub1 var val1 x v vLi x0 H2 H1).
    intros name VarMod.
    apply (NoVarNamed name).
    apply (VariableIsUsedAndModifiedInSubstitution_parallel_left).
    assumption.
    destruct H.
    subst.
    apply (IHsub2 var x val2 v x1 vLi' H4 H3).
    intros name VarMod.
    apply (NoVarNamed name).
    apply (VariableIsUsedAndModifiedInSubstitution_parallel_right).
    assumption.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    destruct (NoVarNamed s).
    apply (VariableIsUsedAndModifiedInSubstitution_affectation).
    inversion H3;subst.
    +
      destruct H0.
      reflexivity.
    +
      inversion H9;subst.
      *
        apply (absurd_isStateWhereValChanged_noVariableNamed v stChanged s s0);assumption.
      *
        exists subStatesQueueChanged.
        split;reflexivity.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    exists vLi.
    split;reflexivity.
Qed.

Lemma substitutionInASmallerState :
  forall (sub : substitution)
         (var val1 val2 : string)
         (v : bState)
         (vLi vLi' : bStateList),
    substitution_step (Automata_B_state var val1 (ConsBStateList v vLi))
                      sub
                      (Automata_B_state var val2 (ConsBStateList v vLi')) ->
    substitutionContainsNoSelect sub ->
    (forall (name : string),
       variableIsUsedAndModifiedInSubstitution name sub ->
       noVariableIsNamedNInBState name v /\ var <> name) ->
    substitution_step (Automata_B_state var val1 vLi) sub
                      (Automata_B_state var val2 vLi').
Proof.
  induction sub.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect.
    inversion NoSelect.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    destruct (substitutionDoesNotChangeStateStructure sub1 var val1 (ConsBStateList v vLi) s2 H2).
    destruct H.
    subst.
    inversion NoSelect;subst.
    destruct (substitutionPreserveSubStateStructure2 sub1 var val1 x v vLi x0 H2 H1).
    intros name VarMod.
    apply (NoVarNamed name).
    apply (VariableIsUsedAndModifiedInSubstitution_parallel_left).
    assumption.
    destruct H.
    subst.
    apply (Substitution_step_parallel) with (Automata_B_state var x x1).
    +
      apply (IHsub1 var x x v vLi x1 H2 H1).
      intros name VarMod.
      apply (NoVarNamed name).
      apply (VariableIsUsedAndModifiedInSubstitution_parallel_left).
      assumption.
    +
      apply (IHsub2 var x val2 v x1 vLi' H4 H3).
      intros name VarMod.
      apply (NoVarNamed name).
      apply (VariableIsUsedAndModifiedInSubstitution_parallel_right).
      assumption.                         
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    destruct (NoVarNamed s).
    apply (VariableIsUsedAndModifiedInSubstitution_affectation).
    inversion H3;subst.
    +
      destruct H0. reflexivity.
    +
      inversion H9;subst.
      *
        destruct (aChangedVariableIsInState v v s s0 H8).
        apply (absurd_isValOfVar_noVariableNamed v s x);assumption.
      *
        apply (Substitution_step_affectation).
        apply (IsStateWhereVarChangedForVal_isNotElem).
        assumption.
  -
    intros var val1 val2 v vLi vLi' Step NoSelect NoVarNamed.
    inversion Step;subst.
    apply (Substitution_step_skipSubstitution).
Qed.
