Require Import ASTD.
Require Import String.
Require Import states.
Require Import ASTDstep.

Lemma nameUniqueWithLessTrans :
  forall (name init : string)
         (states : listOfAstdStates)
         (trans : list transition)
         (sF dF : list string)
         (t : transition),
    astdNameAreUniqueInAstd (Automata name states (t::trans) sF dF init) ->
    astdNameAreUniqueInAstd (Automata name states trans sF dF init).
Proof.
  intros name init states trans sF dF t NamesUnique.
  inversion NamesUnique;subst.
  apply (AstdNameAreUnique_automata);assumption.
Qed.

Theorem absurd_noAstdNamed_isAstdNamed :
  forall (name : string)
         (a : astd)
         (P : Prop),
    isAstdNamedN a name ->
    noAstdIsNamedNInAstd name a ->
    P.
Proof.
  intros.
  destruct a;inversion H;inversion H0;subst;absurd (name = name);[assumption|reflexivity|assumption|reflexivity].
Qed.

Theorem absurd_noAstdNamedN_isAstdNamedN_inList :
  forall (states : listOfAstdStates)
         (a : astd)
         (name : string)
         (P : Prop),
    isAstdNamedNFromAstdList a name states ->
    noAstdIsNamedNInSubAstdList name states ->
    P.
Proof.
  induction states.
  -
    intros a name P isAstdNamed.
    inversion isAstdNamed.
  -
    intros a' name P isAstdNamed noAstdNamed.
    inversion isAstdNamed;inversion noAstdNamed;subst.
    +
      apply (absurd_noAstdNamed_isAstdNamed name a P);assumption.
    +
      apply (IHstates a' name P H2 H8).
Qed.

Theorem isAstdNamedN_uniqueInAstdList :
  forall (a1 a2 : astd)
         (states : listOfAstdStates)
         (name : string),
    isAstdNamedNFromAstdList a1 name states ->
    isAstdNamedNFromAstdList a2 name states ->
    astdNameAreUniqueInAstdList states ->
    a1 = a2.
Proof.
  induction states.
  -
    intros name.
    intros IsAstdNamed.
    inversion IsAstdNamed.
  -
    intros name IsAstdNamed1 IsAstdNamed2 NameUnique.
    inversion IsAstdNamed1;inversion IsAstdNamed2;subst.
    +
      reflexivity.
    +
      inversion NameUnique;subst.
      destruct (H4 name).
      *
        apply (absurd_noAstdNamed_isAstdNamed name a (a = a2) H2 H).
      *
        apply (absurd_noAstdNamedN_isAstdNamedN_inList states a2 name (a = a2) H7 H ).
    +
      inversion NameUnique;subst.
      destruct (H4 name).
      *
        apply (absurd_noAstdNamed_isAstdNamed name a (a1 = a) H7 H).
      *
        apply (absurd_noAstdNamedN_isAstdNamedN_inList states a1 name (a1 = a) H2 H).
    +
      inversion NameUnique.
      apply (IHstates name H2 H7 H3).
Qed.

Theorem astdNamesUniqueInASubAstd :
  forall (states : listOfAstdStates)
         (name : string)
         (a : astd),
    isAstdNamedNFromAstdList a name states ->
    astdNameAreUniqueInAstdList states ->
    astdNameAreUniqueInAstd a.
Proof.
  induction states.
  -
    intros name a.
    intros IsAstdNamed.
    inversion IsAstdNamed.
  -
    intros name a0.
    intros IsAstdNamed NameUnique.
    inversion NameUnique;subst.
    inversion IsAstdNamed;subst.
    +
      assumption.
    +
      apply (IHstates name a0 H5 H2).
Qed.    

Theorem subAstdWellDefined :
  forall (states : listOfAstdStates)
         (name : string)
         (a : astd),
    isAstdNamedNFromAstdList a name states ->
    isAstdListWellDefined states ->
    isWellDefined a.
Proof.
  induction states.
  -
    intros name a.
    intros IsAstdNamed.
    inversion IsAstdNamed.
  -
    intros name a0.
    intros IsAstdNamed NameUnique.
    inversion NameUnique;subst.
    inversion IsAstdNamed;subst.
    +
      assumption.
    +
      apply (IHstates name a0 H4 H1).
Qed.    


Theorem noAstdNamedInASubState :
  forall (a : astd)
         (name name' : string)
         (states : listOfAstdStates),
    noAstdIsNamedNInSubAstdList name states ->
    isAstdNamedNFromAstdList a name' states ->
    noAstdIsNamedNInAstd name a.
Proof.
  induction states.
  +
    intros NoAstdNamed IsAstdNamed.
    inversion IsAstdNamed.
  +
    intros NoAstdNamed IsAstdNamed.
    inversion NoAstdNamed;subst.
    inversion IsAstdNamed;subst;[assumption|].
    apply (IHstates H3 H4).
Qed.

Lemma noAstdNamedNInSubQueue :
  forall (states states2 : listOfAstdStates)
         (name : string),
    noAstdIsNamedNInSubAstdList name states ->
    isASubQueueOfStates states2 states ->
    noAstdIsNamedNInSubAstdList name states2.
Proof.
  induction states.
  -
    intros states2 name NoAstdNamed IsASubQueue.
    inversion IsASubQueue;subst.
    apply (NoAstdIsNamedNInSubAstdList_empty).
  -
    intros states2 name NoAstdNamed IsASubQueue.
    inversion IsASubQueue;subst;[assumption|].
    inversion NoAstdNamed;subst.
    apply (IHstates states2 name H4 H1).
Qed.

Lemma historicStateInHInitIsInitial :
  forall (states : listOfAstdStates)
         (h : histState)
         (name : string)
         (a : astd)
         (s : state),
    isHistoricStateOfNFromHistory s name h ->
    isAstdNamedNFromAstdList a name states ->
    isHInit h states ->
    isAutomata a ->
    astdNameAreUniqueInAstdList states ->
    isInitialStateOfAstd s a.
Proof.
  induction states.
  -
    intros h name a s IsHist IsAstdNamed IHInit.
    inversion IsAstdNamed.
  -
    intros h name a0 s IsHist IsAstdNamed IsHInit IsAuto NamesUnique.
    inversion NamesUnique;subst.
    inversion IsHInit;inversion IsAstdNamed;subst.
    +
      inversion IsAuto.
    +
      apply (IHstates h name a0 s IsHist H9 H4 IsAuto H2).
    +
      inversion H7;subst. inversion H4;subst. inversion H12;subst.
      inversion IsHist;subst. assumption.
      destruct (H3 name). inversion H. destruct H17. reflexivity.
      apply (absurd_noAstdNamed_isHistoricState states hInit_rest name s).
      assumption.
      apply (HInitIsAnHistoricState states hInit_rest H8).
      assumption.
    +
      inversion IsHist;subst.
      *
        destruct (H3 name0). destruct a;inversion H4;subst;inversion H;subst;[destruct H9;reflexivity| destruct H16;reflexivity].
      apply (absurd_noAstdNamedN_isAstdNamedN_inList states a0 name0);assumption.
      *
        apply (IHstates hInit_rest name a0 s);assumption.
Qed.    
