Require Import states.
Require Import B_state.
Require Import ASTD.
Require Import B_langage.
Require Import String.
Require Import translation.
Require Import B_semantics.
Require Import mergeWorks.
Require Import Theorems.
Require Import modifyPrePostWorks.
Require Import stateEquivalence.
Require Import wellConstructed.
Require Import theoremsOnUniqueAstdList.
Require Import theoremsOnNames.
Require Import theoremsOnUniqueBState.

Lemma translateStateListMakesProperState :
  forall (states : listOfAstdStates)
         (nameOfA nameOfState : string)
         (opes : list operation)
         (vLi : bStateList)
         (v' : bState)
         (l : operation_label),
    translateStateList states nameOfA = Some opes ->
    variableNamesAreUniqueInBState (Automata_B_state nameOfA nameOfState vLi) ->
    astdNameAreUniqueInAstdList states ->
    b_step (Automata_B_state nameOfA nameOfState vLi) opes l v' ->
      exists (vLi' : bStateList)
             (vSub : bState),
        listOfBStateIsListOfBStateWhereStateNamedNIsReplacedByState vLi' vLi nameOfState vSub /\
        v' = (Automata_B_state nameOfA nameOfState vLi').
Proof.
  induction states.
  -
    intros name nameOfState opes vLi v'  l TranslateList BNamesUnique AstdNamesUnique Step.
    simpl in TranslateList.
    inversion TranslateList;subst.
    inversion Step.
    inversion H1.
  -
    intros nameOfA nameOfState opes vLi v' label.
    intros Translate BNameUnique AstdNamesUnique Step.
    simpl in Translate.
    case_eq (translateAstd a);intros;rewrite H in Translate;[|discriminate].
    case_eq (translateStateList states nameOfA);intros;rewrite H0 in Translate;[|discriminate].
      Admitted.

Theorem aStateIsNamedNIfTranslateStateListStep :
  forall (states : listOfAstdStates)
         (opes : list operation)
         (name nameOfState : string)
         (l : operation_label)
         (v v' : bState),
    b_step v opes l v'  ->
    translateStateList states name = Some opes ->
    isValOfVarInState nameOfState name v ->
    variableNamesAreUniqueInBState v ->
    exists (a : astd),
      isAstdNamedNFromAstdList a nameOfState states.
Proof.
  induction states.
  -
    intros opes name nameOfState l v v' step translateList.
    simpl in translateList.
    inversion translateList;subst.
    inversion step.
    inversion H1.
  -
    intros opes name nameOfState l v v' step translateList isValOfVar nameUnique.
    simpl in translateList.
    case_eq (translateAstd a);intros;rewrite H in translateList;[|discriminate].
    case_eq (translateStateList states name);intros;rewrite H0 in translateList;[|discriminate].
    destruct (mergeIsOneOrTheOther (modifyPrePost l0 name (get_name a)) l1 opes l v v' translateList).
    apply (modifyPrePostMakesWellConstructed).
    apply (translateStateListIsWellConstructed states name l1 H0).
    assumption.
    +
      rewrite (isValOfVarInState_uniqueInBState v name nameOfState (get_name a)).
      exists a.
      apply (IsAstdNamedNFromAstdList_isElem).
      apply (get_nameWorks2).
      assumption.
      apply (isValOfVar_modifyPrePost l0 name (get_name a) l v v' H1).
      assumption.
    +
      destruct (IHstates l1 name nameOfState l v v' H1 H0 isValOfVar nameUnique).
      exists x.
      apply (IsAstdNamedNFromAstdList_isNotElem).
      assumption.
Qed.

Theorem translateStateListWithMerge :
  forall (a : astd)
         (name nameOfState : string)
         (states : listOfAstdStates)
         (vLi' vLi : bStateList)
         (l : operation_label)
         (opes1 opes2 opes : list operation),
    translateStateList states name = Some opes1 ->
    mergeOperationList opes2 opes1 = Some opes ->
    operationsAreWellConstructed opes2 ->
    b_step (Automata_B_state name nameOfState vLi) opes l (Automata_B_state name nameOfState vLi') ->
    isAstdNamedN a nameOfState ->
    astdNameAreUniqueInAstdList (ConsAstdStates a states) ->
    variableNamesAreUniqueInBState (Automata_B_state name nameOfState vLi) ->
    b_step (Automata_B_state name nameOfState vLi) opes2 l (Automata_B_state name nameOfState vLi').
Proof.
  intros a name nameOfState states vLi' vLi l opes1 opes2 opes.
  intros Translate Merge WellCons Step isAstdNamed AstdNamesUnique VarNamesUnique.
  destruct (mergeIsOneOrTheOther opes2 opes1 opes l (Automata_B_state name nameOfState vLi)  (Automata_B_state name nameOfState vLi') Merge WellCons).
  apply (translateStateListIsWellConstructed states name opes1 Translate).
  assumption.
  -
    assumption.
  -
    destruct (aStateIsNamedNIfTranslateStateListStep states opes1 name nameOfState l (Automata_B_state name nameOfState vLi) (Automata_B_state name nameOfState vLi') H Translate).
    apply IsValOfVarInState_isElem.
    assumption.
    inversion AstdNamesUnique.
    destruct (H5 nameOfState);subst.
    +
      destruct a;inversion H6;subst;inversion isAstdNamed;subst;absurd (nameOfState = nameOfState);[assumption|reflexivity|assumption|reflexivity].
    +
      apply absurd_noAstdNamedN_isAstdNamedN_inList with (states := states) (a :=  x) (name :=  nameOfState);assumption.
Qed.