type bool =
| True
| False

type 'a option =
| Some of 'a
| None

type ('a, 'b) prod =
| Pair of 'a * 'b

type 'a list =
| Nil
| Cons of 'a * 'a list

type sumbool =
| Left
| Right

val bool_dec : bool -> bool -> sumbool

val in_dec : ('a1 -> 'a1 -> sumbool) -> 'a1 -> 'a1 list -> sumbool

type ascii =
| Ascii of bool * bool * bool * bool * bool * bool * bool * bool

val ascii_dec : ascii -> ascii -> sumbool

type string =
| EmptyString
| String of ascii * string

val string_dec : string -> string -> sumbool

type guard =
  string
  (* singleton inductive, whose constructor was Gu *)

type transition_label =
| Transition_label of string * (string, string) prod list

type transition =
| Local of string * string * transition_label * guard * bool
| From_sub of string * string * string * transition_label * guard * bool
| To_sub of string * string * string * transition_label * guard * bool

type astd =
| Elem of string
| Automata of string * listOfAstdStates * transition list * string list
   * string list * string
and listOfAstdStates =
| EmptyAstdStates
| ConsAstdStates of astd * listOfAstdStates

type operation_label =
| Label of string * (string, string) prod list

type predicate =
| Equality of string * string
| And of predicate * predicate
| Or of predicate * predicate
| TruePredicate
| FalsePredicate
| IsIn of string * string list
| Guard of string

type substitution =
| Select of selectList
| Parallel of substitution * substitution
| Affectation of string * string
| SkipSubstitution
and selectList =
| EmptySelectList
| ConsSelectList of predicate * substitution * selectList

type operation =
| Operation of operation_label * predicate * substitution

val get_name : astd -> string

val finalPredicate : astd -> predicate

val makeDFinalPred : string list -> listOfAstdStates -> string -> predicate

val getAstdFromStateList : string -> listOfAstdStates -> astd option

val initializeAllVariable : astd -> substitution

val initializeFromAstdList : listOfAstdStates -> substitution

val getGuard : guard -> string

val makePrecondition :
  transition -> string -> listOfAstdStates -> predicate option

val makePostcondition :
  transition -> string -> listOfAstdStates -> substitution option

val getTransitionLabel : transition -> transition_label

val transformTransitionLabel : transition_label -> operation_label

val paramAreEquals :
  (string, string) prod list -> (string, string) prod list -> bool

val labelsAreEqual : operation_label -> transition_label -> bool

val addTransition :
  transition -> operation list -> string -> listOfAstdStates -> operation
  list option

val translateAllTransitions :
  transition list -> operation list -> string -> listOfAstdStates ->
  operation list option

val opLabelsAreEqual : operation_label -> operation_label -> bool

val addOperation : operation -> operation list -> operation list option

val mergeOperationList :
  operation list -> operation list -> operation list option

val modifyPrePost : operation list -> string -> string -> operation list

val translateAstd : astd -> operation list option

val translateStateList : listOfAstdStates -> string -> operation list option

