Require Import ASTD.
Require Import B_semantics.
Require Import B_langage.
Require Import String.
Require Import List.
Require Import Bool.
Import ListNotations.

Definition get_name (a : astd) : string :=
  match a with
    |Elem name => name
    |Automata name _ _ _ _ _ => name
  end.

Fixpoint isInStringList st stringList :=
  match stringList with
    |nil => false
    |he::ta => if string_dec st he
               then true
               else isInStringList st ta
  end.

Fixpoint getAstdNamedNFromAstdList searchedName li :=
  match li with
    |EmptyAstdStates => None
    |ConsAstdStates a queue =>
     if string_dec (get_name a) searchedName
     then Some a
     else getAstdNamedNFromAstdList searchedName queue
  end.

Fixpoint finalPredicate a :=
  match a with
    |Elem e => TruePredicate
    |Automata name states transitions sFinal dFinal init =>
     match dFinal with
       |nil => match sFinal with
                 |nil => FalsePredicate
                 |_::_ => IsIn name sFinal
               end
       |he::ta =>
        let dFinalPred := makeDFinalPred dFinal states name in
        match sFinal with
          |nil => dFinalPred
          |_ :: _ => Or (IsIn name sFinal)
                        dFinalPred
        end
     end
  end
with makeDFinalPred dFinal states name :=
       match states with
         |EmptyAstdStates => FalsePredicate
         |ConsAstdStates he ta =>
          let finalPredQueue := makeDFinalPred dFinal ta name in
          let nameOfHead := get_name he in
          if in_dec string_dec nameOfHead dFinal
          then Or (And (Equality name nameOfHead)
                       (finalPredicate he))
                  finalPredQueue
          else finalPredQueue
       end.

Fixpoint getAstdFromStateList name astdList :=
  match astdList with
    |EmptyAstdStates => None
    |ConsAstdStates he ta =>
     if string_dec (get_name he) name
     then Some he
     else getAstdFromStateList name ta
  end.            

Function get_state_list astd :=
  match astd with
    |Elem n => None
    |Automata _ states _ _ _ _ => Some states
  end.

Fixpoint initializeAllVariable astd :=
  match astd with
    |Elem e => SkipSubstitution
    |Automata name states _ _ _ init =>
     Parallel (Affectation name init)
              (initializeFromAstdList states)
  end
with initializeFromAstdList statesList :=
       match statesList with
         |EmptyAstdStates => SkipSubstitution
         |ConsAstdStates headAstd queueAstdList =>
          Parallel (initializeAllVariable headAstd)
                   (initializeFromAstdList queueAstdList)
       end.

Definition getGuard gu :=
  match gu with
    |Gu g => g
  end.

Definition makePrecondition transition nameOfTranslatedAstd listOfStates :=
  match transition with
    |Local fromState toState astdTransition astdPredicate finale =>
     let cAll := Equality nameOfTranslatedAstd fromState in
     let cGuard := Guard (getGuard astdPredicate) in
     (match getAstdFromStateList fromState listOfStates with
        |None => None
        |Some fromAstd => if finale then
                            let cFinal := finalPredicate fromAstd in
                            Some (And (And cAll cGuard)
                                      cFinal)
                          else
                            Some (And cAll cGuard)
      end)
    |To_sub fromState toState throughState astdTransition astdPredicate finale =>
     let cAll := Equality nameOfTranslatedAstd fromState in
     let cGuard := Guard (getGuard astdPredicate) in
     (match getAstdFromStateList fromState listOfStates with
          |None => None
          |Some fromAstd => if (finale) then
                              let cFinal := finalPredicate fromAstd in
                              Some (And (And cAll cGuard)
                                        cFinal)
                            else
                              Some (And cAll cGuard)
        end)
    |From_sub fromState toState throughState astdTransition astdPredicate finale =>
     let cAll := Equality nameOfTranslatedAstd throughState in
     let cGuard := Guard (getGuard astdPredicate) in
     let cToSub := Equality throughState fromState in
       (match getAstdFromStateList throughState listOfStates with
          |None => None
          |Some throughAstd =>
           (match throughAstd with
              |Elem s => None
              |Automata nameThroughState listOfSubStates _ _ _ _ =>
               (match getAstdFromStateList fromState listOfSubStates with
                  |None => None
                  |Some fromAstd =>
                   if (finale) then
                     let cFinale := finalPredicate fromAstd in
                     Some (And (And (And cAll cGuard)
                                    cToSub)
                               cFinale)
                   else
                     Some (And (And cAll cGuard) cToSub)
                end)
            end)
        end)
  end.

Definition makePostcondition transition nameOfTranslatedAstd listOfStates :=
  match transition with
    |Local fromState toState sigma gu finale =>
     let tAll := Affectation nameOfTranslatedAstd toState in
     match getAstdFromStateList toState listOfStates with
       |None => None
       |Some a => let tInitN2 := initializeAllVariable a in
                  Some (Parallel tAll tInitN2)
     end
    |From_sub fromState toState throughState sigma gu finale =>
     let tAll := Affectation nameOfTranslatedAstd toState in
     match getAstdFromStateList toState listOfStates with
       |None => None
       |Some a => let tInitN2 := initializeAllVariable a in
                  Some (Parallel tAll tInitN2)
     end
    |To_sub fromState toState throughState sigma gu final =>
     let tAll := Affectation nameOfTranslatedAstd throughState in
     let tToSub := Affectation throughState toState in
     match getAstdFromStateList throughState listOfStates with
       |None => None
       |Some (Elem a) => None
       |Some (Automata nameThroughState listThroughState _ _ _ _) =>
        if (string_dec toState "H*") then
          Some tAll
        else
          if (string_dec toState "H") then
            let initHInit := initializeFromAstdList listThroughState in
            Some (Parallel tAll initHInit)
          else
            match getAstdFromStateList toState listThroughState with
              |None => None
              |Some a => let tInitFrom := initializeAllVariable a in
                         Some (Parallel (Parallel tAll tToSub)
                                        tInitFrom)
            end
     end
  end.

Definition getTransitionLabel tr :=
  match tr with
    |Local _ _ la _ _ => la
    |From_sub _ _ _ la _ _ => la
    |To_sub _ _ _ la _ _ => la
  end.

Definition transformTransitionLabel l :=
  match l with
    |Transition_label name param => Label name param
  end.

Fixpoint paramAreEquals param1 param2 :=
  match (param1,param2) with
    |([],[]) => true
    |([],_::_) => false
    |(_::_,[]) => false
    |((p1,p1T)::qParam1,(p2,p2T)::qParam2) =>
     if (string_dec p1 p2)
     then
       if (string_dec p1T p2T) then
         paramAreEquals qParam1 qParam2
       else
         false
     else
       false
  end.

Definition labelsAreEqual label1 label2 :=
match (label1,label2) with
  |((Label name1 param1),(Transition_label name2 param2)) =>
   if (string_dec name1 name2) then
     paramAreEquals param1 param2
   else
     false
end.



Fixpoint addTransition (tr : transition) (li : list operation) (name : string) (s : listOfAstdStates) :=
  match (makePrecondition tr name s, makePostcondition tr name s) with
    |(None, _) => None
    |(_, None) => None
    |(Some pre, Some post) =>
     (match li with
       |[ ] => Some  [(Operation (transformTransitionLabel (getTransitionLabel tr)) pre (Select (ConsSelectList pre post EmptySelectList)))]
       |op::qLi =>
        (match op with
          |Operation label preOp postOp =>
           if labelsAreEqual label (getTransitionLabel tr) then
             (match postOp with
               |Select li => Some ((Operation label (Or pre preOp) (Select (ConsSelectList pre post li))):: qLi)
               |_ => None
             end)
           else
             match addTransition tr qLi name s with
               |None => None
               |Some li => Some (op :: li )
             end
        end)
     end)
  end.

Fixpoint translateAllTransitions (trans : list transition) (li : list operation) (name : string) (s : listOfAstdStates) :=
  match trans with
    |[] => Some li
    |tr::qTrans => match translateAllTransitions qTrans li name s with
                     |None => None
                     |Some trLi =>
                      addTransition tr trLi name s
                   end
    end.

Definition opLabelsAreEqual label1 label2 :=
  match (label1,label2) with
    |(Label name1 param1, Label name2 param2) =>
     if string_dec name1 name2 then
       paramAreEquals param1 param2
     else
       false
  end.

Fixpoint addOperation op opList :=
  match op with
    |Operation label pre post =>
     match opList with
       |[] => Some [op]
       |hOpList::qOpList =>
        match hOpList with
          |Operation labelHOp preHOp postHOp =>
           if opLabelsAreEqual label labelHOp then
             match postHOp with
               |Select li =>
                Some ((Operation label
                                 (Or pre preHOp)
                                 (Select (ConsSelectList pre
                                                         post
                                                         li)))::qOpList)
               |_ => None
             end
           else
             match (addOperation op qOpList) with
               |None => None
               |Some li => Some (hOpList::li)
             end
        end
     end
  end.

Fixpoint mergeOperationList aInB qLiInB :=
  match aInB with
    |[] => Some qLiInB
    |op::qAInB =>
     match (mergeOperationList qAInB qLiInB) with
       |None => None
       |Some li =>
        addOperation op li
     end
  end.

Fixpoint modifyPrePost aInB nameOfA nameOfState :=
  match aInB with
    |[] => []
    |op::qAInB =>
     match op with
       |Operation label pre post =>
        (Operation label
                   (And (Equality nameOfA nameOfState) pre)
                   (Select (ConsSelectList (And (Equality nameOfA nameOfState) pre)
                                           (Parallel (Affectation nameOfA nameOfState)
                                                     post) EmptySelectList)))::(modifyPrePost qAInB nameOfA nameOfState)
     end
  end.

Fixpoint translateAstd a :=
  match a with
    |Elem _ => Some []
    |(Automata name states trans sFinal dFinal init) =>
     match translateStateList states name with
       |None => None
       |Some li =>
        match translateAllTransitions trans nil name states with
          |Some opLi => mergeOperationList opLi li
          |None => None
        end
     end
  end
with translateStateList li name :=
       match li with
         |EmptyAstdStates => Some []
         |ConsAstdStates a qLi =>
          match (translateAstd a,translateStateList qLi name) with
            |(None,_) => None
            |(_,None) => None
            |(Some aInB,Some qLiInB) => 
             mergeOperationList (modifyPrePost aInB name (get_name a)) qLiInB
          end
       end.

Extraction "translation.ml" translateAstd.