Require Import B_langage.
Require Import B_state.
Require Import B_semantics.
Require Import translation.
Require Import ASTD.
Require Import String.

Inductive isWellConstructed : predicate -> substitution -> Prop :=
|IsWellConstructed_elem :
   forall (p : predicate)
          (s : substitution),
     isWellConstructed p (Select (ConsSelectList p s EmptySelectList))
|IsWellConstructed_cons :
   forall (p p0 : predicate)
          (li : selectList)
          (s : substitution),
     isWellConstructed p0 (Select li) ->
     isWellConstructed (Or p p0) (Select (ConsSelectList p s li)).

Inductive operationsAreWellConstructed : list operation -> Prop :=
|OperationsAreWellConstructed_nil :
   operationsAreWellConstructed nil
|OperationsAreWellConstructed_cons :
   forall (l : operation_label)
          (p : predicate)
          (s : substitution)
          (li : list operation),
     operationsAreWellConstructed li ->
     isWellConstructed p s ->
     operationsAreWellConstructed ((Operation l p s)::li).
          
Theorem wellConstructedGuardIsTrue :
  forall (li : selectList)
         (p : predicate)
         (v v' : bState),
    isWellConstructed p (Select li) ->
    substitution_step v (Select li) v' ->
    predicateIsTrueInState p v.
Proof.
  induction li.
  -
    intros p v v' WellCons Step.
    inversion Step.
  -
    intros p0 v v' WellCons Step.
    inversion Step;subst.
    +
      inversion WellCons;subst.
      *
        assumption.
      *
        apply (PredicateIsTrueInState_orLeft).
        assumption.
    +
      inversion WellCons;subst;[inversion H4 | ].
      apply (PredicateIsTrueInState_orRight).
      apply (IHli p2 v v' H1 H4).
Qed.

Lemma wellConstructed_preservedByAddOp :
  forall (a : operation)
         (li liRes : list operation),
    operationsAreWellConstructed li ->
    operationsAreWellConstructed (a::nil) ->
    addOperation a li = Some liRes ->
    operationsAreWellConstructed liRes.
Proof.
  induction li.
  -
    intros liRes WellConsLi WellConsA AddOp.
    simpl in AddOp.
    destruct a.
    inversion AddOp.
    apply (OperationsAreWellConstructed_cons).
    apply (OperationsAreWellConstructed_nil).
    inversion WellConsA.
    assumption.
  -
    intros liRes WellConsLi WellConsA AddOp.
    simpl in AddOp.
    destruct a.
    destruct a0.
    destruct (opLabelsAreEqual o o0).
    +
      destruct s0;inversion AddOp.
      inversion WellConsLi;subst.
      apply (OperationsAreWellConstructed_cons).
      assumption.
      apply (IsWellConstructed_cons).
      assumption.
    +
      case_eq (addOperation (Operation o p s) li);intros;rewrite H in AddOp;[ |discriminate].
      inversion AddOp;subst.
      inversion WellConsLi;subst.
      apply (OperationsAreWellConstructed_cons).
      apply (IHli l H2 WellConsA H).
      assumption.
Qed.

Lemma wellConstructed_preservedByMerge :
  forall (li1 li2 liRes : list operation),
    operationsAreWellConstructed li2 ->
    operationsAreWellConstructed li1 ->
    mergeOperationList li1 li2 = Some liRes ->
    operationsAreWellConstructed liRes.
Proof.
  induction li1.
  -
    intros li2 liRes WellConsLi2 WellConsLi1 Merge.
    simpl in Merge.
    inversion Merge;subst.
    assumption.
  -
    intros li2 liRes WellConsLi2 WellConsLi1 Merge.
    simpl in Merge.
    case_eq (mergeOperationList li1 li2);intros;rewrite H in Merge;[ |discriminate].
    inversion WellConsLi1;subst.
    apply (wellConstructed_preservedByAddOp (Operation l0 p s) l).
    +
      apply (IHli1 li2 l WellConsLi2 H2 H).
    +
      apply (OperationsAreWellConstructed_cons).
      apply (OperationsAreWellConstructed_nil).
      assumption.
    +
      assumption.
Qed.

Theorem addTransitionMakesProperWellConstructedOp :
  forall (t : transition)
         (opes opesRes : list operation)
         (name : string)
         (states : listOfAstdStates),
    operationsAreWellConstructed opes ->
    addTransition t opes name states = Some opesRes ->
    operationsAreWellConstructed opesRes.
Proof.
  induction opes.
  -
    intros opesRes name states WellCons AddTrans.
    simpl in AddTrans.
    case_eq (makePrecondition t name states);intros;rewrite H in AddTrans;[ | discriminate].
    case_eq (makePostcondition t name states);intros;rewrite H0 in AddTrans;[ | discriminate].
    inversion AddTrans.
    apply (OperationsAreWellConstructed_cons).
    +
      apply (OperationsAreWellConstructed_nil).
    +
      apply (IsWellConstructed_elem).
  -
    intros opesRes name states WellCons AddTrans.
    simpl in AddTrans.
    inversion WellCons;subst.
    case_eq (makePrecondition t name states);intros;rewrite H in AddTrans;[ | discriminate].
    case_eq (makePostcondition t name states);intros;rewrite H0 in AddTrans;[ | discriminate].
    destruct (labelsAreEqual l (getTransitionLabel t)).
    +
      destruct s;inversion AddTrans.
      apply (OperationsAreWellConstructed_cons).
      assumption.
      apply (IsWellConstructed_cons).
      assumption.
    +
      case_eq (addTransition t opes name states);intros;rewrite H3 in AddTrans;[ |discriminate].
      inversion AddTrans;subst.
      apply (OperationsAreWellConstructed_cons).
      apply (IHopes l0 name states H1 H3).
      assumption.
Qed.

Theorem translateAllTransitionsMakeWellConstructedOperations :
  forall (name : string)
         (trans : list transition)
         (states : listOfAstdStates)
         (opes : list operation),
    translateAllTransitions trans nil name states = Some opes ->
    operationsAreWellConstructed opes.
Proof.
  induction trans.
  -
    intros trans opes TranslateAll.
    simpl in TranslateAll.
    inversion TranslateAll.
    apply (OperationsAreWellConstructed_nil).
  -
    intros states opes TranslateAll.
    simpl in TranslateAll.
    case_eq (translateAllTransitions trans nil name states);intros;rewrite H in TranslateAll;[ | discriminate].
    apply (addTransitionMakesProperWellConstructedOp a l opes name states).
    apply (IHtrans states l H).
    assumption.
Qed.

Theorem modifyPrePostMakesWellConstructed :
  forall (opes : list operation)
         (name nameOfA : string),
    operationsAreWellConstructed (modifyPrePost opes name nameOfA).
Proof.
  induction opes.
  -
    intros name nameOfA.
    simpl.
    apply OperationsAreWellConstructed_nil.
  -
    intros name nameOfA.
    simpl.
    destruct a.
    apply (OperationsAreWellConstructed_cons).
    +
      apply (IHopes).
    +
      apply (IsWellConstructed_elem).
Qed.

Theorem translationIsWellConstructed :
  forall (a : astd)
         (opes : list operation),
    translateAstd a = Some opes ->
    operationsAreWellConstructed opes.
Proof.
  apply (astd_mut (fun (a : astd) =>
                     forall (opes : list operation),
                       translateAstd a = Some opes ->
                       operationsAreWellConstructed opes)
                  (fun (states : listOfAstdStates) =>
                     forall (opes : list operation)
                            (name : string),
                       translateStateList states name = Some opes ->
                       operationsAreWellConstructed opes)).
  -
    intros s opes Translate.
    simpl in Translate.
    inversion Translate;subst.
    apply OperationsAreWellConstructed_nil.
  -
    intros name states IHstates trans dF sF init opes.
    simpl.
    intros Translate.
    case_eq (translateStateList states name);intros;rewrite H in Translate;[ |discriminate].
    case_eq (translateAllTransitions trans nil name states);intros;rewrite H0 in Translate;[ |discriminate].
    apply (wellConstructed_preservedByMerge l0 l).
    apply (IHstates l name H).
    apply (translateAllTransitionsMakeWellConstructedOperations name trans states l0 H0).
    assumption.
  -
    intros opes name Translate.
    simpl in Translate.
    inversion Translate;subst.
    apply OperationsAreWellConstructed_nil.
  -
    intros a IHa states IHstates opes name.
    simpl.
    intros Translate.
    case_eq (translateAstd a);intros;rewrite H in Translate;[ |discriminate].
    case_eq (translateStateList states name);intros;rewrite H0 in Translate;[ |discriminate].
    apply (wellConstructed_preservedByMerge (modifyPrePost l name (get_name a)) l0).
    apply (IHstates l0 name H0).
    apply (modifyPrePostMakesWellConstructed).
    assumption.
Qed.


Theorem translateStateListIsWellConstructed :
  forall (states : listOfAstdStates)
         (name : string)
         (opes : list operation),
    translateStateList states name = Some opes ->
    operationsAreWellConstructed opes.
Proof.
  induction states.
  -
    intros name opes Translate.
    simpl in Translate.
    inversion Translate;subst.
    apply OperationsAreWellConstructed_nil.
  -
    intros name opes Translate.
    simpl in Translate.
    case_eq (translateAstd a);intros;rewrite H in Translate;[ |discriminate].
    case_eq (translateStateList states name);intros;rewrite H0 in Translate;[ |discriminate].
    apply (wellConstructed_preservedByMerge (modifyPrePost l name (get_name a)) l0).
    apply (IHstates name l0 H0).
    apply (modifyPrePostMakesWellConstructed).
    assumption.
Qed.